#ifndef __NATLIB_NLALLOC_H__
#define __NATLIB_NLALLOC_H__
/****************************************************************************/
/** nlalloc.h                                                              **/
/** ---------------------------------------------------------------------- **/
/** User defined memory allocator definitions                              **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** October 20, 1996 by Oleg Kagan                                         **/
/****************************************************************************/

//============================================================================
#include <nlobject.h>
#include <nllock.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

//============================================================================
class _NATLIB_DECLSPEC_ TAllocator : public TNatlibObject {
public:
	//------------------------------------------------------------------------
	typedef TNatlibObject TParent;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
		TException, TNatlibObject::TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE_LOW_MEMORY(TXLowMemory, TException, _NATLIB_DECLSPEC_);
	// May be thrown by some allocators if requested size can not be allocated
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE_OPERATION_IS_NOT_ALLOWED(
		TXInvalidAllocationSize, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OPERATION_IS_NOT_ALLOWED*/;

	//------------------------------------------------------------------------
	TAllocator();
	//------------------------------------------------------------------------
	~TAllocator();

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE(TAllocator) = 0;

	// resturns in r_nAllocated size that really have been allocated
	// r_nAllocated always bigger or equal than requested
	//------------------------------------------------------------------------
	virtual void* Allocate(size_t, size_t& r_nAllocated) = 0;
	//------------------------------------------------------------------------
	void* Allocate(size_t nRequested) 
	{
		size_t nAllocated; return Allocate(nRequested, nAllocated);
	} /*Allocate(size_t)*/
	//------------------------------------------------------------------------
	virtual void Deallocate(void*) = 0;

	// Used to access single allocation size in fixed containers and/or 
	// to optimize allocations (0 - means any)
	// Returns number of the resulting allocation size that is always bigger 
	// or equal to requested or 0. 
	// NOTE: May throw TXInvalidAllocationSize
	//------------------------------------------------------------------------
	virtual size_t SetAllocationSize(size_t nSize) {return 0;}
	//------------------------------------------------------------------------
	virtual size_t GetAllocationSize() const {return 0;}

	// May be called by user to optimize future allocations
	//------------------------------------------------------------------------
	virtual void Reserve(size_t) {}

	// returns ~size_t(0) if unable to calculate
	//------------------------------------------------------------------------
	virtual size_t Free() const {return ~size_t(0);};

	//------------------------------------------------------------------------
	virtual void CollectGarbage() {}
} /*TAllocator*/;

// TStdAllocator
//============================================================================
class _NATLIB_DECLSPEC_ TStdAllocator : public TAllocator {
public:
	//------------------------------------------------------------------------
	typedef TAllocator TParent;

	//------------------------------------------------------------------------
	TStdAllocator();
	//------------------------------------------------------------------------
	~TStdAllocator();

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE(TStdAllocator);

	//------------------------------------------------------------------------
	void* Allocate(size_t nRequested, size_t& r_nAllocated);
	//------------------------------------------------------------------------
	void Deallocate(void* pMemory);

	//------------------------------------------------------------------------
	nl_ulong_t Allocations() const {return nAllocations;}

protected:
	//------------------------------------------------------------------------
	nl_ulong_t nAllocations;
} /*TStdAllocator*/;

// TStdFixedAllocator
//============================================================================
class _NATLIB_DECLSPEC_ TStdFixedAllocator : public TStdAllocator {
public:
	//------------------------------------------------------------------------
	typedef TStdAllocator TParent;

	//------------------------------------------------------------------------
	TStdFixedAllocator(size_t nAllocationSize = 0);
	//------------------------------------------------------------------------
	~TStdFixedAllocator();

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE(TStdFixedAllocator);

	// throws TXInvalidAllocationSize if requested size is not equal 
	// to nAllowedSize
	//------------------------------------------------------------------------
	void* Allocate(size_t, size_t&);

	//------------------------------------------------------------------------
	size_t SetAllocationSize(size_t nSize) {return nAllowedSize = nSize;}
	//------------------------------------------------------------------------
	size_t GetAllocationSize() const {return nAllowedSize;}

protected:
	//------------------------------------------------------------------------
	size_t nAllowedSize;
} /*TStdFixedAllocator*/;

// TAllocationAware
// virtual base for any object that should be allocation aware
// (must be inherited virtually)
//============================================================================
class _NATLIB_DECLSPEC_ TAllocationAware {
public:
	// returns true if object has been allocated dynamically (e.g. using
	// operator new)
	//------------------------------------------------------------------------
	bool IsDynamic() const {return bDynamic;}

	//------------------------------------------------------------------------
	void* operator new(size_t);

protected:
	// used to process right this pointer in case of virtual inheritance
	//------------------------------------------------------------------------
	explicit TAllocationAware(void* pThis);

private:
	// TLastDynamicAllocation
	//------------------------------------------------------------------------
	typedef class _NATLIB_DECLSPEC_ TLastDynamicAllocation {
	public:
		//--------------------------------------------------------------------
		TLastDynamicAllocation();
		//--------------------------------------------------------------------
		~TLastDynamicAllocation();

		//--------------------------------------------------------------------
		void Set(void* pNew) {pLock->Lock(pNew); pLast = pNew;}
		//--------------------------------------------------------------------
		void Clear() {void* pKey = pLast; pLast = null; pLock->Unlock(pKey);}
		//--------------------------------------------------------------------
		void* Get() const {return pLast;}

		//--------------------------------------------------------------------
		void SetLock(TLock& r_Lock) {delete pLock; pLock = &r_Lock;}

	protected:
		//--------------------------------------------------------------------
		TLock* pLock;
		void* pLast;
	} typedef_TLastDynamicAllocation;

	//------------------------------------------------------------------------
	bool bDynamic;
	static TLastDynamicAllocation g_LastDynamicAllocation;
} /*TAllocationAware*/;

// TAllocationEnvelopeOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TAllocationEnvelopeOf {
public:
	//------------------------------------------------------------------------
	T Data;

	//------------------------------------------------------------------------
	TAllocationEnvelopeOf(const T&);
	//------------------------------------------------------------------------
	TAllocationEnvelopeOf(const TAllocationEnvelopeOf<T>&);
	//------------------------------------------------------------------------
	~TAllocationEnvelopeOf();
	
	//------------------------------------------------------------------------
	void* operator new(size_t, void* pPlacement) {return pPlacement;}
} /*TAllocationEnvelopeOf<T>*/;

// TPointerTo template
// Used to encapsulate pointers to use them as classes
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TPointerTo : 
	public virtual TNatlibObject 
{
public:
	//------------------------------------------------------------------------
	typedef TNatlibObject TParent;
	//------------------------------------------------------------------------
	typedef TPointerTo<T> TPointer;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
		TException, TNatlibObject::TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(TXNotAssigned, TException, _NATLIB_TEMPLATE_DECLSPEC_);
	NLEXCEPT_DECLARE_LOW_MEMORY(
		TXLowMemory, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;

	//------------------------------------------------------------------------
	TPointerTo() : pData(null) {}
	//------------------------------------------------------------------------
	explicit TPointerTo(T* pDataPtr) : pData(pDataPtr) {}
	//------------------------------------------------------------------------
	TPointerTo(const TPointer& c_Source) : pData(c_Source.pData) {}
	//------------------------------------------------------------------------
	~TPointerTo() {}

	//------------------------------------------------------------------------
	void AssertValid() const {TParent::AssertValid();}

	//------------------------------------------------------------------------
	TStdString Dump() const;

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TPointerTo, T);

	//------------------------------------------------------------------------
	void Assign(T* pDataNew) {pData = pDataNew;}
	//------------------------------------------------------------------------
	void Unassign() {pData = null;}
	//------------------------------------------------------------------------
	T& Dereference() const
	{
		if (!pData) 
			throw TXNotAssigned(THIS_CLASS);
		return *pData;
	} /*Dereference() const*/

	//------------------------------------------------------------------------
	bool IsNull() const {return !pData;}

	//------------------------------------------------------------------------
	TPointer& operator=(T* pDataNew) {Assign(pDataNew); return *this;}

	//------------------------------------------------------------------------
	TPointer& operator=(const TPointer& c_Source)
	{
		pData = c_Source.pData;
		return *this;
	} /*operator=(const TPointer&)*/

	//------------------------------------------------------------------------
	T& operator*() const {return Dereference();}
	//------------------------------------------------------------------------
	T* operator->() const {return &Dereference();}

	//------------------------------------------------------------------------
	operator T*() const {return pData;}

protected:
	//------------------------------------------------------------------------
	T* pData;
} /*TPointerTo<T>*/;

// TAllocationEnvelopeOf template implementation
//============================================================================
template <class T> TAllocationEnvelopeOf<T>::TAllocationEnvelopeOf(

	const T& c_Data
	
): 
	Data(c_Data) 
{} /*TAllocationEnvelopeOf<T>(const T&)*/

//----------------------------------------------------------------------------
template <class T> TAllocationEnvelopeOf<T>::TAllocationEnvelopeOf(

	const TAllocationEnvelopeOf<T>& c_Source
	
):
	Data(c_Source.Data)
{} /*TAllocationEnvelopeOf<T>(const TAllocationEvelopeOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> TAllocationEnvelopeOf<T>::~TAllocationEnvelopeOf() 
{} /*~TAllocationEnvelopeOf<T>()*/

// TPointerTo template implementation
//============================================================================
template <class T> TStdString TPointerTo<T>::Dump() const
{
	TStdString sDump(TParent::Dump());

	if (IsNull()) sDump += " is null";
	else {
		char szDump[1024];
		sprintf(szDump, " points to %s at %p", typeid(*pData).name(), pData);
		sDump += szDump;
	} /*if*/
	return sDump;
} /*TPointerTo<T>::Dump() const*/

//----------------------------------------------------------------------------
template <class T> NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TPointerTo, T)

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLALLOC_DECLSPEC(T, Declspec) \
	NLOBJECT_DECLSPEC(T, Declspec); \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TAllocationEnvelopeOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TPointerTo), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLALLOC_H__)*/
