#ifndef __NATLIB_NLAUTOPT_H__
#define __NATLIB_NLAUTOPT_H__
/****************************************************************************/
/** nlautopt.h                                                             **/
/** ---------------------------------------------------------------------- **/
/** Provides exception safe way to use operator new()                      **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** October 20, 1996 by Oleg Kagan                                         **/
/****************************************************************************/

//============================================================================
#include <nlobject.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TAutoPointerReferenceManagerOf template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TAutoPointerReferenceManagerOf {
public:
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE_BASE(TException, _NATLIB_TEMPLATE_DECLSPEC_);
	NLEXCEPT_DECLARE(
		TXInvalidDecrementOfReferenceFlag, TException, 
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(
		TXDestructWhileReferencesExist, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(TXNotAssigned, TException, _NATLIB_TEMPLATE_DECLSPEC_);

	//------------------------------------------------------------------------
	explicit TAutoPointerReferenceManagerOf(
		T* pInitData, bool bInitArray
	):
		nReferenceFlag(1), pData(pInitData), bArray(bInitArray)
	{
		if (!pData) 
			throw TXNotAssigned(THIS_CLASS);
	} /*TAutoPointerReferenceManagerOf(T*)*/

	//------------------------------------------------------------------------
	~TAutoPointerReferenceManagerOf()
	{
		if (nReferenceFlag) 
			throw TXDestructWhileReferencesExist(THIS_CLASS);
		if (bArray) delete[] pData;
		else delete pData;
		pData = null;
	} /*~TAutoPointerReferenceManagerOf()*/

	//------------------------------------------------------------------------
	nl_ulong_t References() const {return nReferenceFlag;}

	//------------------------------------------------------------------------
	T& Value() const {AssertValid(); return *pData;}

	//------------------------------------------------------------------------
	void IncReferences() {++nReferenceFlag;}

	//------------------------------------------------------------------------
	void DecReferences()
	{
		if (nReferenceFlag == 0) {
			nReferenceFlag = 1; // make nonexclusive (for robustness)
			throw TXInvalidDecrementOfReferenceFlag(THIS_CLASS);
		} /*if*/
		--nReferenceFlag;
	} /*DecReferences()*/

	//------------------------------------------------------------------------
	void AssertValid() const
	{
		if (!pData) 
			throw TXNotAssigned(THIS_CLASS);
	} /*AssertValid*/

	// Used by auto pointer to close reference manager without deleteing
	// the data
	//------------------------------------------------------------------------
	void Unassign() {pData = null;}

	//------------------------------------------------------------------------
	bool IsArray() const {return bArray;}

protected:
	// the real number of references is nReferencesFlag + 1
	//------------------------------------------------------------------------
	nl_ulong_t nReferenceFlag;
	//------------------------------------------------------------------------
	T* pData;
	bool bArray;
} /*TAutoPointerReferenceManagerOf<T>*/;

// TAutoPointerTo template
// Automaticaly counts references to object and deletes it if no more
// references exists
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TAutoPointerTo :
	public virtual TNatlibObject
{
public:
	//------------------------------------------------------------------------
	typedef TNatlibObject TParent;
	//------------------------------------------------------------------------
	typedef TAutoPointerReferenceManagerOf<T> TReferenceManager;
	//------------------------------------------------------------------------
	typedef TAutoPointerTo<T> TAutoPointer;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
		TException, TNatlibObject::TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(TXNotAssigned, TException, _NATLIB_TEMPLATE_DECLSPEC_);
	NLEXCEPT_DECLARE_LOW_MEMORY(
		TXLowMemory, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE_LOW_MEMORY*/;

	//------------------------------------------------------------------------
	TAutoPointerTo();
	//------------------------------------------------------------------------
	explicit TAutoPointerTo(T*, bool bArray);
	//------------------------------------------------------------------------
	TAutoPointerTo(const TAutoPointer&);
	//------------------------------------------------------------------------
	~TAutoPointerTo();

	//------------------------------------------------------------------------
	void AssertValid() const;

	//------------------------------------------------------------------------
	TStdString Dump() const;

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TAutoPointerTo, T);

	//------------------------------------------------------------------------
	bool IsArray() const;

	//------------------------------------------------------------------------
	void Assign(T*, bool bArray);

	// if bDestruct == false a pointing obect will not be destructed
	// return null if object have been destructed or bDestruct = true
	//------------------------------------------------------------------------
	T* Unassign(bool bDestruct = true);

	//------------------------------------------------------------------------
	T& Dereference() const;
	//------------------------------------------------------------------------
	T& RawDereference() const;

	//------------------------------------------------------------------------
	bool IsNull() const {return !pReferenceManager;}

	//------------------------------------------------------------------------
	TAutoPointer& operator=(T* pData)
	{
		Assign(pData, bArrayDefault); return *this;
	} /*operator=(T*)*/

	//------------------------------------------------------------------------
	TAutoPointer& operator=(const TAutoPointer& c_AutoPtr)
	{
		Assign(c_AutoPtr.pReferenceManager); return *this;
	} /*operator=(TAutoPointer&)*/

	//------------------------------------------------------------------------
	T& operator*() const {return Dereference();}
	//------------------------------------------------------------------------
	operator T*() const {return pReferenceManager ? &RawDereference() : null;}

protected:
	//------------------------------------------------------------------------
	TReferenceManager* pReferenceManager;
	bool bArrayDefault;

	//------------------------------------------------------------------------
	void Assign(TReferenceManager*);
} /*TAutoPointerTo<T>*/;

// TAutoPointerToClass template 
//============================================================================
template <class TClass> class _NATLIB_TEMPLATE_DECLSPEC_ TAutoPointerToClass : 
	public TAutoPointerTo<TClass> 
{
public:
	//------------------------------------------------------------------------
	typedef TAutoPointerTo<TClass> TParent;

	//------------------------------------------------------------------------
	TAutoPointerToClass() : TParent() {}
	//------------------------------------------------------------------------
	explicit TAutoPointerToClass(TClass* pClass, bool bArray) : 
		TParent(pClass, bArray) 
	{} /*TAutoPointerToClass(TClass*, bool)*/
	//------------------------------------------------------------------------
	TAutoPointerToClass(const TAutoPointerToClass<TClass>& r_hSource) :
		TParent(r_hSource)
	{} /*TAutoPointerToClass(const TAutoPointerToClass<TClass>&)*/

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TAutoPointerToClass, TClass);

	//------------------------------------------------------------------------
	TAutoPointerToClass<TClass>& operator=(
		const TAutoPointerToClass<TClass>& c_hSource
	){
		TParent::operator=(c_hSource); return *this;
	} /*operator=(const TAutoPointerToClass<TClass>&)*/
	//------------------------------------------------------------------------
	TAutoPointerToClass<TClass>& operator=(
		const TAutoPointerTo<TClass>& c_hSource
	){
		TParent::operator=(c_hSource); return *this;
	} /*operator=(const TAutoPointerTo<TClass>&)*/

	//------------------------------------------------------------------------
	TAutoPointerToClass<TClass>& operator=(TClass* pClass)
	{
		TParent::operator=(pClass); return *this;
	} /*operator=(T*)*/

	//------------------------------------------------------------------------
	TClass* operator->() const {return &Dereference();}
} /*TAutoPinterToClass<TClass>*/;

// TAutoPointerTo template implementation
//============================================================================
template <class T> TAutoPointerTo<T>::TAutoPointerTo() :
	pReferenceManager(null), bArrayDefault(false)
{} /*TAutoPointerTo<T>()*/

//----------------------------------------------------------------------------
template <class T> TAutoPointerTo<T>::TAutoPointerTo(

	T* pData,
	bool bArray

):
	pReferenceManager(null),
	bArrayDefault(bArray)
{
	Assign(pData, bArray);
} /*TAutoPointerTo<T>(T*, bool)*/

//----------------------------------------------------------------------------
template <class T> TAutoPointerTo<T>::TAutoPointerTo(

	const TAutoPointerTo<T>& c_AutoPointer

):
	pReferenceManager(null),
	bArrayDefault(c_AutoPointer.bArrayDefault)
{
	Assign(c_AutoPointer.pReferenceManager);
} /*TAutoPointerTo<T>(const TAutoPointer&)*/

//----------------------------------------------------------------------------
template <class T> TAutoPointerTo<T>::~TAutoPointerTo()
{
	Unassign(true);
} /*~TAutoPointerTo<T>()*/

//----------------------------------------------------------------------------
template <class T> void TAutoPointerTo<T>::AssertValid() const
{
	TParent::AssertValid();
	if (pReferenceManager) pReferenceManager->AssertValid();
} /*TAutoPointerTo<T>::AssertValid() const*/

//----------------------------------------------------------------------------
template <class T> TStdString TAutoPointerTo<T>::Dump() const
{
	TStdString sDump(TParent::Dump());

	if (IsNull()) sDump += " is null";
	else {
		// Get Address
		//-------------
		char szDump[1024];
		T* pData = &(pReferenceManager->Value());
		sprintf(szDump, " points to %s at %p", typeid(*pData).name(), pData);
		sDump += szDump;
	} /*if*/
	return sDump;
} /*TAutoPointerTo<T>::Dump() const*/

//----------------------------------------------------------------------------
template <class T> NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TAutoPointerTo, T)

//----------------------------------------------------------------------------
template <class T> bool TAutoPointerTo<T>::IsArray() const
{
	return (pReferenceManager && pReferenceManager->IsArray());
} /*TAutoPointerTo<T>::IsArray() const*/

//----------------------------------------------------------------------------
template <class T> void TAutoPointerTo<T>::Assign(

	T* pData,
	bool bArray

){
	bArrayDefault = bArray;
	Unassign(true);
	if (pData) {
		NLEXCEPT_CLASS_NEW(
			TReferenceManager(pData, bArray), pReferenceManager
		) /*NLEXCEPT_CLASS_NEW*/;
	} /*if*/
} /*TAutoPointerTo<T>::Assign(T*)*/

//----------------------------------------------------------------------------
template <class T> T* TAutoPointerTo<T>::Unassign(

	bool bDestruct

){
	T* pData = null;
	if (pReferenceManager) {
		pReferenceManager->DecReferences();
		if (!pReferenceManager->References()) {
			if (!bDestruct) {
				pData = &(pReferenceManager->Value());
				pReferenceManager->Unassign(); // Become invalid...
			} /*if*/
			delete pReferenceManager;
			pReferenceManager = null; // valid again...
		} /*if*/
	} /*if*/
	return pData;
} /*TAutoPointerTo<T>::Unassign()*/

//----------------------------------------------------------------------------
template <class T> T& TAutoPointerTo<T>::Dereference() const
{
	if (!pReferenceManager) 
		throw TXNotAssigned(THIS_CLASS);
	return pReferenceManager->Value();
} /*TAutoPointerTo<T>::Dereference() const*/

//----------------------------------------------------------------------------
template <class T> T& TAutoPointerTo<T>::RawDereference() const
{
	return pReferenceManager->Value();
} /*TAutoPointerTo<T>::RawDereference() const*/

//----------------------------------------------------------------------------
template <class T> void TAutoPointerTo<T>::Assign(

	TAutoPointerReferenceManagerOf<T>* pNewReferenceManager

){
	Unassign(true);
	if (pNewReferenceManager) {
		pNewReferenceManager->IncReferences();
		pReferenceManager = pNewReferenceManager;
	} /*if*/
} /*TAutoPointerTo<T>::Assign(T*)*/

// TAutoPointerToClass template implementation
//============================================================================
template <class TClass> 
NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TAutoPointerToClass, TClass);

//============================================================================	  
NATLIB_NAMESPACE_END

//============================================================================
#define NLAUTOPT_DECLSPEC(T, Declspec) \
	NLOBJECT_DECLSPEC(T, Declspec); \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TAutoPointerReferenceManagerOf), \
		T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TAutoPointerTo), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; 

#endif /*!defined(__NATLIB_NLAUTOPT_H__)*/
