#ifndef __NATLIB_NLAVLTRE_H__
#define __NATLIB_NLAVLTRE_H__
/****************************************************************************/
/** nlavltre.h                                                             **/
/** ---------------------------------------------------------------------- **/
/** Binary tree implementation based on AVL tree data structure            **/
/** (See N.Wirth "Algorithms and Data Structures")                         **/
/** Note: Copy constructor for template argument have to be available      **/
/** Only single element of specified value (by comparison) can be stored   **/
/** in the binary tree (set)                                               **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** October 22, 1996 by Oleg Kagan                                         **/
/****************************************************************************/

//============================================================================
#include <nlbintre.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

//============================================================================
#define NLAVLTRE_ALLOCATOR_DEFAULT NLBINTRE_ALLOCATOR_DEFAULT

// TAvlTreeSubdelete
// Internal tool used by InsertDelete while perform delete operation
//============================================================================
class _NATLIB_DECLSPEC_ TAvlTreeSubdelete {
public:
	//------------------------------------------------------------------------
	typedef TBinaryTreeNode TNode;

    // Balancing subtree after changes in left subtree
    //------------------------------------------------------------------------
    static bool BalanceL(TNode* &r_pNode);
    // -//-//- in right subtree
    //------------------------------------------------------------------------
    static bool BalanceR(TNode* &r_pNode);

    //------------------------------------------------------------------------
    TNode*& r_pNode;

    //------------------------------------------------------------------------
    explicit TAvlTreeSubdelete(TNode*& r_pDelNode) : r_pNode(r_pDelNode) {}

    //------------------------------------------------------------------------
    bool operator()(TNode*&);
} /*TAvlTreeSubdelete*/;

// TAvlTreeInsertDelete template
// provides ability to insert new data into the tree with balancing or
// delete data from the tree with balancing
// (Note: Only TAVLTree should operate it)
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TAvlTreeInsertDelete {
public:
    //------------------------------------------------------------------------
    typedef TBinaryTreeNode TNode;
	typedef TBinaryTreeNodeOf<T> TDataNode;
	typedef TBinaryTreeOf<T> TBinaryTree;
    typedef TBinaryTree::TSortingComparator TSortingComparator;
    typedef TBinaryTree::TSize TSize;

    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE_BASE(TXBadParameter, _NATLIB_TEMPLATE_DECLSPEC_);
    NLEXCEPT_DECLARE_BASE(TXNoDataDefined, _NATLIB_TEMPLATE_DECLSPEC_);

    // Constructing InsertDelete specifying tree that it should operate on
    // (Note: Specified tree must be TAvlTreeOf<T>
    //------------------------------------------------------------------------
    inline explicit TAvlTreeInsertDelete(
    	TNode*& r_pRoot, TSortingComparator&, TSize&
    ) /*TAvlTreeInsertDelete(TNode*&, TSortingComparator&, TSize&)*/;

    // Insert new node (if equal by sorting comparison data already exists in
    // the tree it will be replaced by the new)
    //------------------------------------------------------------------------
    void InsertNode(TNode&);

	// Deletes data from the tree
    //------------------------------------------------------------------------
    void DeleteData(const T&);

    // Internal
    //------------------------------------------------------------------------
    bool Insert(TNode*& r_pNode);
    //------------------------------------------------------------------------
    bool Delete(TNode*& r_pNode);

    //------------------------------------------------------------------------
    bool IsComplete() const {return bComplete;}

    //------------------------------------------------------------------------
    TNode* NodeToDeletePtr() const {return pNodeToDelete;}

protected:
    //------------------------------------------------------------------------
    TSortingComparator* pComparator;
    TNode** p_pRoot;
    TSize* p_nSize;
    bool bComplete;

	//------------------------------------------------------------------------
    union {
    	const T* pc_DataToDelete; // pointer to data to delete
        TNode* pNodeToInsert; // pointer to node to insert
    } /*union*/;
    TNode* pNodeToDelete; // pNode to delete
} /*TAvlTreeInsertDelete*/;

// TAvlTreeOf template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TAvlTreeOf : public TBinaryTreeOf<T> {
public:
	//------------------------------------------------------------------------
    typedef TBinaryTreeOf<T> TParent;

    //------------------------------------------------------------------------
    typedef TAvlTreeOf<T> TAvlTree;

    //------------------------------------------------------------------------
    typedef TAvlTreeInsertDelete<T> TInsertDelete;
    friend TInsertDelete;

    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
		TException, TBinaryTreeOf<T>::TException, 
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE(TXBadNode, TException, _NATLIB_TEMPLATE_DECLSPEC_);
	NLEXCEPT_DECLARE(
		TXHeightDifferenceTooBig, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE(
		TXBadBalanceFlag, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;

	//------------------------------------------------------------------------
    TAvlTreeOf(
		const TSortingComparator& c_Comparator,
		const TAllocator& = NLAVLTRE_ALLOCATOR_DEFAULT()
	) /*TAvlTreeOf*/;
    // Copy constructor
    //------------------------------------------------------------------------
    TAvlTreeOf(const TAvlTreeOf<T>&);
    //------------------------------------------------------------------------
    explicit TAvlTreeOf(const TBinaryTree&);
    //------------------------------------------------------------------------
    TAvlTreeOf(const TBinaryTree&, const TAllocator&);
    //------------------------------------------------------------------------
	TAvlTreeOf(
		const TContainer&, const TSortingComparator&, 
		const TAllocator& = NLAVLTRE_ALLOCATOR_DEFAULT()
	) /*TAvlTreeOf(const TContainer&, const TSortingComparator&, ...)*/;

    //------------------------------------------------------------------------
    void AssertValid() const;

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TAvlTreeOf, T);

	//------------------------------------------------------------------------
    inline TAvlTree& operator=(const TAvlTree& c_Source)
    {
    	Let(c_Source); return *this;
    } /*operator=(const TAvlTree&)*/
	//------------------------------------------------------------------------
    inline TAvlTree& operator=(const TContainer& c_Source)
	{
    	Let(c_Source); return *this;
    } /*operator=(const TContainerOf<T>&)*/

	//------------------------------------------------------------------------
    inline TAvlTree& operator+=(const T& c_Data)
    {
    	Insert(c_Data); return *this;
    } /*operator+=(const T&)*/
	//------------------------------------------------------------------------
    inline TAvlTree& operator+=(const TContainer& c_Source)
    {
    	Insert(c_Source); return *this;
    } /*operator+=(const TContainer&)*/
	//------------------------------------------------------------------------
    inline TAvlTree& operator-=(const TConditionOf<T>& c_Condition)
    {
    	Delete(c_Condition, indexAll); return *this;
    } /*operator-=(const TConditionOf<T>&)*/
    //------------------------------------------------------------------------
    inline TAvlTree& operator-=(const TComparatorOf<T>& c_Comparator)
    {
		Delete(c_Comparator, indexAll); return *this;
    } /*operator-=(const TComparatorOf<T>&)*/
    //------------------------------------------------------------------------
    inline TAvlTree& operator-=(const T& c_Data)
    {
    	Delete(c_Data); return *this;
    } /*operator-=(const T&)*/
    //------------------------------------------------------------------------
    inline TAvlTree& operator-=(const TContainer& c_Container)
    {
    	Delete(c_Container); return *this;
    } /*operator-=(const TContainer&)*/

protected:
	//------------------------------------------------------------------------
    virtual void InsertNode(TNode&, bool& r_bInserted);
    //------------------------------------------------------------------------
    virtual TSize DeleteNode(const TComparatorOf<T>&, TIndex);

    //------------------------------------------------------------------------
	bool DeleteNodeByData(const T&);

    //------------------------------------------------------------------------
    nl_long_t CheckNodeBalance(const TNode*) const;
} /*TAvlTreeOf*/;

// TAvlTreeInsertDelete template implementation
//============================================================================
template <class T> TAvlTreeInsertDelete<T>::TAvlTreeInsertDelete(

    TBinaryTreeNode*& r_pRoot,
    TEtalonComparatorOf<T>& r_Comparator,
    nl_ulong_t& r_nSize

):
	p_pRoot(&r_pRoot), pComparator(&r_Comparator), p_nSize(&r_nSize),
    bComplete(false), pNodeToDelete(null)
{} /*TAvlTreeInsertDelete(TNode*&, TSize&, TSortingComparator&)*/

//----------------------------------------------------------------------------
template <class T> void TAvlTreeInsertDelete<T>::InsertNode(

	TBinaryTreeNode& r_Node

){
	bComplete = false;

	// Prepare insert node and pointer to it
    //---------------------------------------
    pNodeToInsert = &r_Node;
    r_Node.pLeft = r_Node.pRight = null;
    r_Node.Balance = TNode::balanceEqual;

    // Prepare insert comparison (using tree sorting comparison)
    //-----------------------------------------------------------
    TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
    	*pComparator
    ) /*hComparator*/;
    hComparator->SetEtalon(((TDataNode&)r_Node).Data);

    // Perform insert operation
    //--------------------------
    Insert(*p_pRoot);

    // Prepare to exit
    //-----------------
    pNodeToInsert = null;
} /*TAvlTreeInsertDelete<T>::InsertNode(TNode*)*/

//----------------------------------------------------------------------------
template <class T> void TAvlTreeInsertDelete<T>::DeleteData(

	const T& c_Data

){
	bComplete = false;

    // Prepare insert comparison (using tree sorting comparison)
	//-----------------------------------------------------------
	TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
    	*pComparator
    ) /*hComparator*/;
    hComparator->SetEtalon(c_Data);

	// Perform delete operation
	//--------------------------
    Delete(*p_pRoot);
} /*TAvlTreeInsertDelete<T>::DeleteByComparator()*/

//----------------------------------------------------------------------------
template <class T> bool TAvlTreeInsertDelete<T>::Insert(

	TBinaryTreeNode*& r_pNode

){
	bool bRet;
	TNode* p1;
    TNode* p2;

	if (r_pNode == null) {
    	// Connect node
        //--------------
		r_pNode = pNodeToInsert;

		// Creation & insertion are OK ...
		//---------------------------------
		++(*p_nSize);
        bComplete = true;
		bRet = true;
	}
	else {
		switch ((*pComparator)(((TDataNode*&)r_pNode)->Data)) {
			case TSortingComparator::compLess: {
				if ((bRet = Insert(r_pNode->pLeft)) != 0) {
					// Left subtree bigger
					//---------------------
					switch (r_pNode->Balance) {
						case TNode::balanceRight :
							r_pNode->Balance = TNode::balanceEqual;
                            bRet = false;
                            break;

						case TNode::balanceEqual :
							r_pNode->Balance = TNode::balanceLeft;
                            break;

						case TNode::balanceLeft :
							// Balance
							//---------
							p1 = r_pNode->pLeft;
							if (p1->Balance == TNode::balanceLeft) {
								// Once LL-Rotate
								//----------------
								r_pNode->pLeft = p1->pRight;
								p1->pRight = r_pNode;
                                r_pNode->Balance = TNode::balanceEqual;
								r_pNode = p1;
							}
							else {
								// Double LR-Rotate
								//------------------
								p2 = p1->pRight;
								p1->pRight = p2->pLeft;
								p2->pLeft = p1;
                                r_pNode->pLeft = p2->pRight;
								p2->pRight = r_pNode;
								r_pNode->Balance = (
									(p2->Balance == TNode::balanceLeft) ?
										TNode::balanceRight :
										TNode::balanceEqual
                                ) /*r_pNode->Balance*/;
								p1->Balance = (
									(p2->Balance == TNode::balanceRight) ?
										TNode::balanceLeft :
										TNode::balanceEqual
                                ) /*p1->Balance*/;
								r_pNode = p2;
							} /*if*/
							r_pNode->Balance = TNode::balanceEqual;
							bRet = false;
					} /*switch (r_pNode->Balance)*/
				} /*if*/
				break;
			} /*case TSortingComparator::compLess*/

			case TSortingComparator::compGreater : {
				// Right subtree bigger
				//----------------------
				if ((bRet = Insert(r_pNode->pRight)) != 0) {
					switch (r_pNode->Balance) {
						case TNode::balanceLeft:
							r_pNode->Balance = TNode::balanceEqual;
                            bRet = false;
                            break;

						case TNode::balanceEqual:
							r_pNode->Balance = TNode::balanceRight;
                            break;

						case TNode::balanceRight:
							// Balance
							//---------
							p1 = r_pNode->pRight;
							if (p1->Balance == TNode::balanceRight) {
								// Once RR-Rotate
								//----------------
								r_pNode->pRight = p1->pLeft;
								p1->pLeft = r_pNode;
                                r_pNode->Balance = TNode::balanceEqual;
								r_pNode = p1;
							}
							else {
								// Double RR-Rotate
								//------------------
								p2 = p1->pLeft;
								p1->pLeft = p2->pRight;
								p2->pRight = p1;
                                r_pNode->pRight = p2->pLeft;
								p2->pLeft = r_pNode;
								r_pNode->Balance = (
									(p2->Balance == TNode::balanceRight) ?
										TNode::balanceLeft :
										TNode::balanceEqual
                                ) /*r_pNode->Balance*/;
								p1->Balance = (
									(p2->Balance == TNode::balanceLeft) ?
										TNode::balanceRight :
										TNode::balanceEqual
                                ) /*p1->Balance*/;
								r_pNode = p2;
							} /*if*/
							r_pNode->Balance = TNode::balanceEqual;
							bRet = false;
					} /*switch (r_pNode->nBalance)*/
				} /*if*/
				break;
			} /*case TSortingComparator::compGreater*/

			case TSortingComparator::compEqual: {
            	// Prepare new node
                //------------------
				pNodeToInsert->pLeft = r_pNode->pLeft;
                pNodeToInsert->pRight = r_pNode->pRight;
                pNodeToInsert->Balance = r_pNode->Balance;

                // Connect node
                //--------------
                pNodeToDelete = r_pNode;
                r_pNode = pNodeToInsert;

                bComplete = true;
                bRet = false;
				break;
			} /*TSortingComparator::compEqual*/

			default: {
				throw TAvlTreeOf<T>::TXUndefinedComparisonResult(
                    THIS_CLASS
				) /*TAvlTreeOf<T>::TXUndefinedComparisonResult*/;
			} /*default*/
		} /*switch*/
	} /*if*/
	return bRet;
} /*TAvlTreeInsertDelete<T>::Insert(TNode*&)*/

//----------------------------------------------------------------------------
template <class T> bool TAvlTreeInsertDelete<T>::Delete(

	TBinaryTreeNode*& r_pNode

){
	bool bRet;
	if (r_pNode == null) {
		// Key not found
		//---------------
		bRet = false;
	}
	else {
		switch ((*pComparator)(((TDataNode*)r_pNode)->Data)) {
			case TSortingComparator::compLess: {
				if ((bRet = Delete(r_pNode->pLeft)) != 0)
					bRet = TAvlTreeSubdelete::BalanceL(r_pNode);
				break;
			} /*case TSortingComparator::compLess*/

			case TSortingComparator::compGreater: {
				if ((bRet = Delete(r_pNode->pRight)) != 0)
				bRet = TAvlTreeSubdelete::BalanceR(r_pNode);
				break;
			} /*case TSortingComparator::compGreater*/

			case TSortingComparator::compEqual: {
				// Erase *r_pNode
				//----------------
				TNode* pq = r_pNode;
				if (pq->pRight == null) {
					r_pNode = pq->pLeft;
					bRet = true;
				}
				else if (!pq->pLeft) {
					r_pNode = pq->pRight;
					bRet = true;
				}
				else {
					TAvlTreeSubdelete SubDelete(r_pNode);
					if ((bRet = SubDelete(pq->pLeft)) != 0) {}
					r_pNode->pLeft = pq->pLeft;

					if (bRet) bRet = TAvlTreeSubdelete::BalanceL(r_pNode);
				} /*if*/
				pNodeToDelete = pq;
				--(*p_nSize);
                bComplete = true;
				break;
			} /*case TSortingComparator::compEqual*/

			default: {
				throw TAvlTreeOf<T>::TXUndefinedComparisonResult(
                	THIS_CLASS
                ) /*throw TAvlTree::TXUndefinedComparisonResult*/;
			} /*default*/
		} /*switch*/
	} /*if*/
	return bRet;
} /*TAvlTreeInsertDelete<T>::Delete(TNode*&)*/

// TAvlTreeOf template implementation
//============================================================================
template <class T> TAvlTreeOf<T>::TAvlTreeOf(

	const TEtalonComparatorOf<T>& c_Comparator,
	const TAllocator& c_Allocator /*= NLAVLTRE_ALLOCATOR_DEFAULT()*/

):
	TParent(c_Comparator, c_Allocator)
{} /*TAvlTreeOf<T>(const TSortingComparator&, const TAllocator&)*/

//----------------------------------------------------------------------------
template <class T> TAvlTreeOf<T>::TAvlTreeOf(

	const TAvlTreeOf<T>& c_Tree

):
	TParent(c_Tree.ComparatorConst(), c_Tree.AllocatorConst())
{
	Insert(c_Tree);
} /*TAvlTreeOf<T>(const TAvlTree&)*/

//----------------------------------------------------------------------------
template <class T> TAvlTreeOf<T>::TAvlTreeOf(

	const TBinaryTreeOf<T>& c_Tree

):
	TParent(c_Tree.ComparatorConst(), c_Tree.AllocatorConst())
{
	Insert(c_Tree);
} /*TAvlTreeOf<T>(const TAvlTree&)*/

//----------------------------------------------------------------------------
template <class T> TAvlTreeOf<T>::TAvlTreeOf(

	const TBinaryTreeOf<T>& c_Tree,
	const TAllocator& c_Allocator

):
	TParent(c_Tree.ComparatorConst(), c_Allocator)
{
	Insert(c_Tree);
} /*TAvlTreeOf<T>(const TAvlTree&, const TAllocator&)*/

//----------------------------------------------------------------------------
template <class T> TAvlTreeOf<T>::TAvlTreeOf(

	const TContainerOf<T>& c_Container,
    const TEtalonComparatorOf<T>& c_Comparator,
	const TAllocator& c_Allocator /*= NLAVLTRE_ALLOCATOR_DEFAULT()*/

):
	TParent(c_Comparator, c_Allocator)
{
	Insert(c_Container);
} /*TAvlTreeOf<T>(const TContainer&, const TSortingComparator&, ...)*/

//----------------------------------------------------------------------------
template <class T> void TAvlTreeOf<T>::AssertValid() const
{
    TParent::AssertValid();
    if (pRoot) CheckNodeBalance(pRoot);
} /*TAvlTreeOf<T>::AssertValid() const*/

//----------------------------------------------------------------------------
template <class T> NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TAvlTreeOf, T)

//----------------------------------------------------------------------------
template <class T> void TAvlTreeOf<T>::InsertNode(

	TBinaryTreeNode& r_Node,
	bool& r_bInserted

){
	r_bInserted = false;
	TAvlTreeInsertDelete<T> Inserter(pRoot, Comparator(), nSize);
    Inserter.InsertNode(r_Node);
    r_bInserted = Inserter.IsComplete();
    DeallocateNode(Inserter.NodeToDeletePtr());
} /*TAvlTreeOf<T>::InsertNode(TNode&, bool&)*/;

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TAvlTreeOf<T>::DeleteNode(

	const TComparatorOf<T>& c_Comparator,
    nl_ulong_t nIndex 

){
	TSize nDeleted = 0;
	if (nIndex != indexAll) {
		TNode* pNodeToDelete = GetNodePtr(pRoot, c_Comparator, nIndex);

        if (pNodeToDelete) {
			if (!DeleteNodeByData(((TDataNode*)pNodeToDelete)->Data)) 
				throw TException(THIS_CLASS);
        	++nDeleted;
        } /*if*/
    } /*if*/
    else {
   		const TEtalonComparator* pc_EtalonComparator = (
        	dynamic_cast<const TEtalonComparator*>(&c_Comparator)
        ) /*pc_EtalonComparator*/;

    	TNode* pNodeToDelete = GetAnyNodePtr(pRoot, c_Comparator);

        while (pNodeToDelete) {
        	if (
            	pc_EtalonComparator &&
                (
					pc_EtalonComparator->EtalonPtr() == 
					&(((TDataNode*)pNodeToDelete)->Data)
				)
            ){
				throw TXAttemptToDeleteReferencedData(THIS_CLASS);
			} /*if*/

			if (!DeleteNodeByData(((TDataNode*)pNodeToDelete)->Data)) 
				throw TException(THIS_CLASS);

			++nDeleted;
            pNodeToDelete = GetAnyNodePtr(pRoot, c_Comparator);
        } /*while*/
    } /*if*/
    return nDeleted;
} /*TAvlTreeOf<T>::DeleteNode(const TComparatorOf<T>&, TIndex)*/

//----------------------------------------------------------------------------
template <class T> bool TAvlTreeOf<T>::DeleteNodeByData(

	const T& c_Data

){
    TAvlTreeInsertDelete<T> Deleter(pRoot, Comparator(), nSize);
    Deleter.DeleteData(c_Data);

    if (!Deleter.IsComplete()) return false;

    DeallocateNode(Deleter.NodeToDeletePtr());
    return true;
} /*TAvlTreeOf<T>::DeleteNodeByData(const T&)*/

//----------------------------------------------------------------------------
template <class T> nl_long_t TAvlTreeOf<T>::CheckNodeBalance(

	const TBinaryTreeNode* pc_Node

) const {
	if (pc_Node) {
		TNode* pLeft = pc_Node->pLeft;
		if (pc_Node->pLeft && !pLeft) 
			throw TXBadNode(THIS_CLASS);

		TNode* pRight = pc_Node->pRight;
		if (pc_Node->pRight && !pRight) 
			throw TXBadNode(THIS_CLASS);

		nl_ulong_t nLeftHeight = CheckNodeBalance(pLeft);
		nl_ulong_t nRightHeight = CheckNodeBalance(pRight);

		nl_long_t nBalance = nl_long_t(nRightHeight) - nl_long_t(nLeftHeight);

		if ((nBalance > nl_long_t(1)) || (nBalance < nl_long_t(-1)))
			throw TXHeightDifferenceTooBig(THIS_CLASS);

		if (nBalance != nl_long_t(pc_Node->Balance)) 
			throw TXBadBalanceFlag(THIS_CLASS);

		return (
			1 + ((nLeftHeight > nRightHeight) ? nLeftHeight : nRightHeight)
		) /*return*/;
	}
	else return 0;
} /*TAvlTreeOf<T>::CheckNodeBalance(const TNode*) const*/

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLAVLTRE_DECLSPEC(T, Declspec) \
	NLBINTRE_DECLSPEC(T, Declspec); \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TAvlTreeInsertDelete), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
    NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TAvlTreeOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLAVLTRE_H__)*/
