#ifndef __NATLIB_NLBINTRE_H__
#define __NATLIB_NLBINTRE_H__
/****************************************************************************/
/** nlbintre.h                                                             **/
/** ---------------------------------------------------------------------- **/
/** Container implementation based on binary tree data structure           **/
/** Note: Copy contructor for template argument must be available          **/
/** Only single element of specified value (by comparison) can be stored   **/
/** in the binary tree (set)                                               **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** October 22, 1996 by Oleg Kagan                                         **/
/****************************************************************************/

//============================================================================
#include <nlscont.h>
#include <nlalloc.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

//============================================================================
#define NLBINTRE_ALLOCATOR_DEFAULT TStdFixedAllocator

// TBinaryTreeNode
//============================================================================
class _NATLIB_DECLSPEC_ TBinaryTreeNode {
public:
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE_BASE(TXBadParameter, _NATLIB_DECLSPEC_);

	//------------------------------------------------------------------------
	static TBinaryTreeNode*& GetMaxNodePtr(TBinaryTreeNode*&);
	//------------------------------------------------------------------------
	static TBinaryTreeNode*& GetMinNodePtr(TBinaryTreeNode*&);
	// Disconnects specified node from the tree returns pointer to node that
	// should be deleted
	//------------------------------------------------------------------------
	static TBinaryTreeNode* DisconnectNode(TBinaryTreeNode*&);

	// TBalance
    //------------------------------------------------------------------------
    typedef enum {
    	balanceLeft = -1,
        balanceEqual = 0,
        balanceRight = 1
    } TBalance;

    //------------------------------------------------------------------------
    TBalance Balance;

	//------------------------------------------------------------------------
	TBinaryTreeNode* pLeft;  // Left subtree pointer
	TBinaryTreeNode* pRight;  // Right subtree pointer

	//------------------------------------------------------------------------
	TBinaryTreeNode() : Balance(balanceEqual), pLeft(null), pRight(null) {}
} /*TBinaryTreeNode*/;

// TBinaryTreeNode template
// Used by TBinaryTree template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TBinaryTreeNodeOf :
	public TBinaryTreeNode
{
public:
	//------------------------------------------------------------------------
	T Data;

	//------------------------------------------------------------------------
	explicit TBinaryTreeNodeOf(const T& c_Data) : Data(c_Data) {}
	//------------------------------------------------------------------------
	~TBinaryTreeNodeOf() {}

	//------------------------------------------------------------------------
	void* operator new(size_t, void* pPlacement) {return pPlacement;}
} /*TBinaryTreeNodeOf<T>*/;

// TBinaryTreeExecutorOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TBinaryTreeExecutorOf {
public:
	//------------------------------------------------------------------------
	typedef TOperationOn< TBinaryTreeNodeOf<T> > TNodeOperation;

	//------------------------------------------------------------------------
	explicit TBinaryTreeExecutorOf(TNodeOperation& r_NodeOperation) :
		pOperation(&r_NodeOperation)
	{} /*TBinaryTreeExecutorOf(TNodeOperation&)*/

	//------------------------------------------------------------------------
	void Preorder(TBinaryTreeNode*);
	//------------------------------------------------------------------------
	void Inorder(TBinaryTreeNode*);
	//------------------------------------------------------------------------
	void Postorder(TBinaryTreeNode*);

protected:
	//------------------------------------------------------------------------
	TNodeOperation* pOperation;
} /*TBinaryTreeExecutorOf<T>*/;

// TBinaryTreeApplyOperationOn template
//============================================================================
template <class T> 
class _NATLIB_TEMPLATE_DECLSPEC_ TBinaryTreeApplyOperationOn :
	public TOperationOn< TBinaryTreeNodeOf<T> >
{
public:
	//------------------------------------------------------------------------
	typedef TOperationOn< TBinaryTreeNodeOf<T> > TParent;

	//------------------------------------------------------------------------
	typedef TBinaryTreeNodeOf<T> TNode;
	typedef TOperationOn<T> TDataOperation;

	//------------------------------------------------------------------------
	TBinaryTreeApplyOperationOn(TDataOperation& r_DataOperation) :
		pOperation(&r_DataOperation)
	{} /*TBinaryTreeApplyOperationOn(TDataOperation&)*/

	//------------------------------------------------------------------------
	void operator()(TNode& r_Node) {pOperation->operator()(r_Node.Data);}

protected:
	//------------------------------------------------------------------------
	TDataOperation* pOperation;

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(TBinaryTreeApplyOperationOn, T)
} /*TBinaryTreeApplyOperationOn<T>*/;

// TBinaryTreeApplyConstantOperationOn template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TBinaryTreeApplyConstantOperationOn :
	public TConstantOperationOn< TBinaryTreeNodeOf<T> >
{
public:
	//------------------------------------------------------------------------
	typedef TConstantOperationOn< TBinaryTreeNodeOf<T> > TParent;

	//------------------------------------------------------------------------
	typedef TBinaryTreeNodeOf<T> TNode;
	typedef TConstantOperationOn<T> TConstantDataOperation;

	//------------------------------------------------------------------------
	TBinaryTreeApplyConstantOperationOn(
		TConstantDataOperation& r_DataOperation
	) :
		pOperation(&r_DataOperation)
	{} /*TBinaryTreeApplyConstantOperationOn(TConstantDataOperation&)*/

	//------------------------------------------------------------------------
	void operator()(const TNode& c_Node) 
	{
		pOperation->operator()(c_Node.Data);
	} /*operator()(const TNode&)*/

protected:
	//------------------------------------------------------------------------
	TConstantDataOperation* pOperation;

	//------------------------------------------------------------------------
	void operator()(TNode& r_Node) {operator()((const TNode&)r_Node);}

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(TBinaryTreeApplyConstantOperationOn, T)
} /*TBinaryTreeApplyConstantOperationOn<T>*/;

// TBinaryTreeCheckExistenceOf template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TBinaryTreeCheckExistenceOf {
public:
	//------------------------------------------------------------------------
	typedef TBinaryTreeNode TNode;
	typedef TBinaryTreeNodeOf<T> TDataNode;
	typedef TContainerOf<T>::TSize TSize;

	//------------------------------------------------------------------------
	explicit TBinaryTreeCheckExistenceOf(
		const TComparatorOf<T>& c_Comparator
	):
		pc_Comparator(&c_Comparator), nExist(0)
	{} /*TBinaryTreeCheckExistenceOf(const TComparatorOf<T>&)*/

	//------------------------------------------------------------------------
	TSize operator()(const TNode* pc_Node)
	{
		nExist = 0; Check(pc_Node);
		return nExist;
	} /*operator()(const TNode*)*/

protected:
	//------------------------------------------------------------------------
	const TComparatorOf<T>* pc_Comparator;
	TSize nExist;

protected:
	//------------------------------------------------------------------------
	void Check(const TNode* pc_Node);
} /*TBinaryTreeCheckExistenceOf<T>*/;

// TBinaryTreeRebuildHelper
// Helps to rebuild binary tree structure
//============================================================================
class _NATLIB_DECLSPEC_ TBinaryTreeRebuildHelper {
public:
	//------------------------------------------------------------------------
	TBinaryTreeRebuildHelper() {}

	// Returns head of insert list (pRight is Next in list, pLef == null)
	//------------------------------------------------------------------------
	TBinaryTreeNode* BuildInsertList(TBinaryTreeNode* pRoot);

protected:
	//------------------------------------------------------------------------
	TBinaryTreeNode* pInsertPlace;

	//------------------------------------------------------------------------
	void PreorderList(TBinaryTreeNode*);
} /*TBinaryTreeRebuildHelper*/;

// TBinaryTreeOf template
// Note: For all functions that take comparison as parameter the comparison
// must be compatible with sorting comparison specified on construction of
// the tree
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TBinaryTreeOf :
	public TSortingContainerOf<T>
{
public:
	//------------------------------------------------------------------------
	typedef TSortingContainerOf<T> TParent;
	typedef TBinaryTreeOf<T> TBinaryTree;

	//------------------------------------------------------------------------
	typedef TBinaryTreeNode TNode;
	typedef TBinaryTreeNodeOf<T> TDataNode;

	//------------------------------------------------------------------------
	typedef TBinaryTreeCheckExistenceOf<T> TCheckExistence;

	//------------------------------------------------------------------------
	typedef TOperationOn< TBinaryTreeNodeOf<T> > TNodeOperation;
	typedef TConstantOperationOn< TBinaryTreeNodeOf<T> > 
		TNodeConstantOperation;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
		TException, TSortingContainerOf<T>::TException, 
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(
		TXAttemptToDeleteReferencedData, TException, 
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(
		TXAllocatorIsNotDefined, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;

	// Initializes tree structure with specified sorting comparison
	//------------------------------------------------------------------------
	TBinaryTreeOf(
		const TSortingComparator&, 
		const TAllocator& = NLBINTRE_ALLOCATOR_DEFAULT()
	) /*TBinaryTreeOf(const TSortingComparator&, const TAllocator&)*/;
	//------------------------------------------------------------------------
	TBinaryTreeOf(const TBinaryTree&);
	//------------------------------------------------------------------------
	TBinaryTreeOf(
		const TContainer&, const TSortingComparator&, 
		const TAllocator& = NLBINTRE_ALLOCATOR_DEFAULT()
	) /*TBinaryTreeOf(const TContainer&, const TSortingComparator&, ...)*/;

	//------------------------------------------------------------------------
	~TBinaryTreeOf();

	//------------------------------------------------------------------------
	void AssertValid() const;

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TBinaryTreeOf, T);

	// Inserts copy of specified data into the tree
	// (previous copy of specified data will be deleted)
	//------------------------------------------------------------------------
	const T& Insert(const T&);
	// Inserts all elements form specified container
	//------------------------------------------------------------------------
	void Insert(const TContainer&);

	//------------------------------------------------------------------------
	TSize Size() const {return nSize;}
	//------------------------------------------------------------------------
	bool IsEmpty() const {return !pRoot;}

	//------------------------------------------------------------------------
	const T* FastSearch(const TComparatorOf<T>&, TIndex = indexAny) const;
	//------------------------------------------------------------------------
	const T* FastSearch(const T& c_Data) const
	{
		return TParent::FastSearch(c_Data);
	} /*FastSearch(const T&) const*/

	//------------------------------------------------------------------------
	const T* SearchFirst() const;
	//------------------------------------------------------------------------
	const T* SearchLast() const;

	//------------------------------------------------------------------------
	const T* SearchNext(const T&) const;
	//------------------------------------------------------------------------
	const T* SearchPrevious(const T&) const;

	//------------------------------------------------------------------------
	const T* SearchNext(const TComparatorOf<T>&) const;
	//------------------------------------------------------------------------
	const T* SearchPrevious(const TComparatorOf<T>&) const;

	//------------------------------------------------------------------------
	TSize FastExist(const TComparatorOf<T>& c_Comparator) const;
	//------------------------------------------------------------------------
	TSize FastExist(const T& c_Data) const
	{
		return TParent::FastExist(c_Data);
	} /*FastExist(const T&) const*/

	// Deletes data that follows specified condition, using index to determine
	// which data should be deleted (see TSortingContainerOf<T>),
	// returns number of deleted items
	//------------------------------------------------------------------------
	TSize FastDelete(const TComparatorOf<T>&, TIndex = indexAll);

	//------------------------------------------------------------------------
	void Clear();

	//------------------------------------------------------------------------
	virtual void Preorder(TOperation&);
	//------------------------------------------------------------------------
	void Inorder(TOperation&);
	//------------------------------------------------------------------------
	virtual void Postorder(TOperation&);

	//------------------------------------------------------------------------
	virtual void Preorder(TConstantOperation&) const;
	//------------------------------------------------------------------------
	void Inorder(TConstantOperation&) const;
	//------------------------------------------------------------------------
	virtual void Postorder(TConstantOperation&) const;

	// Sets new sorting comparison and rebuilds the tree
	//------------------------------------------------------------------------
	void SetComparator(const TSortingComparator&);

	//------------------------------------------------------------------------
	const TAllocator& AllocatorConst() const {return *pAllocator;}

	//------------------------------------------------------------------------
	inline TBinaryTree& operator=(const TBinaryTree& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const TBinaryTree&)*/
	//------------------------------------------------------------------------
	inline TBinaryTree& operator=(const TContainer& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const TContainer&)*/

	//------------------------------------------------------------------------
	inline TBinaryTree& operator+=(const T& c_Data)
	{
		Insert(c_Data); return *this;
	} /*operator+=(const T&)*/
	//------------------------------------------------------------------------
	inline TBinaryTree& operator+=(const TContainer& c_Container)
	{
		Insert(c_Container); return *this;
	} /*operator+=(const TContainer&)*/
	//------------------------------------------------------------------------
	inline TBinaryTree& operator-=(const TConditionOf<T>& c_Condition)
	{
		Delete(c_Condition, indexAll); return *this;
	} /*operator-=(const TConditionOf<T>&)*/
	//------------------------------------------------------------------------
	inline TBinaryTree& operator-=(const TComparatorOf<T>& c_Comparator)
	{
		Delete(c_Comparator, indexAll); return *this;
	} /*operator-=(const TComparatorOf<T>&)*/
	//------------------------------------------------------------------------
	inline TBinaryTree& operator-=(const T& c_Data)
	{
		Delete(c_Data); return *this;
	} /*operator-=(const T&)*/
	//------------------------------------------------------------------------
	inline TBinaryTree& operator-=(const TContainer& c_Container)
	{
		Delete(c_Container); return *this;
	} /*operator-=(const TContainer&)*/

protected:
	//------------------------------------------------------------------------
	TNode* pRoot;
	TSize nSize;
	TAllocator* pAllocator;

	// Default constructor for children
	//------------------------------------------------------------------------
	TBinaryTreeOf() : pRoot(null), nSize(0) {}

	// Inserts allocated node into the tree structure
	// (if node with similar data already exists in the tree then replace it)
	// r_bInserted output is true if node have been already inserted in the
	// tree in case of exception (whithout exception it is always true)
	//------------------------------------------------------------------------
	virtual void InsertNode(TNode&, bool& r_bInserted);

	// Deletes nodes that follows the specified comparison
	// (returns number of deleted nodes)
	//------------------------------------------------------------------------
	virtual TSize DeleteNode(const TComparatorOf<T>&, TIndex);

	// Disconnects and dealocates node if parameter isn't null,
	// decremens size of the tree (used for to delete node from the tree)
	//------------------------------------------------------------------------
	void DisconnectAndDeallocateNode(TNode*&);

	// Searches for node that follows specified comparison in
	// specified subtree (fastest result is for indexAny)
	// (returns real pointer of the tree structure that points to the node or
	// should point to new node, in last case GetNodePtr == null)
	//------------------------------------------------------------------------
	TNode*& GetNodePtr(TNode*&, const TComparatorOf<T>&, TIndex) const;
	//------------------------------------------------------------------------
	static TNode*& GetAnyNodePtr(TNode*&, const TComparatorOf<T>&);
	//------------------------------------------------------------------------
	static TNode*& GetNextNodePtr(TNode*&, const TComparatorOf<T>&);
	//------------------------------------------------------------------------
	static TNode*& GetPreviousNodePtr(TNode*&, const TComparatorOf<T>&);

	// Creates new node that contains copy of specified data
	//------------------------------------------------------------------------
	virtual TNode* AllocateNode(const T&) const;
	// Destructs specified node
	//------------------------------------------------------------------------
	virtual void DeallocateNode(TNode*) const;

	//------------------------------------------------------------------------
	virtual void RebuildTreeStructure();

	//------------------------------------------------------------------------
	void PreorderNode(TNodeOperation&);
	//------------------------------------------------------------------------
	void InorderNode(TNodeOperation&);
	//------------------------------------------------------------------------
	void PostorderNode(TNodeOperation&);

	//------------------------------------------------------------------------
	void PreorderNodeConst(TNodeConstantOperation&) const;
	//------------------------------------------------------------------------
	void InorderNodeConst(TNodeConstantOperation&) const;
	//------------------------------------------------------------------------
	void PostorderNodeConst(TNodeConstantOperation&) const;

public:
	// TConstFriend
	// This class is used to allow effective printing of the tree
	//------------------------------------------------------------------------
	typedef class TConstFriend {
	protected:
		//--------------------------------------------------------------------
		static const TNode* GetRootPtrConst(const TBinaryTree& c_Tree)
		{
			return c_Tree.pRoot;
		} /*GetTreeHeadPtrConst(const TBinaryTree& c_Tree)*/
	} typedef_TConstFriend;
	friend TConstFriend;
} /*TBinaryTreeOf<T>*/;

// TBinaryTreeCheckExistenceOf template implementation
//============================================================================
template <class T> void TBinaryTreeCheckExistenceOf<T>::Check(

	const TBinaryTreeNode* pc_Node

){
	while (pc_Node) {
		switch (
			(*pc_Comparator)(((const TDataNode*)pc_Node)->Data)
		){
			case TComparator::compLess :
				pc_Node = pc_Node->pLeft;
				break;

			case TComparator::compGreater :
				pc_Node = pc_Node->pRight;
				break;

			case TComparator::compEqual :
				++nExist;

			default:
				Check(pc_Node->pLeft);
				pc_Node = pc_Node->pRight; // tail recursion
		} /*switch*/
	} /*while*/
} /*TBinaryTreeCheckExistenceOf<T>::Check(const TNode*)*/;

// TBinaryTreeExecutorOf template implementation
//============================================================================
template <class T> void TBinaryTreeExecutorOf<T>::Preorder(

	TBinaryTreeNode* pNode

){
	while (pNode) {
		(*pOperation)(*(TBinaryTreeNodeOf<T>*)pNode);
		Preorder(pNode->pLeft);
		pNode = pNode->pRight; // Tail recursion
	} /*while*/
} /*TBinaryTreeExecutorOf<T>::Preorder(TBinaryTreeNode*)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeExecutorOf<T>::Inorder(

	TBinaryTreeNode* pNode

){
	while (pNode) {
		Inorder(pNode->pLeft);
		(*pOperation)(*(TBinaryTreeNodeOf<T>*)pNode);
		pNode = pNode->pRight; // Tail recursion
	} /*while*/
} /*TBinaryTreeExecutorOf<T>::Inorder(TBinaryTreeNode*)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeExecutorOf<T>::Postorder(

	TBinaryTreeNode* pNode

){
	if (pNode) {
		Postorder(pNode->pLeft);
		Postorder(pNode->pRight);
		(*pOperation)(*(TBinaryTreeNodeOf<T>*)pNode);
	} /*if*/
} /*TBinaryTreeExecutorOf<T>::Postorder(TBinaryTreeNode*)*/;

// TBinaryTreeOf template implementation
//============================================================================
template <class T> TBinaryTreeOf<T>::TBinaryTreeOf(

	const TEtalonComparatorOf<T>& c_SortingComparator,
	const TAllocator& c_Allocator /*= NLBINTRE_ALLOCATOR_DEFAULT()*/

):
	pRoot(null), nSize(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Allocator);
	pAllocator->SetAllocationSize(sizeof(TDataNode));

	SetComparator(c_SortingComparator);
} /*TBinaryTreeOf<T>(const TEtalonComparatorOf<T>&, const TAllocator&)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeOf<T>::TBinaryTreeOf(

	const TBinaryTreeOf<T>& c_SourceTree

):
	pRoot(null), nSize(0), pAllocator(null)
{
	pAllocator = CloneObject(*c_SourceTree.pAllocator);
	pAllocator->SetAllocationSize(sizeof(TNode));

	SetComparator(c_SourceTree.Comparator());
	try {Insert(c_SourceTree);}
	catch (...) {Clear(); ClearComparator(); throw;}
} /*TBinaryTreeOf<T>(const TBinaryTreeOf<T>&)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeOf<T>::TBinaryTreeOf(

	const TContainerOf<T>& c_SourceContainer,
	const TEtalonComparatorOf<T>& c_SortingComparator,
	const TAllocator& c_Allocator /*= NLBINTRE_ALLOCATOR_DEFAULT()*/

):
	pRoot(null), nSize(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Allocator);
	pAllocator->SetAllocationSize(sizeof(TNode));

	SetComparator(c_SortingComparator);
	try {Insert(c_SourceContainer);}
	catch (...) {Clear(); ClearComparator(); throw;}
} /*TBinaryTreeOf<T>(const TContainerOf<T>&, ...)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeOf<T>::~TBinaryTreeOf()
{
	Clear();
	delete pAllocator; pAllocator = null;
} /*~TBinaryTreeOf<T>()*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::AssertValid() const
{
	TParent::AssertValid();
	if (!pAllocator)
		throw TXAllocatorIsNotDefined(THIS_CLASS);
} /*TBinaryTreeOf<T>::AssertValid() const*/

//----------------------------------------------------------------------------
template <class T> NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TBinaryTreeOf, T)

//----------------------------------------------------------------------------
template <class T> const T& TBinaryTreeOf<T>::Insert(

	const T& c_Data

){
	TNode* pNode = AllocateNode(c_Data);

	bool bInserted;
	try {
		InsertNode(*pNode, bInserted);
	} /*try*/
	catch (...) {
		if (bInserted) {DeallocateNode(pNode); pNode = null;}
		throw;
	} /*catch(...)*/

	return ((TDataNode*)pNode)->Data;
} /*TBinaryTree::Insert(const T&)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Insert(

	const TContainerOf<T>& c_Source

){
	// Is parameter binary tree?
	//---------------------------
	const TBinaryTree* pSourceTree = (
		dynamic_cast<const TBinaryTree*>(&c_Source)
	) /*pSourceTree*/;

	if (pSourceTree) {
		if (pSourceTree != dynamic_cast<const TBinaryTree*>(this)) {
			// If source is another binary tree then use Preorder iteration
			// for optimal resulting structure
			//--------------------------------------------------------------
			TContainerInsertInto<T> InsertIntoThis(*this);
			pSourceTree->Preorder(InsertIntoThis);
		} /*if*/
	}
	else {
		TSortingContainerOf<T>::Insert(c_Source);
	} /*iff*/
} /*TBinaryTreeOf<T>::Insert(const TContainer&)*/;

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::FastSearch(

	const TComparatorOf<T>& c_Comparator,
	nl_ulong_t nIndex

) const {
	TNode* pTempRoot = pRoot;
	TNode*& r_pNode = (
		(nIndex == indexAny) ?
			GetAnyNodePtr(pTempRoot, c_Comparator) :
			GetNodePtr(pTempRoot, c_Comparator, nIndex)
	) /*r_pNode*/;
	return r_pNode ? &(((TDataNode*&)r_pNode)->Data) : null;
} /*TBinaryTreeOf<T>::FastSearch(const TComparatorOf<T>&, TIndex)*/;

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::SearchFirst() const
{
	if (IsEmpty()) return null;
	const TNode* pc_Found = TNode::GetMinNodePtr(const_cast<TNode*&>(pRoot));
	return pc_Found ? &(((const TDataNode*)pc_Found)->Data) : null;
} /*TBinaryTreeOf<T>::SearchFirst() const*/

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::SearchLast() const
{
	if (IsEmpty()) return null;
	const TNode* pc_Found = TNode::GetMaxNodePtr(const_cast<TNode*&>(pRoot));
	return pc_Found ? &(((const TDataNode*)pc_Found)->Data) : null;
} /*TBinaryTreeOf<T>::SearchLast() const*/

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::SearchNext(

	const T& c_Data

) const {
	TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
		Comparator()
	) /*hComparator*/;
	hComparator->SetEtalon(c_Data);
	return SearchNext(*hComparator);
} /*TBinaryTreeOf<T>::SearchNext(const T&) const*/

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::SearchPrevious(

	const T& c_Data

) const {
	TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
		Comparator()
	) /*hComparator*/;
	hComparator->SetEtalon(c_Data);
	return SearchPrevious(*hComparator);
} /*TBinaryTreeOf<T>::SearchPrevious(const T&) const*/

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::SearchNext(

	const TComparatorOf<T>& c_Comparator

) const {
	const TNode* pc_Found = (
		GetNextNodePtr(const_cast<TNode*&>(pRoot), c_Comparator)
	) /*pc_Found*/;
	return pc_Found ? &(((const TDataNode*)pc_Found)->Data) : null;
} /*TBinaryTreeOf<T>::SearchNext(const TComparatorOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> const T* TBinaryTreeOf<T>::SearchPrevious(

	const TComparatorOf<T>& c_Comparator

) const {
	const TNode* pc_Found = (
		GetPreviousNodePtr(const_cast<TNode*&>(pRoot), c_Comparator)
	) /*pc_Found*/;
	return pc_Found ? &(((const TDataNode*)pc_Found)->Data) : null;
} /*TBinaryTreeOf<T>::SearchPrevious(const TComparatorOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TBinaryTreeOf<T>::FastExist(

	const TComparatorOf<T>& c_Comparator

) const {
	TCheckExistence CheckExistence(c_Comparator);
	return CheckExistence(pRoot);
} /*TBinaryTreeOf<T>::FastExist(const TComparatorOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TBinaryTreeOf<T>::FastDelete(

	const TComparatorOf<T>& c_Comparator,
	nl_ulong_t nIndex

){
	return DeleteNode(c_Comparator, nIndex);
} /*TBinaryTreeOf<T>::FastDelete(const TComparatorOf<T>&, TIndex)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Clear()
{
	while (!IsEmpty())
		DisconnectAndDeallocateNode(pRoot);
	nSize = 0;
} /*TBinaryTreeOf<T>::Clear()*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Preorder(

	TOperationOn<T>& r_Operation

){
	TBinaryTreeApplyOperationOn<T> ApplyOperation(r_Operation);
	PreorderNode(ApplyOperation);
} /*TBinaryTreeOf<T>::Preorder(TConstantOperation&) const*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Inorder(

	TOperationOn<T>& r_Operation

){
	TBinaryTreeApplyOperationOn<T> ApplyOperation(r_Operation);
	InorderNode(ApplyOperation);
} /*TBinaryTreeOf<T>::Inorder(TConstantOperation&) const*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Postorder(

	TOperationOn<T>& r_Operation

){
	TBinaryTreeApplyOperationOn<T> ApplyOperation(r_Operation);
	PostorderNode(ApplyOperation);
} /*TBinaryTreeOf<T>::Postorder(TConstantOperation&) const*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Preorder(

	TConstantOperationOn<T>& r_Operation

) const {
	TBinaryTreeApplyConstantOperationOn<T> ApplyOperation(
		r_Operation
	) /*ApplyOperation*/;
	PreorderNodeConst(ApplyOperation);
} /*TBinaryTreeOf<T>::Preorder(TConstantOperation&) const*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Inorder(

	TConstantOperationOn<T>& r_Operation

) const {
	TBinaryTreeApplyConstantOperationOn<T> ApplyOperation(
		r_Operation
	) /*ApplyOperation*/;
	InorderNodeConst(ApplyOperation);
} /*TBinaryTreeOf<T>::Inorder(TConstantOperation&) const*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::Postorder(

	TConstantOperationOn<T>& r_Operation

) const {
	TBinaryTreeApplyConstantOperationOn<T> ApplyOperation(
		r_Operation
	) /*ApplyOperation*/;
	PostorderNodeConst(ApplyOperation);
} /*TBinaryTreeOf<T>::Postorder(TConstantOperation&) const*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::SetComparator(

	const TEtalonComparatorOf<T>& c_Comparator

){
	TParent::SetComparator(c_Comparator);
	RebuildTreeStructure();
} /*TBinaryTreeOf<T>::SetComparator(const TSortingComparator&)*/

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::InsertNode(

	TBinaryTreeNode& r_Node,
	bool& r_bInserted

){
	// Init insertion flag
	//---------------------
	r_bInserted = false;

	// Address of pointer in the structure of the tree that
	// should point to new node
	//------------------------------------------------------
	TNode** p_pInsertPlace;

	// Search for place to insert new node
	//-------------------------------------
	{
		// Initialze search comparison
		//------------------------------
		TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
			Comparator()
		) /*hComparator*/;
		hComparator->SetEtalon(((TDataNode&)r_Node).Data);
		p_pInsertPlace = &GetAnyNodePtr(pRoot, *hComparator);
	} /*Search for place to insert new node*/

	// Check if similar data is already in the tree
	//----------------------------------------------
	if (*p_pInsertPlace) {
		// Save pointer to node that should be deleted
		//---------------------------------------------
		TNode* pNodeToDelete = *p_pInsertPlace;

		// Insert node in place of the old one
		//-------------------------------------
		r_bInserted = true;
		r_Node.pLeft = (*p_pInsertPlace)->pLeft;
		r_Node.pRight = (*p_pInsertPlace)->pRight;
		*p_pInsertPlace = &r_Node;

		// Delete old node
		//-----------------
		pNodeToDelete->pLeft = pNodeToDelete->pRight = null;
		DeallocateNode(pNodeToDelete);
	}
	else {
		// Insert node at apropriate place
		//---------------------------------
		r_bInserted = true;
		r_Node.pLeft = r_Node.pRight = null;
		*p_pInsertPlace = &r_Node;

		// New node have been inserted
		//-----------------------------
		++nSize;
	} /*if*/
} /*TBinaryTreeOf<T>::InsertNode(TNode&)*/;

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TBinaryTreeOf<T>::DeleteNode(

	const TComparatorOf<T>& c_Comparator,
	nl_ulong_t nIndex

){
	TSize nDeleted = 0;
	if (nIndex != indexAll) {
		TNode** p_pNodeToDelete = &GetNodePtr(pRoot, c_Comparator, nIndex);
		if ((nDeleted = (*p_pNodeToDelete != null)) != 0)
			DisconnectAndDeallocateNode(*p_pNodeToDelete);
	} /*if*/
	else {
		const TEtalonComparator* pc_EtalonComparator = (
			dynamic_cast<const TEtalonComparator*>(&c_Comparator)
		) /*pc_EtalonComparator*/;
		TNode** p_pNodeToDelete = &GetAnyNodePtr(pRoot, c_Comparator);
		while (*p_pNodeToDelete) {
			if (
				pc_EtalonComparator &&
				(
					pc_EtalonComparator->EtalonPtr() ==
					&(((TDataNode*)*p_pNodeToDelete)->Data)
				)
			){
				throw TXAttemptToDeleteReferencedData(THIS_CLASS);
			} /*if*/
			DisconnectAndDeallocateNode(*p_pNodeToDelete);
			++nDeleted;
			p_pNodeToDelete = &GetAnyNodePtr(pRoot, c_Comparator);
		} /*while*/
	} /*if*/
	return nDeleted;
} /*TBinaryTreeOf<T>::DeleteNode(const TComparatorOf<T>&, ...)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::DisconnectAndDeallocateNode(

	TBinaryTreeNode*& r_pNode

){
	TNode* pNodeToDeallocate = TNode::DisconnectNode(r_pNode);
	if (pNodeToDeallocate) {
		DeallocateNode(pNodeToDeallocate);
		if (!nSize) 
			throw TXInvalidSize(THIS_CLASS);
		--nSize;
	} /*if*/
} /*TBinaryTreeOf<T>::DisconnectAndDeallocateNode(TNode&*)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeNode*& TBinaryTreeOf<T>::GetNodePtr(

	TBinaryTreeNode*& r_pNode,
	const TComparatorOf<T>& c_Comparator,
	nl_ulong_t nIndex

) const {
	TNode** p_pResult = &GetAnyNodePtr(r_pNode, c_Comparator);
	if (*p_pResult && nIndex) {
		// Try to find first by value
		//----------------------------
		TNode** p_pNode = &((*p_pResult)->pLeft);
		while (*p_pNode) {
			switch (c_Comparator(((TDataNode*)*p_pNode)->Data)) {
				case TSortingComparator::compLess:
					p_pNode = &((*p_pNode)->pLeft);
					break;

				case TSortingComparator::compGreater:
					p_pNode = &((*p_pNode)->pRight);
					break;

				case TSortingComparator::compEqual:
					p_pResult = p_pNode;
					p_pNode = &((*p_pResult)->pLeft);
					break;

				default:
					throw TXUndefinedComparisonResult(THIS_CLASS);
			} /*switch*/
		} /*while*/

		if (*p_pResult && --nIndex) {
			// Prepare comparison
			//--------------------
			TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
				Comparator()
			) /*hComparator*/;
			hComparator->SetEtalon(((TDataNode*)*p_pResult)->Data);
			do {
				p_pResult = &GetNextNodePtr(
					const_cast<TNode*&>(pRoot), *hComparator
				) /*GetNextNodePtr*/;
			} while (--nIndex && *p_pResult);
		} /*while*/
	} /*if*/
	return *p_pResult;
} /*TBinaryTreeOf<T>::GetNodePtr(TNode*&, TComparatorOf&, TIndex)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeNode*& TBinaryTreeOf<T>::GetAnyNodePtr(

	TBinaryTreeNode*& r_pNode,
    const TComparatorOf<T>& c_Comparator

){
	TNode** p_pNode = &r_pNode;
    while (*p_pNode) {
    	switch (c_Comparator(((TDataNode*)*p_pNode)->Data)) {
            case TSortingComparator::compLess:
                p_pNode = &((*p_pNode)->pLeft);
                break;

            case TSortingComparator::compGreater:
                p_pNode = &((*p_pNode)->pRight);
                break;

            case TSortingComparator::compEqual:
				return *p_pNode;

            default:
                throw TXUndefinedComparisonResult(THIS_PLACE);
        } /*switch*/
    } /*while*/
	return *p_pNode;
} /*TBinaryTreeOf<T>::GetAnyNodePtr(TNode*&,const TComparatorOf&)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeNode*& TBinaryTreeOf<T>::GetNextNodePtr(

	TBinaryTreeNode*& r_pSubtree,
    const TComparatorOf<T>& c_Comparator

){
	TNode** p_pCurrent = &r_pSubtree;
	TNode** p_pNext = null;

	while (*p_pCurrent) {
    	switch (c_Comparator(((TDataNode*)*p_pCurrent)->Data)) {
        	case TComparator::compLess :
				p_pNext = p_pCurrent;
				p_pCurrent = &((*p_pCurrent)->pLeft);
                break;

			case TComparator::compEqual:
            case TComparator::compGreater:
            	p_pCurrent = &((*p_pCurrent)->pRight);
                break;

            default:
            	throw TXUndefinedComparisonResult(THIS_PLACE);
        } /*switch (c_Comparator((*p_pCurrent)->Data))*/
	} /*while*/;

	return p_pNext ? *p_pNext : *p_pCurrent;
} /*TBinaryTreeOf<T>::GetNextNodePtr(TNode*&, const TComparatorOf&)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeNode*& TBinaryTreeOf<T>::GetPreviousNodePtr(

	TBinaryTreeNode*& r_pSubtree,
    const TComparatorOf<T>& c_Comparator

){
	TBinaryTreeNode** p_pCurrent = &r_pSubtree;
	TBinaryTreeNode** p_pPrevious = null;

	while (*p_pCurrent) {
		switch (c_Comparator(((TDataNode*)*p_pCurrent)->Data)) {
            case TComparator::compLess :
            case TComparator::compEqual :
            	p_pCurrent = &((*p_pCurrent)->pLeft);
                break;

        	case TComparator::compGreater :
				p_pPrevious = p_pCurrent;
				p_pCurrent = &((*p_pCurrent)->pRight);
                break;

            default:
            	throw TXUndefinedComparisonResult(THIS_PLACE);
        } /*switch (c_Comparator((*p_pCurrent)->Data))*/
	} /*while*/;
	return p_pPrevious ? *p_pPrevious : *p_pCurrent;
} /*TBinaryTreeOf<T>::GetNextNodePtr(TNode*&, const TComparatorOf&)*/;

//----------------------------------------------------------------------------
template <class T> TBinaryTreeNode* TBinaryTreeOf<T>::AllocateNode(

	const T& c_Data

) const {
	return new (pAllocator->Allocate(sizeof(TDataNode))) TDataNode(c_Data);
} /*TBinaryTreeOf<T>::AllocateNode(const T&)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::DeallocateNode(

	TBinaryTreeNode* pNode

) const {
	if (pNode) {
		typedef TBinaryTreeNodeOf<T> _TBinaryTreeNodeOf;
		((TDataNode*)pNode)->~_TBinaryTreeNodeOf();
		pAllocator->Deallocate(pNode);
	} /*if*/
} /*TBinaryTreeOf<T>::DeallocateNode(TNode*)*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::RebuildTreeStructure()
{
	// Save root of the structure
    //----------------------------
    TBinaryTreeRebuildHelper RebuildHelper;

    // Build insert list
    //-------------------
	TNode* pInsertList = RebuildHelper.BuildInsertList(pRoot);

    // "Clear" the tree
    //------------------
    pRoot = null;
    nSize = 0;

	bool bInserted;
	TNode* pNext = null;
	try {
		while (pInsertList) {
        	pNext = pInsertList->pRight;
            pInsertList->pLeft = pInsertList->pRight = null;
            bInserted = false;
        	InsertNode(*pInsertList, bInserted);
            pInsertList = pNext;
        } /*while*/
    } /*try*/
    catch (...) {
    	// Now we doing maximum to save elements of the tree...
        // than tree become invalid but can be recovered... or deleted
        //--------------------------------------------------------------
        if (bInserted) pInsertList = pNext;
        else pInsertList->pRight = pNext;
		if (pInsertList) {pInsertList->pLeft = pRoot; pRoot = pInsertList;}
        throw;
    } /*catch (...)*/
} /*TBinaryTreeOf<T>::RebuildTreeStructure()*/;

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::PreorderNode(

	TOperationOn< TBinaryTreeNodeOf<T> >& r_Operation

){
	try {
    	TBinaryTreeExecutorOf<T> Executor(r_Operation);
        Executor.Preorder(pRoot);
    } /*try*/
    catch (...) {RebuildTreeStructure(); throw;}
    RebuildTreeStructure();
} /*TBinaryTreeOf<T>::PreorderNode(TNodeOperation&)*/

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::InorderNode(

	TOperationOn< TBinaryTreeNodeOf<T> >& r_Operation

){
	try {
       	TBinaryTreeExecutorOf<T> Executor(r_Operation);
        Executor.Inorder(pRoot);
    } /*try*/
    catch (...) {RebuildTreeStructure(); throw;}
    RebuildTreeStructure();
} /*TBinaryTreeOf<T>::InorderNode(TNodeOperation&)*/

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::PostorderNode(

	TOperationOn< TBinaryTreeNodeOf<T> >& r_Operation

){
	try {
       	TBinaryTreeExecutorOf<T> Executor(r_Operation);
        Executor.Postorder(pRoot);
    } /*try*/
    catch (...) {RebuildTreeStructure(); throw;}
    RebuildTreeStructure();
} /*TBinaryTreeOf<T>::PostorderNode(TNodeOperation&)*/

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::PreorderNodeConst(

	TConstantOperationOn< TBinaryTreeNodeOf<T> >& r_Operation

) const {
	try {
    	TBinaryTreeExecutorOf<T> Executor(r_Operation);
        Executor.Preorder(pRoot);
    } /*try*/
    catch (const TOperation::TXAbort&) {}
} /*TBinaryTreeOf<T>::PreorderNodeConst(TNodeOperation&) const*/

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::InorderNodeConst(

	TConstantOperationOn< TBinaryTreeNodeOf<T> >& r_Operation

) const {
	try {
    	TBinaryTreeExecutorOf<T> Executor(r_Operation);
        Executor.Inorder(pRoot);
    } /*try*/
    catch (const TOperation::TXAbort&) {}
} /*TBinaryTreeOf<T>::InorderNodeConst(TNodeOperation&) const*/

//----------------------------------------------------------------------------
template <class T> void TBinaryTreeOf<T>::PostorderNodeConst(

	TConstantOperationOn< TBinaryTreeNodeOf<T> >& r_Operation

) const {
	try {
    	TBinaryTreeExecutorOf<T> Executor(r_Operation);
        Executor.Postorder(pRoot);
    } /*try*/
    catch (const TOperation::TXAbort&) {}
} /*TBinaryTreeOf<T>::PostorderNodeConst(TNodeOperation&) const*/

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLBINTRE_DECLSPEC(T, Declspec) \
	NLSCONT_DECLSPEC(T, Declspec); \
    NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TBinaryTreeNodeOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
    NLOPERAT_DECLSPEC( \
		NATLIB_NAMESPACE_MEMBER1(TBinaryTreeNodeOf, T), Declspec \
	) /*NLOPERAT_DECLSPEC*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TBinaryTreeExecutorOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
    NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TBinaryTreeApplyOperationOn), \
		T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
    	NATLIB_NAMESPACE_MEMBER(TBinaryTreeApplyConstantOperationOn), \
		T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TBinaryTreeCheckExistenceOf), \
		T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
    NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TBinaryTreeOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLBINTRE_H__)*/
