#ifndef __NATLIB_NLCLONT_H__
#define __NATLIB_NLCLONT_H__
/****************************************************************************/
/** nlcont.h                                                               **/
/** ---------------------------------------------------------------------- **/
/** Natlib abstract container base                                         **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** October 08, 1996 by Oleg Kagan                                         **/
/****************************************************************************/

//============================================================================
#include <nlcondit.h>
#include <nlcompar.h>
#include <nloperat.h>
#include <nlautopt.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TContainerOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TContainerOf : 
	public virtual TNatlibObject 
{
public:
	//------------------------------------------------------------------------
    typedef TNatlibObject TParent;

	//------------------------------------------------------------------------
    typedef TContainerOf<T> TContainer;
    typedef TEtalonComparatorOf<T> TEtalonComparator;

    //------------------------------------------------------------------------
    typedef nl_ulong_t TSize;
    typedef TSize TIndex;

	// Some constants that can be used access functions
    //------------------------------------------------------------------------
    enum {indexAll = 0, indexAny = indexAll, indexFirst = 1};

    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
		TException, TNatlibObject::TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE_ABORT(TXAbort, TException, _NATLIB_TEMPLATE_DECLSPEC_);
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXOutOfRange, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    typedef TXOutOfRange TXNotFound;

  	//------------------------------------------------------------------------
    typedef TOperationOn<T> TOperation;
    typedef TConstantOperationOn<T> TConstantOperation;

	//------------------------------------------------------------------------
    TStdString Dump() const;

#ifndef _NATLIB_NO_EXTENDED_OVERLOADING_
    //------------------------------------------------------------------------
    NLOBJECT_DECLARE_CLONE1(TContainerOf, T) = 0;
#endif /*_NATLIB_NO_EXTENDED_OVERLOADING_*/

	// Returns number of elements in container
	//------------------------------------------------------------------------
    virtual TSize Size() const = 0;
    //------------------------------------------------------------------------
    virtual bool IsEmpty() const {return !Size();};

	// Searches for elements that follow the specified condition,
    // returns element number nIndex (returns null if not found)
	//------------------------------------------------------------------------
    virtual const T* Search(const TConditionOf<T>&, TIndex = indexAny) const;
	//------------------------------------------------------------------------
    virtual const T* Search(const TComparatorOf<T>&, TIndex = indexAny) const;
	// Throws TXNotFound if can not find
    //------------------------------------------------------------------------
    const T& Find(const TConditionOf<T>&, TIndex = indexAny) const;
    //------------------------------------------------------------------------
    const T& Find(const TComparatorOf<T>&, TIndex = indexAny) const;

	// Returns number of elements in container that follows the condition
	//------------------------------------------------------------------------
    virtual TSize Exist(const TConditionOf<T>&) const;
    //------------------------------------------------------------------------
    virtual TSize Exist(const TComparatorOf<T>&) const;

	// Inserts element into container
    //------------------------------------------------------------------------
	virtual const T& Insert(const T&) = 0;
	// Inserts all elements from specified container
	//------------------------------------------------------------------------
	virtual void Insert(const TContainer&);

    // Clears this container than inserts all elements from specified
    //------------------------------------------------------------------------
    virtual void Let(const TContainer&);

	// Deletes elements that follow the specified condition
    // (returns number of deleted elements)
    //------------------------------------------------------------------------
    virtual TSize Delete(const TConditionOf<T>&, TIndex = indexAll) = 0;
    //------------------------------------------------------------------------
    virtual TSize Delete(const TComparatorOf<T>&, TIndex = indexAll);

	// Clears contents of the container (deletes everything)
    //------------------------------------------------------------------------
    virtual void Clear() = 0;

	// Applies the specified operation for all elements in th econtainer
    //------------------------------------------------------------------------
	virtual void ForEach(TOperation&) = 0;
	// Applies the specified operation for all elements that follow the
    // specified condition
	//------------------------------------------------------------------------
    virtual void ForEachThat(const TConditionOf<T>&, TOperation&);

    //------------------------------------------------------------------------
	virtual void ForEach(TConstantOperation&) const = 0;
	//------------------------------------------------------------------------
	void ForEachConst(TConstantOperation& r_Operation) const 
	{
		ForEach(r_Operation);
	} /*ForEachConst(TConstantOperation&) const*/

	//------------------------------------------------------------------------
    virtual void ForEachThat(
		const TConditionOf<T>&, TConstantOperation&
	) const /*ForEachThat(const TConditionOf<T>&, TConstantOperation&)*/;

	//------------------------------------------------------------------------
    void ForEachThatConst(
		const TConditionOf<T>& c_Condition, 
		TConstantOperation& r_Operation
	) const {
		ForEachThat(c_Condition, r_Operation);
	} /*ForEachThatConst(const TConditionOf<T>&, ...) const*/

    //------------------------------------------------------------------------
    virtual bool IsEqualByComparator(
		const TContainer&, const TEtalonComparator&
	) const /*IsEqualByComparator(const TContainer&, const ...) const*/;

	//------------------------------------------------------------------------
    inline TContainer& operator=(const TContainer& c_Source)
    {
    	Let(c_Source); return *this;
    } /*operator=(const TContainer&)*/

    //------------------------------------------------------------------------
    inline TContainer& operator+=(const T& c_Data)
    {
    	Insert(c_Data); return *this;
    } /*operator+=(const T&)*/
    //------------------------------------------------------------------------
    inline TContainer& operator+=(const TContainer& c_Source)
    {
    	Insert(c_Source); return *this;
    } /*operator+=(const TContainer&)*/

    //------------------------------------------------------------------------
    inline TContainer& operator-=(const TConditionOf<T>& c_Condition)
    {
    	Delete(c_Condition); return *this;
    } /*operator-=(const TConditionOf<T>&)*/
    //------------------------------------------------------------------------
    inline TContainer& operator-=(const TComparatorOf<T>& c_Comparator)
    {
    	Delete(c_Comparator); return *this;
    } /*operator-=(const TComparatorOf<T>&)*/
} /*TContainerOf<T>*/;

// TContainerInsertInto template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TContainerInsertInto :
	public TConstantOperationOn<T>
{
public:
	//------------------------------------------------------------------------
    typedef TConstantOperationOn<T> TParent;
    //------------------------------------------------------------------------
    typedef TContainerOf<T> TContainer;

    //------------------------------------------------------------------------
    typedef basic_TXLowMemory TXLowMemory;

    //------------------------------------------------------------------------
    TContainerInsertInto(TContainer& r_Target) : pTarget(&r_Target) {}
    //------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TContainerInsertInto, T);

    //------------------------------------------------------------------------
    void operator()(const T&);

protected:
    //------------------------------------------------------------------------
	TContainer* pTarget;

private:
    //------------------------------------------------------------------------
    inline void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TContainerInsertInto<T>*/;

// TContainerSearchByIndex template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TContainerSearchByIndex :
	public TConstantOperationOn<T>
{
public:
   	//------------------------------------------------------------------------
	typedef TConstantOperationOn<T> TParent;
    //------------------------------------------------------------------------
	typedef TContainerOf<T> TContainer;
    //------------------------------------------------------------------------
    typedef TContainer::TIndex TIndex;
    //------------------------------------------------------------------------
    typedef const T* TSearchDataPtr;

    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE_DATA_HOLDER(
    	TXFound, TSearchDataPtr, TConstantOperationOn<T>::TException,
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*DECLARE_XINFO_DATA*/;

    //------------------------------------------------------------------------
    TContainerSearchByIndex(
		const TConditionOf<T>& c_Condition, TIndex nIndex
	):
		pCondition(&c_Condition),
        nCounter(nIndex ? nIndex : TContainer::indexFirst)
    {} /*TContainerSearchByIndex(const TConditionOf<T>&, TIndex)*/

    //------------------------------------------------------------------------
    void operator()(const T&);

protected:
   	//------------------------------------------------------------------------
    const TConditionOf<T>* pCondition;
    TIndex nCounter;

private:
    //------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(TContainerSearchByIndex, T)

    //------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TContainerSearchByIndex<T>*/;

// TContainerCountByCondition template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TContainerCountByCondition :
	public TConstantOperationOn<T>
{
public:
    //------------------------------------------------------------------------
    typedef TConstantOperationOn<T> TParent;
    //------------------------------------------------------------------------
    typedef TContainerOf<T>::TSize TSize;

    //------------------------------------------------------------------------
    TContainerCountByCondition(const TConditionOf<T>& c_Condition) :
        pc_Condition(&c_Condition), nFound(0)
    {} /*TContainerCountByCondition(const TConditionOf<T>&)*/

    //------------------------------------------------------------------------
    TSize Found() const {return nFound;}

    //------------------------------------------------------------------------
    void operator()(const T&);

protected:
    //------------------------------------------------------------------------
    const TConditionOf<T>* pc_Condition;
    TSize nFound;

private:
    //------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(TContainerCountByCondition, T)

    //------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TContainerCountByCondition<T>*/;

// TContainerApplyByCondition template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TContainerApplyByCondition :
	public TOperationOn<T>
{
public:
	//------------------------------------------------------------------------
	typedef TOperationOn<T> TParent;
    //------------------------------------------------------------------------
    typedef TContainerOf<T>::TOperation TOperation;

    //------------------------------------------------------------------------
    TContainerApplyByCondition(
        const TConditionOf<T>& c_Condition, TOperation& r_Operation
    ):
        pc_Condition(&c_Condition), pOperation(&r_Operation)
    {} /*TContainerApplyByCondition(const TConditionOf<T>&, TOperation&)*/

    //------------------------------------------------------------------------
    void operator()(T& r_Data);

protected:
    //------------------------------------------------------------------------
    const TConditionOf<T>* pc_Condition;
    TOperation* pOperation;

private:
    //------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(TContainerApplyByCondition, T)
} /*TContainerApplyByCondition<T>*/;

// TContainerConstantApplyByCondition template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TContainerConstantApplyByCondition :
    public TConstantOperationOn<T>
{
public:
	//------------------------------------------------------------------------
    typedef TConstantOperationOn<T> TParent;
    //------------------------------------------------------------------------
    typedef TContainerOf<T>::TConstantOperation TConstantOperation;

    //------------------------------------------------------------------------
    TContainerConstantApplyByCondition(
        const TConditionOf<T>& c_Condition,
		TConstantOperation& r_Operation
    ):
        pc_Condition(&c_Condition), pOperation(&r_Operation)
    {} /*TContainerConstantApplyByCondition(const TConditionOf<T>&, ...)*/

    //------------------------------------------------------------------------
    void operator()(const T&);

protected:
    //------------------------------------------------------------------------
    const TConditionOf<T>* pc_Condition;
    TConstantOperation* pOperation;

private:
    //------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(
		TContainerConstantApplyByCondition, T
	) /*NLOBJECT_CLONE_IS_NOT_ALLOWED1*/

    //------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TContainerConstantApplyByCondition<T>*/;

// TContainerCheckEqualityByComparator template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TContainerCheckEqualityByComparator :
	public TConstantOperationOn<T>
{
public:
	//------------------------------------------------------------------------
    typedef TConstantOperationOn<T> TParent;
	//------------------------------------------------------------------------
    typedef TContainerOf<T> TContainer;

    //------------------------------------------------------------------------
    typedef TEtalonComparatorOf<T> TEtalonComparator;

    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
		TXNotEqual, TContainerOf<T>::TException, 
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;

	//------------------------------------------------------------------------
	TContainerCheckEqualityByComparator(
    	const TContainer& c_Container1, const TContainer& c_Container2,
        TEtalonComparator& c_EtalonComparator
    ) /*TContainerCheckEqualityByComparator(...)*/;

    //------------------------------------------------------------------------
	void operator()(const T&);

protected:
	//------------------------------------------------------------------------
    const TContainer* pc_Container1;
    const TContainer* pc_Container2;
    //------------------------------------------------------------------------
    TAutoRestoreEtalonComparatorHandleOf<T> hComparator;

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(
		TContainerCheckEqualityByComparator, T
	) /*NLOBJECT_CLONE_IS_NOT_ALLOWED1*/;

private:
    //------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TContainerCheckEqualityByComparator<T>*/;

// TContainerOf template implementation
//============================================================================
template <class T> TStdString TContainerOf<T>::Dump() const
{
	TStdString sDump(TParent::Dump());

	// Add size
    //----------
    {
	    char szDump[256];
	    sprintf(szDump, " size: %lu", (unsigned long)Size());
	    sDump += szDump;
    } /*Add size*/

    return sDump;
} /*TContainerOf<T>::Dump() const*/

//----------------------------------------------------------------------------
template <class T> const T* TContainerOf<T>::Search(

	const TConditionOf<T>& c_Condition,
    nl_ulong_t nIndex

) const {
	TContainerSearchByIndex<T> Search(c_Condition, nIndex);
    try {
		ForEach(Search);
    } /*try*/
    catch (TContainerSearchByIndex<T>::TXFound& r_XFound) {
    	return r_XFound.Value();
    } /*catch (TSearchByIndex::TXFound&)*/
    return null;
} /*TContainerOf<T>::Search(const TConditionOf<T>&, TIndex)*/

//----------------------------------------------------------------------------
template <class T> const T* TContainerOf<T>::Search(

	const TComparatorOf<T>& c_Comparator,
    nl_ulong_t nIndex

) const {
	TIsEqualByComparatorOf<T> IsEqual(c_Comparator);
	return Search(IsEqual, nIndex);
} /*TContainerOf<T>::Search(const TComparatorOf<T>&, TIndex)*/

//----------------------------------------------------------------------------
template <class T> const T& TContainerOf<T>::Find(

	const TConditionOf<T>& c_Condition,
    nl_ulong_t nIndex

) const {
	const T* pc_Data = Search(c_Condition, nIndex);
	if (!pc_Data) 
		throw TXNotFound(THIS_CLASS);
    return *pc_Data;
} /*TContainerOf<T>::Find(const TConditionOf<T>&, TIndex)*/

//----------------------------------------------------------------------------
template <class T> const T& TContainerOf<T>::Find(

	const TComparatorOf<T>& c_Comparator,
    nl_ulong_t nIndex

) const {
	const T* pc_Data = Search(c_Comparator, nIndex);
	if (!pc_Data) 
		throw TXNotFound(THIS_CLASS);
    return *pc_Data;
} /*TContainerOf<T>::Find(const TComparatorOf<T>&, TIndex)*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TContainerOf<T>::Exist(

	const TConditionOf<T>& c_Condition

) const {
	TContainerCountByCondition<T> CountByCondition(c_Condition);
    ForEach(CountByCondition);
    return CountByCondition.Found();
} /*TContainerf<T>::Exist(const TConditionOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TContainerOf<T>::Exist(

	const TComparatorOf<T>& c_Comparator

) const {
	TIsEqualByComparatorOf<T> IsEqual(c_Comparator);
    return Exist(IsEqual);
} /*TContainerf<T>::Exist(const TComparatorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TContainerOf<T>::Insert(

	const TContainerOf<T>& c_Container

){
	TContainerInsertInto<T> InsertIntoThis(*this);
    c_Container.ForEach(InsertIntoThis);
} /*TContainerOf<T>::Insert(const TContainerOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TContainerOf<T>::Let(

	const TContainerOf<T>& c_Source

){
	if (
		dynamic_cast<const TContainer*>(this) != 
		dynamic_cast<const TContainer*>(&c_Source)
	){
    	Clear();
        Insert(c_Source);
    } /*if*/
} /*TContainerOf<T>::Let(const TContainerOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TContainerOf<T>::Delete(

	const TComparatorOf<T>& c_Comparator,
    nl_ulong_t nIndex

){
	TIsEqualByComparatorOf<T> IsEqual(c_Comparator);
    return Delete(IsEqual, nIndex);
} /*TContainerOf<T>::Delete(const TComparatorOf<T>&, TIndex)*/

//----------------------------------------------------------------------------
template <class T> void TContainerOf<T>::ForEachThat(

	const TConditionOf<T>& c_Condition,
    TOperationOn<T>& r_Operation

){
	TContainerApplyByCondition<T> Apply(c_Condition, r_Operation);
	ForEach(Apply);
} /*TContainerOf<T>::ForEachThat(const TConditionOf<T>&, ...)*/

//----------------------------------------------------------------------------
template <class T> void TContainerOf<T>::ForEachThat(

	const TConditionOf<T>& c_Condition,
    TConstantOperationOn<T>& r_Operation

) const {
	TContainerConstantApplyByCondition<T>
    	Apply(c_Condition, r_Operation);
	ForEach(Apply);
} /*TContainerOf<T>::ForEachThat(const TConditionOf<T>&, ...) const*/

//----------------------------------------------------------------------------
template <class T> bool TContainerOf<T>::IsEqualByComparator(

	const TContainerOf<T>& c_Container,
    const TEtalonComparatorOf<T>& c_Comparator

) const {
	if (
    	dynamic_cast<const TContainerOf<T>*>(this) ==
        dynamic_cast<const TContainerOf<T>*>(&c_Container)
    ) return true;

	if (Size() != c_Container.Size()) return false;

    // Prepare copy of comparator
    //----------------------------
	TAutoPointerTo< TEtalonComparatorOf<T> > hComparator(
		CloneObject(c_Comparator), false
    ) /*hComparator*/;

    if (hComparator.IsNull()) {
    	throw TEtalonComparator::TXCanNotBeCloned(
			THIS_PLACE_CLASS(c_Comparator)
		) /*throw TEtalonComparator::TXCanNotBeCloned*/;
	} /*if*/

    try {
    	TContainerCheckEqualityByComparator<T> CheckEquality(
        	*this, c_Container, *hComparator
        ) /*CheckEquality*/;
    } /*try*/
    catch (TContainerCheckEqualityByComparator<T>::TXNotEqual&) {
    	return false;
    } /*catch (TContainerCheckEqualityByComparator::TXNotEqual&)*/
    return true;
} /*TContainerOf<T>::IsEqualByComparator(...) const*/

// TContainerInsertInto template implementation
//============================================================================
template <class T> 
NLOBJECT_CLONE_RESULT1(TContainerInsertInto, T)
TContainerInsertInto<T>::Clone() const
{
    TContainerInsertInto<T>* pCopy;
    NLEXCEPT_CLASS_NEW(TContainerInsertInto<T>(*pTarget), pCopy);
    return pCopy;
} /*TContainerInsertInto<T>::Clone() const*/

//----------------------------------------------------------------------------
template <class T> void TContainerInsertInto<T>::operator()(

	const T& c_Data

){
	pTarget->Insert(c_Data);
} /*TContainerInsertInto<T>::operator()(const T&)*/

// TContainerSearchByIndex template implementation
//============================================================================
template <class T> void TContainerSearchByIndex<T>::operator()(

	const T& c_Data

){
	if ((*pCondition)(c_Data) && (!--nCounter))
    	throw TXFound(&c_Data, THIS_CLASS);
} /*TContainerSearchByIndex<T>::operator()(const T&)*/

// TContainerCountByCondition template implementation
//============================================================================
template <class T> void TContainerCountByCondition<T>::operator()(

	const T& c_Data

){
	if ((*pc_Condition)(c_Data)) ++nFound;
} /*TContainerCountByCondition<T>::operator(const T&)*/

// TContainerApplyByCondition template implementation
//============================================================================
template <class T> void TContainerApplyByCondition<T>::operator()(

	T& r_Data

){
	if ((*pc_Condition)(r_Data)) (*pOperation)(r_Data);
} /*TContainerApplyByCondition<T>::operator()(T&)*/

// TContainerConstantApplyByCondition template implementation
//============================================================================
template <class T> void TContainerConstantApplyByCondition<T>::operator()(

	const T& c_Data

){
	if ((*pc_Condition)(c_Data)) (*pOperation)(c_Data);
} /*TContainerConstantApplyByCondition<T>::operator()(const T&)*/

// TContainerCheckEqualityByComparator template implementation
//============================================================================
template <class T>
TContainerCheckEqualityByComparator<T>::TContainerCheckEqualityByComparator(

	const TContainerOf<T>& c_Container1,
    const TContainerOf<T>& c_Container2,
    TEtalonComparatorOf<T>& r_EtalonComparator

):
	pc_Container1(&c_Container1), pc_Container2(&c_Container2),
    hComparator(r_EtalonComparator)
{} /*TContainerCheckEqualityByComparator<T>(const TContainer&,...)*/

//----------------------------------------------------------------------------
template <class T> void TContainerCheckEqualityByComparator<T>::operator()(

	const T& c_Data

){
	hComparator->SetEtalon(c_Data);
	if (
    	pc_Container1->Exist(*hComparator) !=
        pc_Container2->Exist(*hComparator)
    ){
		throw TXNotEqual(THIS_CLASS);
	} /*if*/
    hComparator->ClearEtalon();
} /*TContainerCheckEqualityByComparator::operator()(const T&)*/

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLCONT_DECLSPEC(T, Declspec) \
	NLCONDIT_DECLSPEC(T, Declspec); \
    NLCOMPAR_DECLSPEC(T, Declspec); \
    NLOPERAT_DECLSPEC(T, Declspec); \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TContainerOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TContainerInsertInto), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TContainerSearchByIndex), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/;\
	NATLIB_EXPAND_TEMPLATE1( \
    	NATLIB_NAMESPACE_MEMBER(TContainerCountByCondition), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TContainerApplyByCondition), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TContainerConstantApplyByCondition), \
		T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
    NATLIB_EXPAND_TEMPLATE1( \
    	NATLIB_NAMESPACE_MEMBER(TContainerCheckEqualityByComparator), \
		T, Declspec \
    ) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLCLONT_H__)*/
