#ifndef __NATLIB_NLSAFEH_H__
#define __NATLIB_NLSAFEH_H__
/****************************************************************************/
/** nlsafeh.h                                                              **/
/** ---------------------------------------------------------------------- **/
/** Safe object handling implementation                                    **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** June 16, 1996 by Oleg Kagan                                            **/
/****************************************************************************/

//============================================================================
#include <nlobject.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

//============================================================================
class _NATLIB_DECLSPEC_ TSafeHandleBase;

// TSafeHandleAllowed
// The base for all classes that allow usage of safe handles.
// You have to VIRTUALLY derive Your classes ftom this class.
//============================================================================
class _NATLIB_DECLSPEC_ TSafeHandleAllowed {
public:
	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE_BASE(TXFatalError, _NATLIB_DECLSPEC_);
    NLEXCEPT_DECLARE(
		TXInvalidReferencesNumber, TXFatalError, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE(
		TXUnableToCreateDescriptor, TXFatalError, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;

	//------------------------------------------------------------------------
    typedef class _NATLIB_DECLSPEC_ TDescriptor {
	public:
    	//--------------------------------------------------------------------
        nl_ulong_t nReferences;
        //--------------------------------------------------------------------
        union {
        	TSafeHandleAllowed* pObject;
			const TSafeHandleAllowed* pc_Object;
        } /*union*/;

        //--------------------------------------------------------------------
        TDescriptor(const TSafeHandleAllowed& c_Object) :
        	nReferences(0), pc_Object(&c_Object)
        {} /*TDescriptor(const TSafeHandleAllowed&)*/
        //--------------------------------------------------------------------
        TDescriptor(TSafeHandleAllowed& r_Object) :
        	nReferences(0), pObject(&r_Object)
        {} /*TDescriptor(TSafeHandleAllowed&)*/

        //--------------------------------------------------------------------
        ~TDescriptor()
		{
		    if (nReferences) {
				throw TSafeHandleAllowed::TXInvalidReferencesNumber(
					THIS_CLASS
				) /*TSafeHandleAllowed::TXInvalidReferencesNumber*/;
			} /*if*/

			// Unregister descriptor if needed
			//---------------------------------
			if (pObject) {pObject->pDescriptor = null; pObject = null;}
		} /*~TDescriptor*/

        //--------------------------------------------------------------------
        void RegisterReference()
        {
        	if (!++nReferences) {
            	throw TSafeHandleAllowed::TXInvalidReferencesNumber(
                	THIS_CLASS
                ) /*TSafeHandleAllowed::TXInvalidReferencesNumber*/;
            } /*if*/
			if (pObject) pObject->EvHandleCreated();
        } /*RegisterReference()*/
        //--------------------------------------------------------------------
        void UnregisterReference()
		{
			if (!nReferences) {
				throw TSafeHandleAllowed::TXInvalidReferencesNumber(
					THIS_CLASS
				) /*TSafeHandleAllowed::TXInvalidReferencesNumber*/;
			} /*if*/
			TSafeHandleAllowed* pObjectToNotify = pObject;
			--nReferences;
			if (!nReferences) delete this;
			if (pObjectToNotify) pObjectToNotify->EvHandleDeleted(); 
		} /*UnregisterReference()*/
    } typedef_TDescriptor;
    friend TDescriptor;

    //------------------------------------------------------------------------
    virtual ~TSafeHandleAllowed();

    //------------------------------------------------------------------------
    nl_ulong_t References() const
    {
    	return pDescriptor ? pDescriptor->nReferences : 0;
    } /*References() const*/

    //------------------------------------------------------------------------
    void InvalidateHandles();

protected:
	//------------------------------------------------------------------------
    TSafeHandleAllowed() : pDescriptor(null) {}
	
	//------------------------------------------------------------------------
    TDescriptor& GetDescriptor() const;

   	//------------------------------------------------------------------------
	virtual void EvHandleCreated() const {}
	//------------------------------------------------------------------------
	virtual void EvHandleDeleted() const {}

    //------------------------------------------------------------------------
    virtual TDescriptor* AllocateDescriptor() const;

private:
	//------------------------------------------------------------------------
    mutable TDescriptor* pDescriptor;

    //------------------------------------------------------------------------
    friend TSafeHandleBase;
} /*TSafeHandleAllowed*/;

// TSafeHandleBase class
//============================================================================
class _NATLIB_DECLSPEC_ TSafeHandleBase : public TNatlibObject {
public:
	//------------------------------------------------------------------------
    typedef TNatlibObject TParent;

	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
		TException, TNatlibObject::TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE(TXNullHandleDereference, TException, _NATLIB_DECLSPEC_);
	NLEXCEPT_DECLARE(
		TXInvalidHandleDereference, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;

	//------------------------------------------------------------------------
    TSafeHandleBase() : pDescriptor(null) {}

    //------------------------------------------------------------------------
    TSafeHandleBase(const TSafeHandleAllowed& c_Target) :
    	pDescriptor(null)
    {
    	Assign(c_Target);
    } /*TSafeHandleBase(const TSafeHandleAllowed&)*/

    //------------------------------------------------------------------------
    TSafeHandleBase(const TSafeHandleBase& c_hSource) :
    	pDescriptor(null)
    {
		Let(c_hSource);
    } /*TSafeHandleBase(const TSafeHandleBase&)*/

	//------------------------------------------------------------------------
    virtual ~TSafeHandleBase();

    //------------------------------------------------------------------------
    void Assign(const TSafeHandleAllowed&);
    //------------------------------------------------------------------------
    void Unassign();
	//------------------------------------------------------------------------
	void Let(const TSafeHandleBase&);

	//------------------------------------------------------------------------
	TSafeHandleAllowed& Dereference() const
	{
		if (!pDescriptor) 
			throw TXNullHandleDereference(THIS_CLASS);
	   	if (!pDescriptor->pObject) 
			throw TXInvalidHandleDereference(THIS_CLASS);
		return *(pDescriptor->pObject);
	} /*Dereference() const*/

   	//------------------------------------------------------------------------
    bool IsNull() const {return !pDescriptor;}
    //------------------------------------------------------------------------
    bool IsValid() const {return !pDescriptor || pDescriptor->pObject;}
    //------------------------------------------------------------------------
    bool IsTargetValid() const {return pDescriptor && pDescriptor->pObject;}

	//------------------------------------------------------------------------
	TSafeHandleBase& operator=(const TSafeHandleBase& c_hSource) 
	{
		Let(c_hSource); return *this;
	} /*operator=(const TSafeHandleBase&)*/
    //------------------------------------------------------------------------
    bool operator==(const TSafeHandleBase& c_hToCompare) const
    {
    	return pDescriptor == c_hToCompare.pDescriptor;
    } /*bool operator==(const TSafeHandleBase&) const*/
    //------------------------------------------------------------------------
    bool operator!=(const TSafeHandleBase& c_hToCompare) const
    {
    	return !operator==(c_hToCompare);
    } /*bool operator==(const TSafeHandleBase&) const*/

	//------------------------------------------------------------------------
	bool operator>(const TSafeHandleBase& c_hToCompare) const
	{
		return nl_ulong_t(pDescriptor) > nl_ulong_t(c_hToCompare.pDescriptor);
	} /*operator>(const TSafeHandleBase& c_hToCompare) const*/
	//------------------------------------------------------------------------
	bool operator<(const TSafeHandleBase& c_hToCompare) const
	{
		return nl_ulong_t(pDescriptor) < nl_ulong_t(c_hToCompare.pDescriptor);
		} /*operator<(const TSafeHandleBase& c_hToCompare) const*/
	//------------------------------------------------------------------------
	bool operator>=(const TSafeHandleBase& c_hToCompare) const
	{
		return (
			nl_ulong_t(pDescriptor) >= nl_ulong_t(c_hToCompare.pDescriptor)
		) /*return*/;
	} /*operator>=(const TSafeHandleBase& c_hToCompare) const*/
	//------------------------------------------------------------------------
	bool operator<=(const TSafeHandleBase& c_hToCompare) const
	{
		return (
			nl_ulong_t(pDescriptor) <= nl_ulong_t(c_hToCompare.pDescriptor)
		) /*return*/;
	} /*operator<=(const TSafeHandleBase& c_hToCompare) const*/

    //------------------------------------------------------------------------
    TSafeHandleAllowed& operator*() const {return Dereference();}
    //------------------------------------------------------------------------
    TSafeHandleAllowed* operator->() const {return &Dereference();}

protected:
    //------------------------------------------------------------------------
    typedef TSafeHandleAllowed::TDescriptor TDescriptor;

protected:
	//------------------------------------------------------------------------
    TSafeHandleAllowed::TDescriptor* pDescriptor;

    //------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED(TSafeHandleBase)
} /*TSafeHandleBase*/;

// TSafeHandleOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TSafeHandleOf : 
	public TSafeHandleBase 
{
public:
	//------------------------------------------------------------------------
    typedef TSafeHandleBase TParent;
    //------------------------------------------------------------------------
    typedef TSafeHandleOf<T> TSafeHandle;

	//------------------------------------------------------------------------
    TSafeHandleOf() : TParent() {}
    //------------------------------------------------------------------------
    TSafeHandleOf(T& r_Target) : TParent(r_Target) {}
    //------------------------------------------------------------------------
    TSafeHandleOf(const TSafeHandle& c_hSource) :
    	TParent(c_hSource)
    {} /*TSafeHandleOf(const TSafeHandle&)*/

    //------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TSafeHandleOf, T);

    //------------------------------------------------------------------------
    void Assign(T& r_Target) {TParent::Assign(r_Target);}
    //------------------------------------------------------------------------
    void Let(const TSafeHandle& c_Source) {TParent::Let(c_Source);}
    //------------------------------------------------------------------------
    void Unassign() {TParent::Unassign();}

    //------------------------------------------------------------------------
    T& Dereference() const;

	//------------------------------------------------------------------------
    TSafeHandle& operator=(const TSafeHandle& c_hSource)
    {
    	Let(c_hSource); return *this;
   	} /*operator=(const TSafeHandle&)*/

    //------------------------------------------------------------------------
    TSafeHandle& operator=(T& r_Target)
    {
    	Assign(r_Target); return *this;
    } /*operator=(T&)*/

    //------------------------------------------------------------------------
    T& operator*() const {return Dereference();}
    //------------------------------------------------------------------------
    T* operator->() const {return &Dereference();}
} /*TSafeHandleOf<T>*/;

// TSafeHandleOf template implementation
//============================================================================
template <class T> NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TSafeHandleOf, T)

//----------------------------------------------------------------------------
template <class T> T& TSafeHandleOf<T>::Dereference() const
{
	T* pTarget = dynamic_cast<T*>(&TParent::Dereference());
	if (!pTarget)
		throw TXInvalidHandleDereference(THIS_CLASS);
	return *pTarget;
} /*Dereference() const*/

//============================================================================    
NATLIB_NAMESPACE_END

//============================================================================
#define NLSAFEH_DECLSPEC(T, Declspec) \
	NLOBJECT_DECLSPEC(T, Declspec); \
    NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TSafeHandleOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLSAFEH_H__)*/
