#ifndef __NATLIB_NLSCONT_H__
#define __NATLIB_NLSCONT_H__
/****************************************************************************/
/** nlscont.h                                                              **/
/** ---------------------------------------------------------------------- **/
/** Natlib Sorting Container is container that is able to compare object   **/
/** by user specified comparison so it can use more effective algorithms   **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** November 09, 1996 by Oleg Kagan                                        **/
/****************************************************************************/

//============================================================================
#include <nlcont.h>
#include <nlautopt.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TSortingContainerOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TSortingContainerOf :
	public TContainerOf<T>
{
public:
	//------------------------------------------------------------------------
    typedef TContainerOf<T> TParent;

    //------------------------------------------------------------------------
    typedef TSortingContainerOf<T> TSortingContainer;
    typedef TEtalonComparatorOf<T> TSortingComparator;

	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
		TException, TContainerOf<T>::TException, 
		_NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(
		TXNoComparatorAssigned, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(
		TXUndefinedComparisonResult, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE(TXBadInorder, TException, _NATLIB_TEMPLATE_DECLSPEC_);
    NLEXCEPT_DECLARE(TXInvalidSize, TException, _NATLIB_TEMPLATE_DECLSPEC_);

    //------------------------------------------------------------------------
    ~TSortingContainerOf();

    //------------------------------------------------------------------------
    void AssertValid() const;

    //------------------------------------------------------------------------
#ifndef _NATLIB_NO_EXTENDED_OVERLOADING_
	NLOBJECT_DECLARE_CLONE1(TSortingContainerOf, T) = 0;
#endif /*!_NATLIB_NO_EXTENDED_OVERLOADING_*/

    // Fast search implements fast (binary for example) search specifically
    // for the container. Restrictions for comparison kind may be declared
	//------------------------------------------------------------------------
    virtual const T* FastSearch(
		const TComparatorOf<T>&, TIndex = indexAny
    ) const /*FastSearch(const TComparatorOf<T>&, TIndex)*/ = 0;
	//------------------------------------------------------------------------
    virtual const T* FastSearch(const T&) const;
	//------------------------------------------------------------------------
    const T* Search(const T& c_Data) const
    {
    	return FastSearch(c_Data);
    } /*Search(const T&) const*/
    //------------------------------------------------------------------------
    const T* Search(
    	const TConditionOf<T>& c_Condition, TIndex nIndex = indexAny
    ) const {
    	return TParent::Search(c_Condition, nIndex);
    } /*Search(const TConditionOf<T>&) const*/
    //------------------------------------------------------------------------
    const T* Search(
    	const TComparatorOf<T>& c_Comparator, TIndex nIndex = indexAny
    ) const {
    	return TParent::Search(c_Comparator, nIndex);
    } /*Search(const TComparatorOf<T>&) const*/

    //------------------------------------------------------------------------
    const T& FastFind(
		const TComparatorOf<T>&, TIndex = indexAny
	) const /*FastFind(const TComparatorOf<T>&, TIndex) const*/;
    //------------------------------------------------------------------------
    const T& FastFind(const T&) const;
    //------------------------------------------------------------------------
    const T& Find(const T& c_Data) const {return FastFind(c_Data);}

	//------------------------------------------------------------------------
    virtual TSize FastExist(const TComparatorOf<T>&) const = 0;
    //------------------------------------------------------------------------
    virtual TSize FastExist(const T&) const;
    //------------------------------------------------------------------------
    TSize Exist(const T& c_Data) const {return FastExist(c_Data);}
    //------------------------------------------------------------------------
    TSize Exist(const TConditionOf<T>& c_Condition) const
    {
    	return TParent::Exist(c_Condition);
    } /*Exist(const TConditionOf<T>&) const*/
    //------------------------------------------------------------------------
    TSize Exist(const TComparatorOf<T>& c_Comparator) const
    {
    	return TParent::Exist(c_Comparator);
    } /*Exist(const TComparatorOf<T>&) const*/

	// Seaches first by value (smallest) data in the container;
	//------------------------------------------------------------------------
    virtual const T* SearchFirst() const = 0;
	// Seaches last by value (greatest) data in the container;
	//------------------------------------------------------------------------
    virtual const T* SearchLast() const = 0;
    // Searches for element next by value (smallest creater than specified)
    //------------------------------------------------------------------------
    virtual const T* SearchNext(const T&) const = 0;
    // Searches for element previous by value (greatest less than specified)
    //------------------------------------------------------------------------
    virtual const T* SearchPrevious(const T&) const = 0;

	//------------------------------------------------------------------------
    const T& FindFirst() const;
	//------------------------------------------------------------------------
    const T& FindLast() const;
    // Searches for element next by value (smallest creater than specified)
    //------------------------------------------------------------------------
    const T& FindNext(const T&) const;
    // Searches for element previous by value (greatest less than specified)
    //------------------------------------------------------------------------
    const T& FindPrevious(const T&) const;

    //------------------------------------------------------------------------
	virtual TSize FastDelete(const TComparatorOf<T>&, TIndex = indexAll) = 0;

	//------------------------------------------------------------------------
    virtual TSize Delete(const TConditionOf<T>&, TIndex = indexAll);

    //------------------------------------------------------------------------
    TSize Delete(
		const TComparatorOf<T>& c_Comparator, TIndex nIndex = indexAll
    ){
		return TParent::Delete(c_Comparator, nIndex);
    } /*Delete(const TComparatorOf<T>&, TIndex = indexAll)*/
    //------------------------------------------------------------------------
    virtual TSize Delete(const T&);
   	//------------------------------------------------------------------------
    virtual TSize Delete(const TContainer&);

    // Applies specified operation in order defined by Sorting comparison
    //------------------------------------------------------------------------
    virtual void Inorder(TOperationOn<T>&) = 0;
    //------------------------------------------------------------------------
    virtual void Inorder(TConstantOperationOn<T>&) const = 0;

    //------------------------------------------------------------------------
    virtual void ForEach(TOperation&);
    //------------------------------------------------------------------------
    virtual void ForEach(TConstantOperation&) const;

	// Returns true if containers have same size and contain same elements
    //------------------------------------------------------------------------
    virtual bool IsEqual(const TContainer&) const;

	//------------------------------------------------------------------------
    const TSortingComparator& ComparatorConst() const {return Comparator();}

	//------------------------------------------------------------------------
	inline TSortingContainer& operator=(const TSortingContainer& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const TSortingContainer&)*/
	//------------------------------------------------------------------------
	inline TSortingContainer& operator=(const TContainer& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const TContainer&)*/

    //------------------------------------------------------------------------
    inline TSortingContainer& operator+=(const T& c_Data)
    {
    	Insert(c_Data); return *this;
    } /*operator+=(const T&)*/
    //------------------------------------------------------------------------
    inline TSortingContainer& operator+=(const TContainer& c_Source)
    {
    	Insert(c_Source); return *this;
    } /*operator+=(const TContainer&)*/

	//------------------------------------------------------------------------
    inline TSortingContainer& operator-=(const T& c_Data)
    {
    	Delete(c_Data); return *this;
    } /*operator-=(const T&)*/
	//------------------------------------------------------------------------
    inline TSortingContainer& operator-=(const TContainer& c_Source)
    {
    	Delete(c_Source); return *this;
    } /*operator-=(const TContainer&)*/
	//------------------------------------------------------------------------
	inline TSortingContainer& operator-=(const TConditionOf<T>& c_Condition)
	{
		Delete(c_Condition); return *this;
    } /*operator-=(const TConditionOf<T>&)*/
	//------------------------------------------------------------------------
	inline TSortingContainer& operator-=(const TComparatorOf<T>& c_Comparator)
    {
    	Delete(c_Comparator); return *this;
    } /*operator-=(const TComparatorOf<T>&)*/

    //------------------------------------------------------------------------
    bool operator==(const TContainer& c_Container) const
    {
    	return IsEqual(c_Container);
    } /*operator==(const TContainer&) const*/

protected:
    //------------------------------------------------------------------------
    TSortingContainerOf() : pComparator(null) {}

    //------------------------------------------------------------------------
    TSortingContainerOf(const TSortingComparator& c_Comparator):
		pComparator(null)
    {
    	SetComparator(c_Comparator);
    } /*TSortingContainerOf(const TSortingComparator&)*/

	// Returns current sort comparison
    //------------------------------------------------------------------------
    TSortingComparator& Comparator() const
    {
    	if (!pComparator)
			throw TXNoComparatorAssigned(THIS_CLASS);
		return *pComparator;
    } /*Comparator() const*/

    // Sets comparison
	//------------------------------------------------------------------------
    void SetComparator(const TSortingComparator&);

    // Clears comparison
    //------------------------------------------------------------------------
    void ClearComparator();

private:
	//------------------------------------------------------------------------
    TSortingComparator* pComparator;
} /*TSortingContainerOf<T>*/;

// TSortingContainerCheckInorder template
// Used for verification of sorting order in TSortingContainer::AssertValid()
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TSortingContainerCheckInorder :
	public TConstantOperationOn<T>
{
public:
	//------------------------------------------------------------------------
	typedef TConstantOperationOn<T> TParent;

    //------------------------------------------------------------------------
    typedef TEtalonComparatorOf<T> TSortingComparator;

    //------------------------------------------------------------------------
    TSortingContainerCheckInorder(
		const TSortingContainerOf<T>& c_Container
    ):
        nNodes(0), pc_Container(&c_Container), pComparator(null)
    {
		pComparator = CloneObject(c_Container.ComparatorConst());
	} /*TSortingContainerCheckInorder(const TSortingComparator&)*/

    //------------------------------------------------------------------------
    ~TSortingContainerCheckInorder() {delete pComparator; pComparator = null;}

    //------------------------------------------------------------------------
    size_t Nodes() const {return nNodes;}

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED1(TSortingContainerCheckInorder, T);

    //------------------------------------------------------------------------
    void operator()(const T&);

protected:
    //------------------------------------------------------------------------
	const TSortingContainerOf<T>* pc_Container;
    TSortingComparator* pComparator;
    //------------------------------------------------------------------------
    size_t nNodes;

private:
	//------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TSortingContainerCheckInorder<T>*/;

// TSortingContainerFindInContainer template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TSortingContainerFindInContainer :
	public TConstantOperationOn<T>
{
public:
	//------------------------------------------------------------------------
    typedef  TConstantOperationOn<T> TParent;
    //------------------------------------------------------------------------
    typedef TSortingContainerOf<T> TSortingContainer;

	//------------------------------------------------------------------------
	TSortingContainerFindInContainer(
    	const TSortingContainer& c_Container
    ):
    	pc_Container(&c_Container)
    {} /*TSortingContainerFindInContainer(const TSortingContainer&)*/

    //------------------------------------------------------------------------
    void operator()(const T& c_Data) {pc_Container->Find(c_Data);}

    //------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TSortingContainerFindInContainer, T);

protected:
	//------------------------------------------------------------------------
    const TSortingContainer* pc_Container;

private:
	//------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TSortingContainerFindInContainer<T>*/;

// TSortingContainerDeleteElement template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TSortingContainerDeleteElement :
	public TConstantOperationOn<T>
{
public:
	//------------------------------------------------------------------------
    typedef TConstantOperationOn<T> TParent;
    typedef TSortingContainerOf<T> TSortingContainer;
    typedef TSortingContainer::TSize TSize;

	//------------------------------------------------------------------------
    TSortingContainerDeleteElement(TSortingContainer& r_Container):
    	pContainer(&r_Container), nDeleted(0)
    {} /*TSortingContainerDeleteElement(TSortingContainer&)*/

    //------------------------------------------------------------------------
    void operator()(const T& c_Data)
    {
    	nDeleted += pContainer->Delete(c_Data);
    } /*operator()(const T&)*/

    //------------------------------------------------------------------------
    TSize Deleted() const {return nDeleted;}

    //------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TSortingContainerDeleteElement, T);

protected:
	//------------------------------------------------------------------------
    TSortingContainer* pContainer;
    TSize nDeleted;

private:
    //------------------------------------------------------------------------
    void operator()(T& r_Data) {operator()((const T&)r_Data);}
} /*TSortingContainerDeleteElement<T>*/;

// TSortingContainerOf template implementation
//============================================================================
template <class T> TSortingContainerOf<T>::~TSortingContainerOf()
{
	ClearComparator();
} /*~TSortingContainer()*/

//----------------------------------------------------------------------------
template <class T> void TSortingContainerOf<T>::AssertValid() const
{
	TSortingContainerCheckInorder<T> CheckInorder(*this);
	Inorder(CheckInorder);

	if (CheckInorder.Nodes() != Size()) 
		throw TXInvalidSize(THIS_CLASS);
} /*TSortingContainerOf<T>::AssertValid()*/

//----------------------------------------------------------------------------
template <class T> const T* TSortingContainerOf<T>::FastSearch(

	const T& c_Data

) const {
	TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
    	Comparator()
    ) /*hComparator*/;
    hComparator->SetEtalon(c_Data);
    return FastSearch(*hComparator, indexAny);
} /*TSortingContainerOf<T>::FastSearch(const T&) const*/;

//----------------------------------------------------------------------------
template <class T> const T& TSortingContainerOf<T>::FastFind(

   	const TComparatorOf<T>& c_Comparator, 
	TIndex nIndex /*= indexAny*/

) const {
   	const T* pc_Data = FastSearch(c_Comparator, nIndex);
	if (!pc_Data) 
		throw TXNotFound(THIS_CLASS);
	return *pc_Data;
} /*TSortingContainerOf<T>::FastFind(const TComparator, ...) const*/;

//------------------------------------------------------------------------
template <class T> const T& TSortingContainerOf<T>::FastFind(

	const T& c_Data
	
) const {
	const T* pc_Data = FastSearch(c_Data);
	if (!pc_Data) 
		throw TXNotFound(THIS_CLASS);
	return *pc_Data;
} /*TSortingContainerOf<T>::FastFind(const T&) const*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TSortingContainerOf<T>::FastExist(

	const T& c_Data

) const {
	TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
    	Comparator()
    ) /*hComparator*/;
    hComparator->SetEtalon(c_Data);
    return FastExist(*hComparator);
} /*TSortingContainerOf<T>::FastExist(const T&) const*/;

//----------------------------------------------------------------------------
template <class T> const T& TSortingContainerOf<T>::FindFirst() const
{
	const T* pc_Found = SearchFirst();
	if (!pc_Found) 
		throw TXNotFound(THIS_CLASS);
	return *pc_Found;
} /*TSortingContainerOf<T>::FindFirst() const*/

//----------------------------------------------------------------------------
template <class T> const T& TSortingContainerOf<T>::FindLast() const
{
	const T* pc_Found = SearchLast();
	if (!pc_Found) 
		throw TXNotFound(THIS_CLASS);
	return *pc_Found;
} /*TSortingContainerOf<T>::FindLast() const*/

//----------------------------------------------------------------------------
template <class T> const T& TSortingContainerOf<T>::FindNext(

	const T& c_Data
	
) const {
	const T* pc_Found = SearchNext(c_Data);
	if (!pc_Found) 
		throw TXNotFound(THIS_CLASS);
	return *pc_Found;
} /*TSortingContainerOf<T>::FindNext(const T&) const*/

//----------------------------------------------------------------------------
template <class T> const T& TSortingContainerOf<T>::FindPrevious(

	const T& c_Data
	
) const {
	const T* pc_Found = SearchPrevious(c_Data);
	if (!pc_Found) 
		throw TXNotFound(THIS_CLASS);
	return *pc_Found;
} /*TSortingContainerOf<T>::FindPrevious(const T&) const*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TSortingContainerOf<T>::Delete(

	const TConditionOf<T>& c_Condition,
    nl_ulong_t nIndex

){
	TSize nDeleted = 0;
	if (nIndex != indexAll) {
		const T* pc_Data = Search(c_Condition, nIndex);
        if (pc_Data) Delete(*pc_Data);
		nDeleted = (pc_Data != null);
    }
    else {
    	const T* pc_Data = Search(c_Condition, indexFirst);
        while (pc_Data) {
        	const T* pc_DataToDelete = pc_Data;

            // Get next by condition
            //-----------------------
            do {
            	pc_Data = SearchNext(*pc_Data);
            } while (pc_Data && !c_Condition(*pc_Data));

            Delete(*pc_DataToDelete);
            ++nDeleted;
        } /*while (pc_Data)*/
    } /*if*/
    return nDeleted;
} /*TSortingContainerOf<T>::Delete(const TConditionOf&, TIndex)*/;

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TSortingContainerOf<T>::Delete(

	const T& c_Data

){
	TAutoRestoreEtalonComparatorHandleOf<T> hComparator(
    	Comparator()
    ) /*hComparator*/;
    hComparator->SetEtalon(c_Data);
	return FastDelete(*hComparator, indexFirst);
} /*TSortingContainerOf<T>::Delete(const T&)*/;

//----------------------------------------------------------------------------
template <class T> 
TSortingContainerOf<T>::TSize TSortingContainerOf<T>::Delete(

	const TContainer& c_Container
	
){
	TSortingContainerDeleteElement<T> DeleteFromThis(*this);
	c_Container.ForEach(DeleteFromThis);
	return DeleteFromThis.Deleted();
} /*TSortingContainerOf<T>::Delete(const TContainer&)*/

//----------------------------------------------------------------------------
template <class T> void TSortingContainerOf<T>::ForEach(

	TOperationOn<T>& r_Operation

){
	Inorder(r_Operation);
} /*TSortingContainerOf<T>::ForEach(TOperation&)*/

//----------------------------------------------------------------------------
template <class T> void TSortingContainerOf<T>::ForEach(

	TConstantOperationOn<T>& r_ConstantOperation

) const {
	Inorder(r_ConstantOperation);
} /*TSortingContainerOf<T>::ForEach(TConstantOperation&) const*/

//----------------------------------------------------------------------------
template <class T> bool TSortingContainerOf<T>::IsEqual(

	const TContainerOf<T>& c_Container

) const {
	return IsEqualByComparator(c_Container, Comparator());
} /*TSortingContainerOf<T>::IsEqual(const TContainer&) const*/

//----------------------------------------------------------------------------
template <class T> void TSortingContainerOf<T>::SetComparator(

	const TEtalonComparatorOf<T>& c_SortingComparator

){
	TAutoPointerTo<TSortingComparator> hNewComparator(
    	CloneObject(c_SortingComparator), false
    ) /*hNewComparator*/;

    if (hNewComparator.IsNull()) {
    	throw TSortingComparator::TXCanNotBeCloned(
			THIS_PLACE_CLASS(c_SortingComparator)
		) /*throw TSortingComparator::TXCanNotBeCloned*/;
	} /*if*/

   	ClearComparator();

	pComparator = hNewComparator.Unassign(false);
} /*TSortingContainerOf::SetComparator(const TSortingComparator&)*/;

//----------------------------------------------------------------------------
template <class T> void TSortingContainerOf<T>::ClearComparator()
{
    try {delete pComparator; pComparator = null;}
    catch (...) {pComparator = null; throw;}
} /*TSortingContainerOf<T>::ClearComparator()*/;

// TSortingContainerCheckInorder template implementation
//============================================================================
template <class T> void TSortingContainerCheckInorder<T>::operator()(

	const T& c_Data

){
    if (pComparator->EtalonPtr()) {
        TComparator::TResult ComparatorResult = (*pComparator)(c_Data);

        if (ComparatorResult == TSortingComparator::compUndefined) {
            throw TSortingContainerOf<T>::TXUndefinedComparisonResult(
                	THIS_PLACE_CLASS(*pc_Container)
            ) /*throw TSortingContainer::TXUndefinedComparisonResult*/;
        } /*if*/

        if (ComparatorResult != TSortingComparator::compLess) {
            throw TSortingContainerOf<T>::TXBadInorder(
				THIS_PLACE_CLASS(*pc_Container)
			) /*throw TSortingContainerOf<T>::TXBadInorder*/;
		} /*if*/
    } /*if*/
    ++nNodes;
    pComparator->SetEtalon(c_Data);
} /*TSortingContainerCheckInorder<T>::operator()(const T&)*/

// TSortingContainerFindInContainer template implementation
//============================================================================
template <class T>
NLOBJECT_CLONE_RESULT1(TSortingContainerFindInContainer, T)
TSortingContainerFindInContainer<T>::Clone() const
{
	TSortingContainerFindInContainer<T>* pCopy;
    NLEXCEPT_CLASS_NEW(
    	TSortingContainerFindInContainer<T>(*pc_Container),
        pCopy
    ) /*NLEXCEPT_CLASS_NEW*/;
    return pCopy;
} /*TSortingContainerFindInContainer<T>::Clone() const*/

// TSortingContainerDeleteElement template implementation
//============================================================================
template <class T>
NLOBJECT_CLONE_RESULT1(TSortingContainerDeleteElement, T)
TSortingContainerDeleteElement<T>::Clone() const
{
	TSortingContainerDeleteElement<T>* pCopy;
    NLEXCEPT_CLASS_NEW(
    	TSortingContainerDeleteElement<T>(*pContainer), pCopy
    ) /*NLEXCEPT_CLASS_NEW*/;
    return pCopy;
} /*TSortingContainerDeleteElement<T>::Clone() const*/

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLSCONT_DECLSPEC(T, Declspec) \
	NLCONT_DECLSPEC(T, Declspec); \
    NATLIB_EXPAND_TEMPLATE1( \
    	NATLIB_NAMESPACE_MEMBER(TSortingContainerCheckInorder), \
		T, Declspec \
    ) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
    	NATLIB_NAMESPACE_MEMBER(TSortingContainerFindInContainer), \
		T, Declspec \
    ) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
    	NATLIB_NAMESPACE_MEMBER(TSortingContainerDeleteElement), \
		T, Declspec \
    ) /*NATLIB_EXPAND_TEMPLATE1*/; \
    NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TSortingContainerOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLSCONT_H__)*/
