#ifndef __NATLIB_NLTIME_H__
#define __NATLIB_NLTIME_H__
/****************************************************************************/
/** nltime.h                                                               **/
/** ---------------------------------------------------------------------- **/
/** Time & Date handling declaration                                       **/
/** ---------------------------------------------------------------------- **/
/** Current impementation is based mostly on C runtime library so some     **/
/** calculations made assuming starting from 01.01.1970, hereafter Epoch.  **/
/** Along with that, absolute time is stored as a number of days passed    **/
/** since Anno Domini and as a number of milliseconds passed since the     **/
/** beginning of the day. Days can be negative as well, but seconds are    **/
/** always positive: to represent -1 seconds construct -1 day and          **/
/** NumberOfMSecInDay - 1000.                                              **/
/** For now, calculations of calendar date are limited to 1752-2038 years, **/
/** but for time difference and comparison the whole range is OK.          **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** May 09, 1997 by Ilya Marchevsky                                        **/
/****************************************************************************/

// Standard includes
//============================================================================
#include <time.h>

// Natlib Basics includes
//============================================================================
#include <nlobject.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TTime
//============================================================================
class _NATLIB_DECLSPEC_ TTime : public TNatlibObject {
public:
	//------------------------------------------------------------------------
    typedef TNatlibObject TParent;

	// TCalendarForm
    //------------------------------------------------------------------------
    typedef struct TCalendarForm {
    	nl_long_t nYear;
		nl_long_t nMonth;
        nl_long_t nDay;
        nl_long_t nHour;
        nl_long_t nMinute;
        nl_long_t nSecond;
        nl_long_t nMSecond;
    } typedef_TCalendarForm;

	// TCalendarFormRequest
    //------------------------------------------------------------------------
	typedef struct TCalendarFormRequest : public TCalendarForm {
        nl_int_t WeekDay;
        nl_int_t YearDay;
        nl_long_t SecondsDiffFromGMT;
    } typedef_TCalendarFormRequest;

	//------------------------------------------------------------------------
	static const nl_long_t c_nDaysOnEpoch;
    static const nl_long_t c_nSecondsInDay;
    static const nl_long_t c_nMSecondsInDay;
    static const nl_long_t c_nDaysInWeek;
    static const nl_long_t c_nMonthsInYear;
    static const nl_long_t c_nHoursInDay;
    static const nl_long_t c_nMinutesInHour;
    static const nl_long_t c_nSecondsInMinute;

    //------------------------------------------------------------------------
	static const nl_int_t const cac_nDaysInMonth[];
	static const nl_int_t const cac_nStartOfEachMonth[];

	//------------------------------------------------------------------------
	static const char* const cac_szMonthNames[];
	static const char* const cac_szWeekDayNames[];

	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
		TException, TNatlibObject::TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXWeekDayOutOfBounds, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXMonthOutOfBounds, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXDateBeyondCalculations, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXMsecOutOfBounds, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXSecondsOutOfBounds, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXMinuteOutOfBounds, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;
    NLEXCEPT_DECLARE_OUT_OF_RANGE(
		TXHoursOutOfBounds, TException, _NATLIB_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OUT_OF_RANGE*/;

	// Calculates internal day storage from Gregorian calendar date.
	// Algorithm 199 from Communications of the ACM, Volume 6, No.
	// 8, (Aug. 1963), p. 444.  Gregorian calendar started on Sep. 14, 1752.
	// This function not valid before that.
	// ToDo: what about future time limitations?
    //------------------------------------------------------------------------
    static TTime FromTime(const TCalendarForm& c_CalendarForm);

    // this functions assumed that will never be called before 01.01.1970
    // Now - Seconds arter midnight 01.01.1970 GMT as time() returns
    // Argument assumed to be result of time() function somehow
    //------------------------------------------------------------------------
    static TTime FromSeconds(time_t SecondsSinceEpoch)
    {
        return (
        	TTime(
            	SecondsSinceEpoch / c_nSecondsInDay + c_nDaysOnEpoch,
                SecondsSinceEpoch % c_nSecondsInDay
            ) /*TTime*/
        ) /*return*/;
    } /*FromSeconds(time_t)*/

    // this functions assumed that will never be called before Epoch
    //------------------------------------------------------------------------
    static TTime Now() {return FromSeconds(time(0));}

    //------------------------------------------------------------------------
    static const char* DayOfWeekName(nl_int_t nDayOfWeek);
    //------------------------------------------------------------------------
    static nl_int_t DayOfWeekFromName(const char* c_szDayName );
    //------------------------------------------------------------------------
    static nl_int_t MonthFromName(const char* c_szMonthName);
    //------------------------------------------------------------------------
    static const char* MonthName(nl_int_t nMonth);
    //------------------------------------------------------------------------
    static nl_int_t DaysInYear(nl_long_t nYear) 
	{
		return LeapYear(nYear) ? 366 : 365;
	} /*DaysInYear(nl_long_t)*/
    //------------------------------------------------------------------------
    static bool LeapYear(nl_long_t nYear)
    {
    	return !(nYear & 3) && (nYear % 100) || !(nYear % 400);
    } /*LeapYear(nl_long_t)*/
    //------------------------------------------------------------------------
    static void SetDefaultFormat(const char* szFormat)
    {
    	strcpy(c_szDefaultFormat, szFormat);
    } /*SetDefaultFormat(const char*)*/

    // Construct a TTime objects with the current date/time
    //------------------------------------------------------------------------
    TTime()
    {
     	const time_t SecondsSinceEpoch = time(0);
        nDays = SecondsSinceEpoch / c_nSecondsInDay + c_nDaysOnEpoch;
        nMSeconds = (SecondsSinceEpoch % c_nSecondsInDay) * 1000UL;
    } /*TTime()*/
    //------------------------------------------------------------------------
    TTime(const TTime& c_Src) :
    	nDays(c_Src.nDays), nMSeconds(c_Src.nMSeconds)
    {} /*TTime(const TTime&)*/

    //------------------------------------------------------------------------
    TTime(
    	nl_long_t nYear, nl_int_t nMonth, nl_int_t nDay,
        nl_int_t nHour = 0, nl_int_t nMinute = 0, 
		nl_int_t nSec = 0, nl_int_t nMSec = 0
    ) /*TTime(nl_long_t, nl_int_t, nl_int_t, nl_int_t = 0, ...)*/;

    //------------------------------------------------------------------------
    TTime(const tm& c_Tm);

	//------------------------------------------------------------------------
    TTime(nl_long_t InitDays, nl_ulong_t InitSec) :
    	nDays(InitDays), nMSeconds(InitSec * nl_ulong_t(1000))
    {} /*TTime(nl_long_t, nl_ulong_t)*/

    //------------------------------------------------------------------------
	TStdString Dump() const;

    //------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE(TTime);

    //------------------------------------------------------------------------
	virtual TStdString AsString(const char* c_szCustomFormat = null) const;

    //------------------------------------------------------------------------
    bool Equal(const TTime& c_Time) const
	{
    	return (
        	(nDays == c_Time.nDays) && (nMSeconds == c_Time.nMSeconds)
        ) /*return*/;
    } /*Equal(const TTime&) const*/

    //------------------------------------------------------------------------
    bool Before(const TTime& c_Time) const
    {
    	return (
        	(nDays < c_Time.nDays) ||
            ((nDays == c_Time.nDays) && (nMSeconds < c_Time.nMSeconds))
        ) /*return*/;
	} /*Before(const TTime&) const*/
    //------------------------------------------------------------------------
    bool After(const TTime& c_Time) const
    {
    	return (
        	(nDays > c_Time.nDays) ||
            (
            	(nDays == c_Time.nDays) &&
                (nMSeconds > c_Time.nMSeconds)
            )
        ) /*return*/;
	} /*After(const TTime&) const*/

    //------------------------------------------------------------------------
    bool InRange(const TTime& c_Low, const TTime& c_High) const
    {
    	return (!Before(c_Low) && !After(c_High));
    } /*InRange(const TTime&, const TTime&) const*/

	//----------------------------------------------------------------------------
    virtual TCalendarForm& Time(TCalendarForm& r_CalendarForm) const;

	// Calculates Gregorian calendar date from internally stored number of
    // days passed since anno domini
	// Algorithm 199 from "Communications of the ACM", Volume 6, No. 8,
	// (Aug. 1963), p. 444.  Gregorian calendar started on Sep. 14, 1752.
	// This function not valid before that.
	// Note: this function first transforms GMT to local and after that
    // calculates
	// ToDo: what about future time limitations?
    //------------------------------------------------------------------------
    TCalendarForm& TimeGMT(TCalendarForm& r_CalendarForm) const;

    //------------------------------------------------------------------------
    nl_long_t Year() const
	{
    	TCalendarForm CalendarForm; return Time(CalendarForm).nYear;
	} /*Year() const*/
    //------------------------------------------------------------------------
    nl_int_t Month() const
	{
		TCalendarForm CalendarForm; return Time(CalendarForm).nMonth;
	} /*Month() const*/
    //------------------------------------------------------------------------
    nl_int_t Day() const
	{
		TCalendarForm CalendarForm; return Time(CalendarForm).nDay;
	} /*Day() const*/
    //------------------------------------------------------------------------
    nl_int_t Hour() const
	{
		TCalendarForm CalendarForm; return Time(CalendarForm).nHour;
	} /*Hour() const*/
    //------------------------------------------------------------------------
    nl_int_t Minute() const
	{
		TCalendarForm CalendarForm; return Time(CalendarForm).nMinute;
	} /*Minute() const*/
    //------------------------------------------------------------------------
    nl_int_t Second() const {return (nMSeconds % 1000) / c_nSecondsInMinute;}
    //------------------------------------------------------------------------
    nl_int_t MSecond() const {return nMSeconds % 1000;}

    //------------------------------------------------------------------------
    nl_int_t DayOfYear() const
    {
    	return (nDays - TTime(Year() - 1, 12).nDays);
    } /*DayOfYear() const*/
    //------------------------------------------------------------------------
    nl_int_t DayOfWeek() const;

    //------------------------------------------------------------------------
    nl_long_t GetDays() const {return nDays;}
    //------------------------------------------------------------------------
    nl_ulong_t GetMSeconds() const {return nMSeconds;}

    //------------------------------------------------------------------------
    void SetDate(nl_long_t nYear, nl_int_t nMonth, nl_int_t nDay);
    //------------------------------------------------------------------------
    void SetTime(
		nl_int_t nHour, nl_int_t nMinute = 0, nl_int_t nSecond = 0, 
		nl_int_t nMSec = 0
	) /*SetTime(nl_int_t, nl_int_t = 0, nl_int_t = 0, nl_int_t = 0)*/;

    //------------------------------------------------------------------------
    void SetDateTime(const TCalendarForm& c_CalendarForm)
    {
    	*this = FromTime(c_CalendarForm);
    } /*SetDateTime(const TCalendarForm&)*/

    //------------------------------------------------------------------------
    void AddYears(nl_long_t nYearsToAdd);
    //------------------------------------------------------------------------
    void AddMonths(nl_long_t nMonthsToAdd);
    //------------------------------------------------------------------------
    void AddWeeks(nl_long_t nWeeksToAdd)
    {
    	nDays += (nWeeksToAdd * c_nDaysInWeek);
    } /*AddWeeks(nl_long_t)*/
    //------------------------------------------------------------------------
    void AddDays(nl_long_t nDaysToAdd) {nDays += nDaysToAdd;}
    //------------------------------------------------------------------------
    void AddMSeconds(nl_ulong_t nMSecToAdd)
	{
    	nDays += (nMSeconds + nMSecToAdd) / c_nMSecondsInDay;
        nMSeconds = (nMSeconds + nMSecToAdd) % c_nMSecondsInDay;
	} /*AddMSeconds(nl_ulong_t)*/
    //------------------------------------------------------------------------
    void AddTime(const TTime& c_ToAdd)
	{
    	AddMSeconds(c_ToAdd.nMSeconds); AddDays(c_ToAdd.nDays);
	} /*AddTime(const TTime&)*/

    //------------------------------------------------------------------------
    void SubtractMSeconds(nl_ulong_t nMSecToSubstract)
	{
	    // take one day to ensure > 0
        //----------------------------
        nMSeconds = (
        	nMSeconds + c_nMSecondsInDay - nMSecToSubstract
        ) /*nMSeconds*/;

    	nDays += nMSeconds/c_nMSecondsInDay - 1;
        nMSeconds %= c_nMSecondsInDay;
	} /*SubtractMSeconds(nl_ulong_t)*/

    //------------------------------------------------------------------------
    void DiffTime(const TTime& c_Sub)
	{
    	SubtractMSeconds(c_Sub.nMSeconds); AddDays(-c_Sub.nDays);
	} /*DiffTime(const TTime&)*/

    //------------------------------------------------------------------------
    tm& ToTm(tm& Tm) const;

    //------------------------------------------------------------------------
    virtual void ToLocal() {AddTime(DiffLocalFromGMT);}
    //------------------------------------------------------------------------
    virtual void ToGreenwich() {DiffTime(DiffLocalFromGMT);}

    //------------------------------------------------------------------------
	TTime& operator=(const TTime& c_Src)
    {
    	nDays = c_Src.nDays;
        nMSeconds = c_Src.nMSeconds;
        return *this;
    } /*operator=(const TTime&)*/

    //------------------------------------------------------------------------
    bool operator<(const TTime& c_Time) const {return Before(c_Time);}
    //------------------------------------------------------------------------
    bool operator<=(const TTime& c_Time) const {return !After(c_Time);}
    //------------------------------------------------------------------------
    bool operator>(const TTime& c_Time) const {return After(c_Time);}
    //------------------------------------------------------------------------
    bool operator>=(const TTime& c_Time) const {return !Before(c_Time);}
    //------------------------------------------------------------------------
    bool operator==(const TTime& c_Time) const {return Equal(c_Time);}
    //------------------------------------------------------------------------
    bool operator!=(const TTime& c_Time) const {return !Equal(c_Time);}
    //------------------------------------------------------------------------
    TTime operator+(const TTime& c_Add) const;
    //------------------------------------------------------------------------
    TTime operator-(const TTime& c_Sub) const;
    //------------------------------------------------------------------------
    TTime operator-() const;
    //------------------------------------------------------------------------
    TTime& operator+=(const TTime& c_Add) {AddTime(c_Add); return *this;}
    //------------------------------------------------------------------------
    TTime& operator-=(const TTime& c_Sub) {DiffTime(c_Sub); return *this;}

protected:
	//------------------------------------------------------------------------
    static nl_long_t SecondsDiffLocalFromGMT;
    static TTime DiffLocalFromGMT;
    static char c_szDefaultFormat[100];  // Format for I/O operations

    //------------------------------------------------------------------------
    static nl_long_t SecondsLocalToGreenwich(nl_ulong_t Seconds)
    {
    	return Seconds - SecondsDiffLocalFromGMT;
    } /*SecondsLocalToGreenwich(nl_ulong_t)*/
    //------------------------------------------------------------------------
    static nl_long_t SecondsGreenwichToLocal(nl_ulong_t Seconds)
    {
    	return Seconds + SecondsDiffLocalFromGMT;
    } /*SecondsGreenwichToLocal(nl_ulong_t)*/

	// Performs case-insensitive lookup of string throw the array of possible
	// values. Returns its index if found, -1 otherwise
	// Note: only strlen(Search) letters are compared
	//------------------------------------------------------------------------
	static nl_int_t PatternLookup(
    	const char* c_szSearch,
        const char* const cac_szPattern[], size_t nPatterns
    ) /*PatternLookup(const char*, const char* const[], size_t)*/;

	//------------------------------------------------------------------------
    nl_long_t nDays;
    nl_ulong_t nMSeconds;

private:
    //------------------------------------------------------------------------
    static void Initialize();
	//------------------------------------------------------------------------
	static const class _NATLIB_DECLSPEC_ TInitializer {
    public:
    	TInitializer() {TTime::Initialize();}
    } Initializer;
    //------------------------------------------------------------------------
	friend class TInitializer;
} /*TTime*/;

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLTIMDAT_DECLSPEC(TData, Declspec) \
	NLOBJECT_DECLSPEC(TData, Declspec);

#endif /*!defined(__NATLIB_NLTIME_H__)*/
