#ifndef __NATLIB_NLVECTOR_H__
#define __NATLIB_NLVECTOR_H__
/****************************************************************************/
/** nlvector.h                                                             **/
/** ---------------------------------------------------------------------- **/
/** Natlib vector is storage of a data by size_t                           **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** July 30, 1997 by Oleg Kagan                                            **/
/****************************************************************************/

//============================================================================
#include <nlstorag.h>
#include <nlalloc.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

//============================================================================
#define NLVECTOR_ALLOCATOR_DEFAULT TStdAllocator

//============================================================================
#define NLVECTOR_INDEX_VALIDITY_GUARD(Index) \
	{if (!IsValidIndex(Index)) throw TXOutOfRange(THIS_CLASS);}
//----------------------------------------------------------------------------
#define NLVECTOR_ELEMENT_SIZE(Type) \
	sizeof(TAllocationEnvelopeOf<Type>)
//----------------------------------------------------------------------------
#define NLVECTOR_MAX_SIZE(Type) \
	(~size_t(0) / NLVECTOR_ELEMENT_SIZE(Type))

//============================================================================
#define NLVECTOR_INVALID_INDEX (~size_t(0))
#define NLVECTOR_NPOS NLVECTOR_INVALID_INDEX

// TVectorOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TVectorOf :
	public TStorageOf<T, size_t>
{
public:
	//------------------------------------------------------------------------
	typedef TStorageOf<T, size_t> TParent;
	typedef TVectorOf<T> TVector;
	typedef TStorageOf<T, size_t> TParentStorage;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
		TException, TParentStorage::TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(
		TXAllocatorIsNotDefined, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE_OPERATION_IS_NOT_ALLOWED(
		TXOperationIsNotllowed, TException, _NATLIB_TEMPLATE_DECLSPEC_
	) /*NLEXCEPT_DECLARE_OPERATION_IS_NOT_ALLOWED*/;

	//------------------------------------------------------------------------
	TVectorOf(const TAllocator& = NLVECTOR_ALLOCATOR_DEFAULT());
	//------------------------------------------------------------------------
	TVectorOf(
		const T& c_Filler,
		size_t nInitSize = 1, size_t nInitCapacity = 1,
		const TAllocator& = NLVECTOR_ALLOCATOR_DEFAULT()
	) /*TVectorOf(const T&, size_t, size_t = 0, const TAllocator&)*/;
	//------------------------------------------------------------------------
	TVectorOf(const TVectorOf<T>&);
	//------------------------------------------------------------------------
	TVectorOf(const TVectorOf<T>&, const TAllocator&);
	//------------------------------------------------------------------------
	TVectorOf(
		const TContainerOf<T>&,
		const TAllocator& = NLVECTOR_ALLOCATOR_DEFAULT()
	) /*TVectorOf(const TContainerOf<T>, const TAllocator&)*/;

	//------------------------------------------------------------------------
	~TVectorOf();

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE1(TVectorOf, T);

	//------------------------------------------------------------------------
	TContainerOf<T>::TSize Size() const {return nSize;}

	// Storage specific functions
	//------------------------------------------------------------------------
	const T* SearchByKeyConst(const size_t& c_uIndex) const;
	//------------------------------------------------------------------------
	T* SearchByKey(const size_t& c_uIndex);

	// Next functions return NLVECTOR_NPOS if failed
	//------------------------------------------------------------------------
	size_t SearchIndex(
		const TConditionOf<T>&, TContainerOf<T>::TIndex = indexFirst
	) const /*SearchIndex(const TConditionOf<T>&, TIndex = indexFirst*/;
	//------------------------------------------------------------------------
	size_t SearchIndex(
		const TComparatorOf<T>&, TContainerOf<T>::TIndex = indexFirst
	) const /*SearchIndex(const TComparatorOf<T>&, TIndex = indexFirst*/;
	//------------------------------------------------------------------------
	size_t BinarySearchIndex(const TComparatorOf<T>&) const;

	// Next functions trow TXNotFound if failed
	//------------------------------------------------------------------------
	size_t FindIndex(const TConditionOf<T>&) const;
	//------------------------------------------------------------------------
	size_t FindIndex(const TComparatorOf<T>&) const;
	//------------------------------------------------------------------------
	size_t BinaryFindIndex(const TComparatorOf<T>&) const;

	//------------------------------------------------------------------------
	size_t FirstKey() const;
	//------------------------------------------------------------------------
	size_t LastKey() const;
	//------------------------------------------------------------------------
	size_t NextKey(const size_t& c_uKey) const
	{
		if (!(nSize && c_uKey < (nSize - 1))) 
			throw TXOutOfRange(THIS_CLASS);
		return c_uKey + 1;
	} /*NextKey(const size_t&) const*/
	//------------------------------------------------------------------------
	size_t PreviousKey(const size_t& c_uKey) const
	{
		if (!(c_uKey && (c_uKey < nSize))) 
			throw TXOutOfRange(THIS_CLASS);
		return c_uKey - 1;
	} /*PreviousKey(const size_t& c_uKey) const*/

	//------------------------------------------------------------------------
	T& Assign(const T&, const size_t&);
	// Absolete and throws TXOperationIsNotAllowed exception
	//------------------------------------------------------------------------
	bool Unassign(const size_t&);

	//------------------------------------------------------------------------
	void Inorder(TStorageOperation&);
	//------------------------------------------------------------------------
	void Inorder(TStorageConstantOperation&) const;
	//------------------------------------------------------------------------
	void Postorder(TStorageOperation&);
	//------------------------------------------------------------------------
	void Postorder(TStorageConstantOperation&) const;

	// Vector specific operations
	//------------------------------------------------------------------------
	virtual TContainerOf<T>::TSize Capacity() const {return nCapacity;}
	
	//------------------------------------------------------------------------
	virtual bool IsValidIndex(size_t uIndex) const {return uIndex < Size();}

	//------------------------------------------------------------------------
	void QuickSort(TEtalonComparatorOf<T>&);
	//------------------------------------------------------------------------
	void QuickSort(TEtalonComparatorOf<T>&, size_t, size_t);
	//------------------------------------------------------------------------
	void QuickSortDirect(TEtalonComparatorOf<T>&, size_t, size_t);
	//------------------------------------------------------------------------
	void QuickSortVirtual(TEtalonComparatorOf<T>&, size_t, size_t);

	// Boolean parameter specifies whether to perform unconditional
	// reallocation
	//------------------------------------------------------------------------
	virtual void Resize(
		size_t nNewSize, size_t nNewCapacity, const T& c_Filler,
		bool bUnconditional
	) /*Resize(size_t, size_t, const T&, bool)*/;
	//------------------------------------------------------------------------
	void Resize(size_t nNewSize, const T& c_Filler)
	{
		Resize(nNewSize, nNewSize, c_Filler, false);
	} /*Resize(size_t, const T&)*/

	// Reduces size of the vector down to specified returns resulting size
	//------------------------------------------------------------------------
	size_t ReduceSize(size_t nNewSize, bool bUnconditional = false);
	//------------------------------------------------------------------------
	size_t ReduceSize(
		size_t nNewSize, size_t nNewCapacity, bool bUnconditional
	) /*ReduceSize(size_t, size_t, bool)*/;

	//------------------------------------------------------------------------
	size_t Reserve(size_t nNewCapacity, bool bUnconditional = true);

	//------------------------------------------------------------------------
	virtual T& Element(size_t uIndex)
	{
		NLVECTOR_INDEX_VALIDITY_GUARD(uIndex);
		return a_Memory[uIndex].Data;
	} /*Element(size_t)*/
	//------------------------------------------------------------------------
	virtual const T& ElementConst(size_t uIndex) const
	{
		NLVECTOR_INDEX_VALIDITY_GUARD(uIndex);
		return a_Memory[uIndex].Data;
	} /*ElementConst(size_t) const*/

	//------------------------------------------------------------------------
	const T& Insert(const T&);
	//------------------------------------------------------------------------
	void Insert(const TContainerOf<T>& c_Source)
	{
		TParent::Insert(c_Source);
	} /*Insert(const TContainerOf<T>&)*/
	//------------------------------------------------------------------------
	void Insert(const TVectorOf<T>&);
	//------------------------------------------------------------------------
	void Insert(const TStorageOf<T, size_t>& c_Storage)
	{
		TParent::Insert(c_Storage);
	} /*Insert(const TStorageOf<T, size_t>&)*/

	//------------------------------------------------------------------------
	TContainerOf<T>::TSize Delete(
		const TConditionOf<T>&, TContainerOf<T>::TIndex = indexAll
	) /*Delete(const TConditionOf<T>&, TIndex = indexAll*/;
	//------------------------------------------------------------------------
	TContainerOf<T>::TSize Delete(
		const TComparatorOf<T>& c_Comparator, 
		TContainerOf<T>::TIndex nIndex = indexAll
	){
		return TParent::Delete(c_Comparator, nIndex);
	} /*Delete(const TComparatorOf<T>&, TIndex = indexAll)*/
	//------------------------------------------------------------------------
	void Clear() {ReduceSize(0, 0, false);}

	//------------------------------------------------------------------------
	void Let(const TContainerOf<T>&);
	//------------------------------------------------------------------------
	void Let(const TVectorOf<T>&);
	//------------------------------------------------------------------------
	void Let(const TStorageOf<T, size_t>& c_Source)
	{
		TParent::Let(c_Source);
	} /*Let(const TStorageOf<T, size_t>&)*/

	//------------------------------------------------------------------------
	void Swap(size_t, size_t);

	//------------------------------------------------------------------------
	T& RawElement(size_t uIndex) {return a_Memory[uIndex].Data;}

	//------------------------------------------------------------------------
	const TAllocator& AllocatorConst() const;

	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator=(const TContainerOf<T>& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const const TContainerOf<T>&)*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator=(const TVectorOf<T>& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const TVectorOf<T>&)*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator=(const TStorageOf<T, size_t>& c_Source)
	{
		Let(c_Source); return *this;
	} /*operator=(const TStorageOf<T, size_t>&)*/

    //------------------------------------------------------------------------
	inline TVectorOf<T>& operator+=(const T& c_Data)
    {
    	Insert(c_Data); return *this;
    } /*operator+=(const T&)*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator+=(const TContainerOf<T>& c_Source)
	{
		Insert(c_Source); return *this;
	} /*operator+=(const TContainerOf<T>&*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator+=(const TVectorOf<T>& c_Source)
    {
    	Insert(c_Source); return *this;
	} /*operator+=(const TVectorOf<T>&)*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator+=(const TStorageOf<T, size_t>& c_Storage)
	{
		Insert(c_Storage); return *this;
	} /*operator+=(const TStorageOf<T, size_t>&)*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator-=(const TConditionOf<T>& c_Condition)
    {
    	Delete(c_Condition); return *this;
    } /*operator-=(const TConditionOf<T>&)*/
	//------------------------------------------------------------------------
	inline TVectorOf<T>& operator-=(const TComparatorOf<T>& c_Comparator)
    {
    	Delete(c_Comparator); return *this;
    } /*operator-=(const TComparatorOf<T>&)*/

	//------------------------------------------------------------------------
	inline T& operator[](size_t uIndex) {return Element(uIndex);}
	//------------------------------------------------------------------------
	inline const T& operator[](size_t uIndex) const
	{
		return ElementConst(uIndex);
	} /*operator[](size_t) const*/

protected:
	//------------------------------------------------------------------------
	TAllocationEnvelopeOf<T>* a_Memory;
	size_t nSize, nCapacity;
	TAllocator* pAllocator;

	//------------------------------------------------------------------------
	static void ConstructElements(
		TAllocationEnvelopeOf<T>* pTarget,
		const T& c_Filler,
		size_t nElementsToConstruct
	) /*ConstructElements(TAllocationEnvelopeOf<T>*, const T&,...)*/;
	//------------------------------------------------------------------------
	static void ConstructElements(
		TAllocationEnvelopeOf<T>* pTarget,
		const TAllocationEnvelopeOf<T>* pSource,
		size_t nElementsToConstruct
	) /*ConstructElements(TAllocationEnvelopeOf<T>*, ...)*/;
	//------------------------------------------------------------------------
	static void DestructElements(TAllocationEnvelopeOf<T>*, size_t);
	//------------------------------------------------------------------------
	static void ReconstructElements(
		TAllocationEnvelopeOf<T>* pTarget,
		const T& c_Filler,
		size_t nElementsToReconstruct
	) /*ReconstructElements(TAllocationEnvelopeOf<T>*, ...)*/;
	//------------------------------------------------------------------------
	static void ReconstructElements(
		TAllocationEnvelopeOf<T>* pTarget,
		const TAllocationEnvelopeOf<T>* pc_Source,
		size_t nElementsToReconstruct
	) /*ReconstructElements(TAllocationEnvelopeOf<T>*, ...)*/;
	//------------------------------------------------------------------------
	TAllocationEnvelopeOf<T>* AllocateMemory(size_t nElements);
	//------------------------------------------------------------------------
	void DeallocateMemory(TAllocationEnvelopeOf<T>*& ra_ToDelete)
	{
		pAllocator->Deallocate(ra_ToDelete); ra_ToDelete = null;
	} /*DeallocateMemory(TAllocationEnvelopeOf<T>*)*/
} /*TVectorOf<T>*/;

// TVectorQuickSortOf template
//============================================================================
template <class T> class _NATLIB_TEMPLATE_DECLSPEC_ TVectorQuickSortOf {
public:
	//------------------------------------------------------------------------
	TVectorQuickSortOf(
		TVectorOf<T>* pVectorToSort,
		TEtalonComparatorOf<T>& r_Comparator
	):
		pVector(pVectorToSort),
		hEtalonRestore(r_Comparator), pComparator(&r_Comparator)
	{} /*TVectorQuickSortOf(TEtalonComparatorOf<T>&)*/

	//------------------------------------------------------------------------
	void operator()(nl_long_t nLeft, nl_long_t nRight);

protected:
	//------------------------------------------------------------------------
	TVectorOf<T>* pVector;
	TAutoRestoreEtalonComparatorHandleOf<T> hEtalonRestore;
	TEtalonComparatorOf<T>* pComparator;
} /*TVectorQuickSortOf<T>*/;

// TVectorVirtualQuickSort template
//============================================================================
template <class T>
class _NATLIB_TEMPLATE_DECLSPEC_ TVectorVirtualQuickSortOf {
public:
	//------------------------------------------------------------------------
	TVectorVirtualQuickSortOf(
		TAllocationEnvelopeOf<T>* a_Memory, size_t nSize,
		TEtalonComparatorOf<T>&
	) /*TVectorVirtualQuickSortOf*/;
	//------------------------------------------------------------------------
	~TVectorVirtualQuickSortOf();

	//------------------------------------------------------------------------
	void operator()(nl_long_t nLeft, nl_long_t nRight);

protected:
	//------------------------------------------------------------------------
	TAllocationEnvelopeOf<T>* a_Memory;
	size_t nSize;
	TAutoRestoreEtalonComparatorHandleOf<T> hEtalonRestore;
	TEtalonComparatorOf<T>* pComparator;
	size_t* a_nIndex;
	size_t* a_nPermutationIndex;

	//------------------------------------------------------------------------
	void Sort(nl_long_t nLeft, nl_long_t nRight);
	//------------------------------------------------------------------------
	void ConstructElement(
		TAllocationEnvelopeOf<T>* pTarget,
		const TAllocationEnvelopeOf<T>* pc_Source
	){
		new (pTarget++) TAllocationEnvelopeOf<T>(*pc_Source);
	} /*ConstructElement(TAllocationEnvelopeOf<T>*, ...)*/;
	//------------------------------------------------------------------------
	void DestructElement(TAllocationEnvelopeOf<T>* pTarget)
	{
    	typedef TAllocationEnvelopeOf<T> _TAllocationEnvelope;
		pTarget->~_TAllocationEnvelope();
	} /*DestructElement(TAllocationEnvelopeOf<T>*)*/
	//------------------------------------------------------------------------
	void ReconstructElement(
			TAllocationEnvelopeOf<T>* pTarget,
			const TAllocationEnvelopeOf<T>* pc_Source
	){
		DestructElement(pTarget);
		ConstructElement(pTarget, pc_Source);
	} /*ReconstructElement(TAllocationEnvelopeOf<T>*, ...)*/;
} /*TVectorVirtualQuickSortOf<T>*/;

// TVectorOf template implementation
//============================================================================
template <class T> TVectorOf<T>::TVectorOf(

	const TAllocator& c_Allocator /*= NLVECTOR_ALLOCATOR_DEFAULT()*/

):
	a_Memory(null), nSize(0), nCapacity(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Allocator);
	pAllocator->SetAllocationSize(0);
} /*TVectorOf<T>(const TAllocator&)*/

//----------------------------------------------------------------------------
template <class T> TVectorOf<T>::TVectorOf(

	const T& c_Filler,
	size_t nInitSize /*= 1*/,
	size_t nInitCapacity /*= 1*/,
	const TAllocator& c_Allocator /*= NLVECTOR_ALLOCATOR_DEFAULT()*/

):
	a_Memory(null), nSize(0), nCapacity(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Allocator);
	pAllocator->SetAllocationSize(0);

	Resize(nInitSize, nInitCapacity, c_Filler, true);
} /*TVectorOf<T>(const T&, size_t, size_t, const TAllocator&)*/

//----------------------------------------------------------------------------
template <class T> TVectorOf<T>::TVectorOf(

	const TVectorOf<T>& c_Source

):
	a_Memory(null), nSize(0), nCapacity(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Source.AllocatorConst());
	pAllocator->SetAllocationSize(0);

	Let(c_Source);
} /*TVectorOf<T>(const TVectorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> TVectorOf<T>::TVectorOf(

	const TVectorOf<T>& c_Source,
	const TAllocator& c_Allocator

):
	a_Memory(null), nSize(0), nCapacity(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Allocator);
	pAllocator->SetAllocationSize(0);

	Let(c_Source);
} /*TVectorOf<T>(const TVectorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> TVectorOf<T>::TVectorOf(

	const TContainerOf<T>& c_Source,
	const TAllocator& c_Allocator /*= NLVECTOR_ALLOCATOR_DEFAULT()*/

):
	a_Memory(null), nSize(0), nCapacity(0), pAllocator(null)
{
	pAllocator = CloneObject(c_Allocator);
	pAllocator->SetAllocationSize(0);

	Let(c_Source);
} /*TVectorOf<T>::TVectorOf(const TContainerOf<T>&, const TAllocator&)*/

//----------------------------------------------------------------------------
template <class T> TVectorOf<T>::~TVectorOf()
{
	ReduceSize(0, 0, true);
	delete pAllocator; pAllocator = null;
} /*TVectorOf<T>::~TVectorOf()*/

//----------------------------------------------------------------------------
template <class T> NLOBJECT_IMPLEMENT_CLONE_DEFAULT1(TVectorOf, T)

//----------------------------------------------------------------------------
template <class T> const T* TVectorOf<T>::SearchByKeyConst(

	const size_t& c_uIndex
	
) const {
	return (c_uIndex < nSize) ? &ElementConst(c_uIndex) : null;
} /*TVectorOf<T>::SearchByKeyConst(const size_t&) const*/

//----------------------------------------------------------------------------
template <class T> T* TVectorOf<T>::SearchByKey(

	const size_t& c_uIndex

){
	return (c_uIndex < nSize) ? &Element(c_uIndex) : null;
} /*TVectorOf<T>::SearchByKey(const size_t&)*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::SearchIndex(

	const TConditionOf<T>& c_Condition,
	nl_ulong_t nSearchCount

) const {
	if (nSearchCount == indexAny) nSearchCount = indexFirst;
	for (size_t i = 0; i < nSize; ++i)
		if (c_Condition(a_Memory[i].Data) && !--nSearchCount) return i;
	return NLVECTOR_NPOS;
} /*TVectorOf<T>::SearchIndex(const TConditionOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::SearchIndex(

	const TComparatorOf<T>& c_Comparator,
	nl_ulong_t nSearchCount

) const {
	TIsEqualByComparatorOf<T> IsEqual(c_Comparator);
	return SearchIndex(IsEqual, nSearchCount);
} /*TVectorOf<T>::SearchIndex(const TComparatorf<T>&)*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::BinarySearchIndex(

	const TComparatorOf<T>& c_Comparator

) const {
	if (!nSize) return NLVECTOR_NPOS;

	size_t uLeft = 0;
	size_t uRight = nSize - 1;
	while (uLeft < uRight) {
		size_t uMiddle = (uLeft + uRight) / 2;
		if (
			c_Comparator(a_Memory[uMiddle].Data) == TComparator::compGreater
		) uLeft = uMiddle + 1;
		else uRight = uMiddle;
	} /*while*/
	return (
		(c_Comparator(a_Memory[uLeft].Data) == TComparator::compEqual) ?
			uLeft : NLVECTOR_NPOS
	) /*return*/;
} /*TVectorOf<T>::BinarySearchIndex(const TComparatorOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::FindIndex(

	const TConditionOf<T>& c_Condition
	
) const {
	size_t nFound = SearchIndex(c_Condition);
	if (nFound == NLVECTOR_NPOS)
		throw TXNotFound(THIS_CLASS);
	return nFound;
} /*TVectorOf<T>::FindIndex(const TConditionOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::FindIndex(

	const TComparatorOf<T>& c_Comparator
	
) const {
	size_t nFound = SearchIndex(c_Comparator);
	if (nFound == NLVECTOR_NPOS)
		throw TXNotFound(THIS_CLASS);
	return nFound;
} /*TVectorOf<T>::FindIndex(const TComparatorOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::BinaryFindIndex(

	const TComparatorOf<T>& c_Comparator

) const {
	size_t nFound = BinarySearchIndex(c_Comparator);
	if (nFound == NLVECTOR_NPOS) 
		throw TXNotFound(THIS_CLASS);
	return nFound;
} /*TVectorOf<T>::BinaryFindIndex(const TComparatorOf<T>&) const*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::FirstKey() const
{
	if (!nSize) 
		throw TXNotFound(THIS_CLASS);
	return size_t(0);
} /*TVectorOf<T>::FirstKey() const*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::LastKey() const
{
	if (!nSize) 
		throw TXNotFound(THIS_CLASS);
	return size_t(nSize - 1);
} /*TVectorOf<T>::LastKey() const*/

//----------------------------------------------------------------------------
template <class T> T& TVectorOf<T>::Assign(

	const T& c_Data,
	const size_t& c_uIndex

){
	NLVECTOR_INDEX_VALIDITY_GUARD(c_uIndex);
	ReconstructElements(&(a_Memory[c_uIndex]), c_Data, 1);
	return a_Memory[c_uIndex].Data;
} /*TVectorOf<T>::Assign(const T&, const size_t&)*/

//----------------------------------------------------------------------------
template <class T> bool TVectorOf<T>::Unassign(

	const size_t& /*c_uIndex*/

){
	throw TXOperationIsNotAllowed(THIS_CLASS); return false;
} /*TVectorOf<T>::Unassign(const size_t&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Inorder(

	TStorageOperationOn<T, size_t>& r_Operation

){
	try {
		for (size_t i = 0; i < nSize; ++i)
			r_Operation(a_Memory[i].Data, i);
	} /*try*/
	catch (const TOperation::TXAbort&) {}
} /*TVectorOf<T>::Inorder(TStorageOperation&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Inorder(

	TStorageConstantOperationOn<T, size_t>& r_Operation

) const {
	try {
		for (size_t i = 0; i < nSize; ++i)
			r_Operation(const_cast<const T&>(a_Memory[i].Data), i);
	} /*try*/
	catch (const TOperation::TXAbort&) {}
} /*TVectorOf<T>::Inorder(TStorageConstantOperation&) const*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Postorder(

	TStorageOperationOn<T, size_t>& r_Operation

){
	try {
		size_t i = nSize;
		while (i) r_Operation(a_Memory[i].Data, --i);
	} /*try*/
	catch (const TOperation::TXAbort&) {}
} /*TVectorOf<T>::Postorder(TStorageOperation&)*/;

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Postorder(

	TStorageConstantOperationOn<T, size_t>& r_Operation

) const {
	try {
		size_t i = nSize;
		while (i)
			r_Operation(const_cast<const T&>(a_Memory[i].Data), --i);
	} /*try*/
	catch (const TOperation::TXAbort&) {}
} /*TVectorOf<T>::Postorder(TStorageConstantOperation&) const*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::QuickSort(

	TEtalonComparatorOf<T>& r_SortComparator
	
){
	if (!nSize) return;
	QuickSort(r_SortComparator, 0, nSize - 1);
} /*TVectorOf<T>::QuickSort(TEtalonComparatorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::QuickSort(

	TEtalonComparatorOf<T>& r_SortComparator,
	size_t nLeft,
	size_t nRight

){
	try {
		QuickSortVirtual(r_SortComparator, nLeft, nRight);
	} /*try*/
	catch (TXLowMemory& /*c_XLowMemory*/) {
		QuickSortDirect(r_SortComparator, nLeft, nRight);
	} /*catch (TXLowMemory&)*/
} /*TVectorOf<T>::QuickSort(TEtalonComparatorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::QuickSortDirect(

	TEtalonComparatorOf<T>& r_SortComparator,
	size_t nLeft,
	size_t nRight

){
	NLVECTOR_INDEX_VALIDITY_GUARD(nLeft);
	NLVECTOR_INDEX_VALIDITY_GUARD(nRight);
	if (nSize) {
		TVectorQuickSortOf<T> Sort(this, r_SortComparator);
		if (nLeft < nRight) Sort(nLeft, nRight);
		else Sort(nRight, nLeft);
	} /*if*/
} /*TVectorOf<T>::QuickSortDirect(TEtalonComparatorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::QuickSortVirtual(

	TEtalonComparatorOf<T>& r_SortComparator,
	size_t nLeft,
	size_t nRight

){
	NLVECTOR_INDEX_VALIDITY_GUARD(nLeft);
	NLVECTOR_INDEX_VALIDITY_GUARD(nRight);

	if (nSize) {
		if (nRight < nLeft) {
			register size_t nTemp = nLeft; nLeft = nRight, nRight = nTemp;
		} /*if*/

		TVectorVirtualQuickSortOf<T> Sort(
			a_Memory + nLeft, nRight - nLeft + 1, r_SortComparator
		) /*Sort*/;
		Sort(0, nRight - nLeft);
	} /*if*/
} /*TVectorOf<T>::QuickSortVirtual(TEtalonComparatorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Resize(

	size_t nNewSize,
	size_t nNewCapacity,
	const T& c_Filler,
	bool bUnconditional

){
	// Reserve memory needed
	//------------------------
	if (nNewSize > nNewCapacity) nNewCapacity = nNewSize;
	
	// Construct new members
	//-----------------------
	if (nNewSize < nSize) {
		ReduceSize(nNewSize, nNewCapacity, bUnconditional);
	}
	else if (nNewSize > nSize) {
		Reserve(nNewCapacity, bUnconditional);
		ConstructElements(a_Memory + nSize, c_Filler, nNewSize - nSize);
		nSize = nNewSize;
	} /*iff*/
} /*TVectorOf<T>::Resize(size_t, size_t, const T&, bool)*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::ReduceSize(

	size_t nNewSize,
	bool bUnconditional /*= false*/

){
	return ReduceSize(nNewSize, nCapacity, bUnconditional);
} /*TArayOf<T>::ReduceSize(size_t, bool)*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::ReduceSize(

	size_t nNewSize,
	size_t nNewCapacity,
	bool bUnconditional

){
	// Reduce size
	//-------------
	if (nSize < nNewSize) nNewSize = nSize;
	DestructElements(a_Memory + nNewSize, nSize - nNewSize);
	nSize = nNewSize;

	// Change capacity
	//-----------------
	if (nNewCapacity < nSize) nNewCapacity = nSize;
	Reserve(nNewCapacity, bUnconditional);

	return nSize;
} /*TArayOf<T>::ReduceSize(size_t, bool)*/

//----------------------------------------------------------------------------
template <class T> size_t TVectorOf<T>::Reserve(

	size_t nNewCapacity,
	bool bUnconditional /*= true*/

){
	if (nNewCapacity < nSize) nNewCapacity = nSize;
	if (!bUnconditional && (nNewCapacity < nCapacity))
		nNewCapacity = nCapacity;

	if (!nNewCapacity) {
		DeallocateMemory(a_Memory);
		nSize = nCapacity = 0; a_Memory = null;
	}
	else if (nNewCapacity != nCapacity) {
		TAllocationEnvelopeOf<T>* a_NewMemory = (
			AllocateMemory(nNewCapacity)
		) /*a_NewMemory*/;

		// Copy elements
		//---------------
		ConstructElements(a_NewMemory, a_Memory, nSize);

		// Save memory attributes to delete
		//----------------------------------
		TAllocationEnvelopeOf<T>* a_MemoryToDelete = a_Memory;

		// New memory attributes
		//-----------------------
		a_Memory = a_NewMemory; nCapacity = nNewCapacity;

		DestructElements(a_MemoryToDelete, nSize);
		DeallocateMemory(a_MemoryToDelete);
	} /*iff*/

	return nCapacity;
} /*TVectorOf<T>::Reserve(size_t, bool)*/

//----------------------------------------------------------------------------
template <class T> const T& TVectorOf<T>::Insert(

	const T& c_Data

){
	size_t nNewCapacity = nCapacity; {
		if (nSize >= nNewCapacity) {
			nNewCapacity += (nSize / 2 + 1);
			if (nNewCapacity < 16) nNewCapacity = 16;
		} /*if*/
	} /*nNewCapacity*/;

	Reserve(nNewCapacity);
	Resize(nSize + 1, nNewCapacity, c_Data, false);

	return Element(nSize - 1);
} /*TVectorOf<T>::Insert(const T&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Insert(

	const TVectorOf<T>& c_Source

){
	ReconstructElements(
		a_Memory, c_Source.a_Memory, min(c_Source.Size(), Size())
	) /*ReconstructElements*/;

	if (c_Source.Size() > Size()) 
		throw TXOutOfRange(THIS_CLASS);
} /*TVectorOf<T>::Insert(const TVectorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> nl_ulong_t TVectorOf<T>::Delete(

	const TConditionOf<T>& c_Condition,
	nl_ulong_t nIndex /*= indexAll*/

){
	TSize nDeleted = 0;

	if (nIndex != indexAll) {
		size_t nKey;
		try {nKey = FindKey(c_Condition, nIndex);}
		catch (TXNotFound&) {return 0;}
		TAllocationEnvelopeOf<T>* pPlacement = a_Memory + nKey;
		ReconstructElements(pPlacement, pPlacement + 1, nSize - nKey - 1);
		nDeleted = 1;
	}
	else {
		for (size_t i = 0; i < nSize; ++i) {
			if (c_Condition(RawElement(i))) {
				TAllocationEnvelopeOf<T>* pPlacement = a_Memory + i;
				ReconstructElements(
					pPlacement, pPlacement + 1, nSize - i - 1
				) /*ReconstructElements*/;
				++nDeleted;
			} /*if*/
		} /*for (i)*/
	} /*iff*/

	size_t nNewSize = nSize - nDeleted;
	size_t nNewCapacity = nCapacity;
#ifdef NEVER
	if (nNewSize < (nCapacity / 3)) {
		(nNewCapacity /= 2) += 1;
		if ((nNewCapacity < 16) && (nCapacity >= 16)) nNewCapacity = 16;
	} /*if*/
#endif /*NEVER*/
	ReduceSize(nNewSize, nNewCapacity, false);

	return nDeleted;
} /*TVectorOf<T>::Delete(const TConditionOf<T>&, TIndex = indexAll)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Let(

	const TContainerOf<T>& c_Container
	
){
	Reserve(c_Container.Size(), false);
	TParent::Let(c_Container);
} /*TVectorOf<T>::Let(const TContaineOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Let(

	const TVectorOf<T>& c_Source

){
	ReduceSize(0, 0, true); // Delete all

	// Allocate new memory
	//---------------------
	TAllocationEnvelopeOf<T>* a_NewMemory = (
		AllocateMemory(c_Source.Size())
	) /*a_NewMemory*/;

	// Copy elements
	//---------------
	ConstructElements(a_NewMemory, c_Source.a_Memory, c_Source.Size());

	a_Memory = a_NewMemory;
	nSize = nCapacity = c_Source.Size();
} /*TVectorOf<T>::Let(const TVectorOf<T>&)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::Swap(

	size_t i,
	size_t j

){
	if (!((i < nSize) && (j < nSize))) 
		throw TXOutOfRange(THIS_CLASS);
	if (i == j) return;

	TAllocationEnvelopeOf<T> Temp(a_Memory[i]);
	ReconstructElements(a_Memory + i, a_Memory + j, 1);
	ReconstructElements(a_Memory + j, &Temp, 1);
} /*TVectorOf<T>::Swap(size_t, size_t)*/

//----------------------------------------------------------------------------
template <class T> const TAllocator& TVectorOf<T>::AllocatorConst() const 
{	
	if (!pAllocator)
		throw TXAllocatorIsNotDefined(THIS_CLASS);
	return *pAllocator;
} /*TVectorOf<T>::AllocatorConst() const */

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::ConstructElements(

	TAllocationEnvelopeOf<T>* pTarget,
	const T& c_Default,
	size_t nElementsToConstruct

){
	if (!pTarget) return;

	TAllocationEnvelopeOf<T>* pEnd = pTarget + nElementsToConstruct;
	while (pTarget != pEnd)
		new (pTarget++) TAllocationEnvelopeOf<T>(c_Default);
} /*TVectorOf<T>:ConstructElements(TAllocationEnvelopeOf<T>*, size_t)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::ConstructElements(

	TAllocationEnvelopeOf<T>* pTarget,
	const TAllocationEnvelopeOf<T>* pc_Source,
	size_t nElementsToConstruct

){
	if (!(pTarget && pc_Source)) return;

	TAllocationEnvelopeOf<T>* pEnd = pTarget + nElementsToConstruct;
	while (pTarget != pEnd)
		new (pTarget++) TAllocationEnvelopeOf<T>(*(pc_Source++));
} /*TVectorOf<T>::ConstructElements(TAllocationEnvelopeOf<T>*, ...)*/;

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::DestructElements(

	TAllocationEnvelopeOf<T>* pTarget,
	size_t nElementsToDestruct

){
	if (!pTarget) return;

    typedef TAllocationEnvelopeOf<T> _TAllocationEnvelope;
	_TAllocationEnvelope* pEnd = pTarget + nElementsToDestruct;
	while (pTarget != pEnd) (pTarget++)->~_TAllocationEnvelope();
} /*TVectorOf<T>::DestructElements(TAllocationEnvelopeOf<T>*, size_t)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::ReconstructElements(

	TAllocationEnvelopeOf<T>* pTarget,
	const T& c_Default,
	size_t nElementsToReconstruct

){
	if (!pTarget) return;
	DestructElements(pTarget, nElementsToReconstruct);
	ConstructElements(pTarget, c_Default, nElementsToReconstruct);
} /*TVectorOf<T>::ReconstructElements(TAllocationEnvelopeOf<T>*, ...)*/

//----------------------------------------------------------------------------
template <class T> void TVectorOf<T>::ReconstructElements(

	TAllocationEnvelopeOf<T>* pTarget,
	const TAllocationEnvelopeOf<T>* pc_Source,
	size_t nElementsToReconstruct
	
){
	if (!(pTarget && pc_Source)) return;
	DestructElements(pTarget, nElementsToReconstruct);
	ConstructElements(pTarget, pc_Source, nElementsToReconstruct);
} /*TVectorOf<T>::ReconstructElements(TAllocationEnvelopeOf<T>*, ...)*/
	
//----------------------------------------------------------------------------
template <class T> TAllocationEnvelopeOf<T>* TVectorOf<T>::AllocateMemory(

	size_t nElements

){
	if (nElements > NLVECTOR_MAX_SIZE(T)) 
		throw TXOutOfRange(THIS_PLACE);

	size_t uToAllocate = nElements * NLVECTOR_ELEMENT_SIZE(T);
	return (
		reinterpret_cast<TAllocationEnvelopeOf<T>*>(
			pAllocator->Allocate(uToAllocate)
		) /*reinterpret_cast<TAllocationEnvelopeOf<T>*>*/
	) /*return*/;
} /*TVectorOf<T>::AllocateMemory(size_t)*/

// TVectorQuickSortOf template implementation
//============================================================================
template <class T> void TVectorQuickSortOf<T>::operator()(

	nl_long_t nLeft,
	nl_long_t nRight

){
	while (nLeft < nRight) {
		nl_long_t i = nLeft;
		nl_long_t j = nRight;

		// Partition
		//-----------
		{
			T Median(pVector->RawElement((nLeft + nRight) / 2));
			pComparator->SetEtalon(Median);
			while (i <= j) {
				while (
					(*pComparator)(pVector->RawElement(i)) ==
						TComparator::compGreater
				) /*while*/ ++i;
				while (
					(*pComparator)(pVector->RawElement(j)) ==
						TComparator::compLess
				) /*while*/ --j;
				if (i <= j) {pVector->Swap(i, j); ++i; --j;}
			} /*while*/
		} /*Partition*/

		if (nLeft >= j) nLeft = i;
		else if (i >= nRight) nRight = j;
		else if ((j - nLeft) <= (nRight - i)) {
			operator()(nLeft, j); nLeft = i;
		}
		else {
			operator()(i, nRight); nRight = j;
		} /*iff*/
	} /*while*/
} /*TVectorQuickSortOf<T>::operator()(size_t, size_t)*/

// TVectorVirtualQuickSortOf template implementation
//============================================================================
template <class T> TVectorVirtualQuickSortOf<T>::TVectorVirtualQuickSortOf(

	TAllocationEnvelopeOf<T>* a_InitMemory,
	size_t nInitSize,
	TEtalonComparatorOf<T>& r_Comparator

):
	a_Memory(a_InitMemory), nSize(nInitSize),
	hEtalonRestore(r_Comparator), pComparator(&r_Comparator),
	a_nIndex(null), a_nPermutationIndex(null)
{
	try {
		// Allocator is not used because instrumental 
		// data is allocated here !!!
		//--------------------------------------------
		NLEXCEPT_CLASS_NEW(size_t[nSize], a_nIndex);
		NLEXCEPT_CLASS_NEW(size_t[nSize], a_nPermutationIndex);
	} /*try*/
	catch (...) {
		try {delete[] a_nIndex;} catch (...) {} 
		a_nIndex = null;
		try {delete[] a_nPermutationIndex;} catch (...) {} 
		a_nPermutationIndex = null;
		throw TXLowMemory(THIS_CLASS);
	} /*catch (...)*/

	for (size_t i = 0; i < nSize; ++i) a_nIndex[i] = i;
} /*TVectorVirtualQuickSortOf<T>(TVectorOf<T>, ...)*/

//----------------------------------------------------------------------------
template <class T> TVectorVirtualQuickSortOf<T>::~TVectorVirtualQuickSortOf()
{
	try {delete[] a_nIndex; a_nIndex = null;}
	catch (...) {
		a_nIndex = null;
		try {delete[] a_nPermutationIndex; a_nPermutationIndex = null;}
		catch (...) {a_nPermutationIndex = null; throw;}
		throw;
	} /*catch (...)*/
	try {delete[] a_nPermutationIndex; a_nPermutationIndex = null;}
	catch (...) {a_nPermutationIndex = null; throw;}
} /*~TVectorVirtualQuickSortOf()*/

//----------------------------------------------------------------------------
template <class T> void TVectorVirtualQuickSortOf<T>::operator()(

	nl_long_t nLeft,
	nl_long_t nRight

){
	Sort(nLeft, nRight);

	// Prepare permutation index
	//---------------------------
	{
		for (size_t i = 0; i < nSize; ++i)
			a_nPermutationIndex[a_nIndex[i]] = i;
	} /*Prepare permutation index*/

	// Change the order in the vector
	//--------------------------------
	{
		for (size_t i = 0; i < nSize; ++i) {
			if (i != a_nIndex[i]) {
				size_t nTarget = i;
				TAllocationEnvelopeOf<T> Temp(a_Memory[nTarget].Data);
				size_t nTempIndex = a_nPermutationIndex[i];
				while (nTarget != nTempIndex) {
					size_t nSource = a_nIndex[nTarget];
					ReconstructElement(
						a_Memory + nTarget, a_Memory + nSource
					) /*ReconstructElement*/;
					a_nIndex[nTarget] = nTarget;
					a_nPermutationIndex[nTarget] = nTarget;
					nTarget = nSource;
				} /*while*/
				ReconstructElement(a_Memory + nTempIndex, &Temp);
				a_nIndex[nTempIndex] = nTempIndex;
			} /*if*/
		} /*for (i)*/
	} /*Change the order in the vector*/
} /*TVectorVirtualQuickSortOf<T>::operator()*/

//============================================================================
template <class T> void TVectorVirtualQuickSortOf<T>::Sort(

	nl_long_t nLeft,
	nl_long_t nRight

){
	while (nLeft < nRight) {
		nl_long_t i = nLeft, j = nRight;

		// Partition
		//-----------
		{
			T Median(a_Memory[a_nIndex[(nLeft + nRight) / 2]].Data);
			pComparator->SetEtalon(Median);
			while (i <= j) {
				while (
					(*pComparator)(a_Memory[a_nIndex[i]].Data) ==
						TComparator::compGreater
				) /*while*/ ++i;
				while (
					(*pComparator)(a_Memory[a_nIndex[j]].Data) ==
						TComparator::compLess
				) /*while*/ --j;
				if (i <= j) {
					size_t nTemp = a_nIndex[i];
					a_nIndex[i] = a_nIndex[j]; a_nIndex[j] = nTemp;
					++i; --j;
				} /*if*/
			} /*while*/
		} /*Partition*/

		if (nLeft >= j) nLeft = i;
		else if (i >= nRight) nRight = j;
		else if ((j - nLeft) <= (nRight - i)) {Sort(nLeft, j); nLeft = i;}
		else {Sort(i, nRight); nRight = j;}
	} /*while*/
} /*TVectorVirtualQuickSortOf<T>::Sort()(size_t, size_t)*/

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
#define NLVECTOR_DECLSPEC(T, Declspec) \
	NLSTORAG_DECLSPEC(T, size_t, Declspec); \
	NLALLOC_DECLSPEC(T, Declspec); \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TVectorVirtualQuickSortOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TVectorQuickSortOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/; \
	NATLIB_EXPAND_TEMPLATE1( \
		NATLIB_NAMESPACE_MEMBER(TVectorOf), T, Declspec \
	) /*NATLIB_EXPAND_TEMPLATE1*/

#endif /*!defined(__NATLIB_NLVECTOR_H__)*/
