#ifndef __NATLIB_TASKS_NTASKS_H__
#define __NATLIB_TASKS_NTASKS_H__
/****************************************************************************/
/** ntasks.h                                                               **/
/** ---------------------------------------------------------------------- **/
/** Tasks primitives                                                       **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** September 01, 1997 by Oleg Kagan                                       **/
/****************************************************************************/

// Natlib Basics includes
//============================================================================
#include <nlexcept.h>
#include <nlalloc.h>
#include <nlsafeh.h>
#include <nlvector.h>

// Natlib Tasks includes
//============================================================================
#include <ntsfbuf.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// prototypes
//============================================================================
class _NATLIB_TASKS_DECLSPEC_ TTask;
class _NATLIB_TASKS_DECLSPEC_ TSignal;
class _NATLIB_TASKS_DECLSPEC_ TSchedulerInterface;

//============================================================================
class _NATLIB_TASKS_DECLSPEC_ TTask :
	public virtual TNatlibObject,
    public virtual TAllocationAware,
	public virtual TSafeHandleAllowed,
	private TSafeBuffer::TRecord
{
public:
	//------------------------------------------------------------------------
    friend TSchedulerInterface;

	//------------------------------------------------------------------------
    typedef TNatlibObject TParent;
	typedef TSafeBuffer::TRecord TBufferRecord;

	//------------------------------------------------------------------------
    typedef TSafeHandleOf<TTask> THandle;

	// Phases of task's life
	//------------------------------------------------------------------------
	typedef enum  {
    	phaseFirst,
		phaseNone = phaseFirst,
		phaseStartup,
		phaseTransaction,
        phaseTerminate,
		phaseCanTerminate,
        phaseException
	} TPhase;

    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
    	TException, TNatlibObject::TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE(
    	TXInvalidSchedulerHandle, TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
    NLEXCEPT_DECLARE_INVALID_PARAMETER(
    	TXInvalidParameter, TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE_INVALID_PARAMETER*/;

	//------------------------------------------------------------------------
    TTask(const TStdString& c_sName);
    //------------------------------------------------------------------------
    virtual ~TTask();

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED(TTask);

	//------------------------------------------------------------------------
	virtual const TStdString& Name() const {return sName;}

    //------------------------------------------------------------------------
    bool IsAutodelete() const 
	{
		return TBufferRecord::IsAutodelete();
	} /*IsAutodelete() const*/

	// Methods that called by scheduler 
	//------------------------------------------------------------------------
protected:
	// Called by scheduler on task's startup
	//------------------------------------------------------------------------
    virtual void EvStartup();
	// Called to notify about the Signal 
	// or to perform the background processing (parameter is null)
    //------------------------------------------------------------------------
    virtual void EvTransaction(const TSignal*);
	// Called right before the termination of the task
    //------------------------------------------------------------------------
    virtual void EvTerminate();
	// Called to check if the task can be terminated safe (true by default)
    //------------------------------------------------------------------------
    virtual bool EvCanTerminate();
	// Called to inform task about exception at specified phase
    //------------------------------------------------------------------------
    virtual void EvException(TPhase, const TNatlibException*);

	// Services to use from inside of child classes
	//------------------------------------------------------------------------
protected:
	// Send the specified signal to the specified process 
	// (THandle::c_hNull means all processes)
    //------------------------------------------------------------------------
    void Send(TSignal&, const THandle& c_hTarget);
	//------------------------------------------------------------------------
	void Send(TSignal&, const TContainerOf<THandle>& c_Targets);
    // Sleep until the next signal arrival
	// (Request scheduler to stop calling the EvTransaction(null))
	//------------------------------------------------------------------------
    void Wait();
	// Starts the child task
	//------------------------------------------------------------------------
	void Start(TTask&, bool bAutodelete = true);
    // Terminates the specified process
    // (if handle specified is null terminate itself)
    //------------------------------------------------------------------------
    void Terminate(const THandle&, bool bUnconditional = false);
    // Terminate itself
    //------------------------------------------------------------------------
    void Terminate(bool bUnconditional = false);
	//------------------------------------------------------------------------
	void GlobalTerminate();

    //------------------------------------------------------------------------
    void ShowException(TPhase);
    //------------------------------------------------------------------------
    void ShowException(const TNatlibException&, TPhase);

	//------------------------------------------------------------------------
	TSchedulerInterface& Scheduler();

private:
	//------------------------------------------------------------------------
    TStdString sName;
    TSafeHandleBase hScheduler;
} /*TTask*/;

// TTaskHandle
//============================================================================
typedef TTask::THandle TTaskHandle;

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
NLSAFEH_DECLSPEC(NATLIB_NAMESPACE_MEMBER(TTask), _NATLIB_TASKS_DECLSPEC_);
typedef NATLIB_NAMESPACE_MEMBER(TTask::THandle) NTASKS_TTaskHandle;
NLVECTOR_DECLSPEC(NTASKS_TTaskHandle, _NATLIB_TASKS_DECLSPEC_);
NLVECTOR_DECLSPEC(NTASKS_TTaskHandle, _NATLIB_TASKS_DECLSPEC_);

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TSignal
//============================================================================
class _NATLIB_TASKS_DECLSPEC_ TSignal :
	public virtual TNatlibObject,
    public virtual TAllocationAware,
    private TSafeBuffer::TRecord
{
public:
	//------------------------------------------------------------------------
	typedef TNatlibObject TParent;
	typedef TSafeBuffer::TRecord TBufferRecord;
    friend TSchedulerInterface;

    //------------------------------------------------------------------------
	TSignal();
	//------------------------------------------------------------------------
	virtual ~TSignal();

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED(TSignal)
    
	// Unregisters signal from signal queue
    //------------------------------------------------------------------------
    void Unregister() {TBufferRecord::Unregister();}

private:
	//------------------------------------------------------------------------
	TVectorOf<TTask::THandle> Destination;
	TSafeHandleBase hScheduler;
} /*TSignal*/;

// TSchedulerInterface
//============================================================================
class _NATLIB_TASKS_DECLSPEC_ TSchedulerInterface :
	public virtual TNatlibObject,
    public virtual TAllocationAware,
	public virtual TSafeHandleAllowed
{
public:
	//------------------------------------------------------------------------
    typedef TNatlibObject TParent;

	//------------------------------------------------------------------------
    typedef TSafeHandleOf<TSchedulerInterface> THandle;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
    	TException, TNatlibObject::TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
    	TXBadSchedulerChain, TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;

    // Servise exception
    // (used to request immediate service from running scheduler)
    //------------------------------------------------------------------------
    NLEXCEPT_DECLARE_BASE(
    	TXServiceRequest, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE_BASE*/;
	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
    	TXWaitRequest, TXServiceRequest, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
    	TXEndTransactionRequest, TXServiceRequest, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE_DATA_HOLDER(
    	TXTerminateRequest, bool, TXServiceRequest, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE_DATA_HOLDER*/;
	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
    	TXGlobalTerminateRequest, TXServiceRequest, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;

	//------------------------------------------------------------------------
    TSchedulerInterface();
    //------------------------------------------------------------------------
    virtual ~TSchedulerInterface();

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED(TSchedulerInterface);

	//------------------------------------------------------------------------
    virtual void Start(TTask&, bool bAutodelete = true) = 0;
	// Start the task using dependent scheduler interface
    //------------------------------------------------------------------------
    virtual void Terminate(
		const TTask::THandle&, bool bUnconditional = false
	) /*Terminate(const TTask::THandle&, bool = false */ = 0;
    //------------------------------------------------------------------------
    virtual void Send(TSignal&, const TTask::THandle& c_hTarget) = 0;
    //------------------------------------------------------------------------
    virtual void Send(TSignal&, const TContainerOf<TTask::THandle>&) = 0;

	//------------------------------------------------------------------------
    virtual void ShowTaskException(
    	TTask&, TTask::TPhase, const TNatlibException* = null
    ) /*ShowTaskException(TTask&,TPhase,const TNatlibException* = null)*/ = 0;

protected:
	// Methods to use by dependent scheduler interfaces
	//------------------------------------------------------------------------
    virtual void InterfaceStart(
		TSchedulerInterface&, TTask&, bool bAutodelete = true
	) /*InterfaceStart(TSchedulerInterface&, TTask&, bool = true)*/ = 0;
    //------------------------------------------------------------------------
    virtual void InterfaceSend(
		TSchedulerInterface&, TSignal&, const TTask::THandle& c_hTarget
	) /*InterfaceSend(TSchedulerInterface&, TSignal&, ...)*/ = 0;
    //------------------------------------------------------------------------
    virtual void InterfaceSend(
		TSchedulerInterface&, TSignal&, const TContainerOf<TTask::THandle>&
	) /*InterfaceSend(TSchedulerInterface&, TSignal&, ...)*/ = 0;
	//------------------------------------------------------------------------
    virtual void InterfaceShowTaskException(
    	TSchedulerInterface&, 
		TTask&, TTask::TPhase, const TNatlibException* = null
    ) /*InterfaceShowTaskException(TSchedulerInterface&, TTask&, ...)*/ = 0;

	// methods to use in children
    //------------------------------------------------------------------------
    static bool IsTaskRegistered(
		const TTask& r_Task, const TSafeBuffer& c_Buffer
	){
    	return r_Task.IsRegistered(c_Buffer);
    } /*IsTaskRegistered(const TTask&, const TSafeBuffer&)*/
    //------------------------------------------------------------------------
    static bool IsTaskRegistered(const TTask& r_Task)
	{
    	return r_Task.IsRegistered();
    } /*IsTaskRegistered(const TTask&)*/

    //------------------------------------------------------------------------
    static TVectorOf<TTask::THandle>& GetSignalDestination(TSignal& r_Signal)
    {
    	return r_Signal.Destination;
    } /*GetSignalDestination(TSignal&)*/
	//------------------------------------------------------------------------
	static void SetSignalDestination(
		TSignal& r_Signal, const TTask::THandle& c_hTarget
	){
		r_Signal.Destination.Resize(1, c_hTarget);
	} /*SetSignalDestination(TSignal&, const TTask::THandle&)*/
	//------------------------------------------------------------------------
	static void SetSignalDestination(
		TSignal& r_Signal, const TContainerOf<TTask::THandle>& c_Targets
	){
		r_Signal.Destination.Let(c_Targets);
	} /*SetSignalDestination(TSignal&, const TContainerOf<TTask::THandle>&)*/;

    //------------------------------------------------------------------------
    static void CallTaskStartup(TTask& r_Task) {r_Task.EvStartup();}
    //------------------------------------------------------------------------
    static void CallTaskTransaction(TTask& r_Task, const TSignal* pc_Signal)
    {
    	r_Task.EvTransaction(pc_Signal);
    } /*CallTaskTransaction(TTask&, const TSignal*)*/
    //------------------------------------------------------------------------
    static void CallTaskTerminate(TTask& r_Task) {r_Task.EvTerminate();}
    //------------------------------------------------------------------------
    static bool CallTaskCanTerminate(TTask& r_Task)
    {
    	return r_Task.EvCanTerminate();
    } /*CallTaskCanTerminate(TTask&)*/
    //------------------------------------------------------------------------
    static void CallTaskException(
    	TTask& r_Task, TTask::TPhase Phase, TNatlibException* pException
    ){
    	r_Task.EvException(Phase, pException);
    } /*CallTaskException(TTask&, TTask::TPhase, TNatlibException*)*/

	//------------------------------------------------------------------------
	static TSchedulerInterface* TaskSchedulerPtr(TTask& r_Task)
	{
		if (!r_Task.hScheduler.IsTargetValid()) return null;
		return (
			&dynamic_cast<TSchedulerInterface&>(
				r_Task.hScheduler.Dereference()
			) /*dynamic_cast<TSchedulerInterface&>*/
		) /*return*/;
	} /*TaskSchedulerPtr(TTask&)*/

	//------------------------------------------------------------------------
	static void SetTaskScheduler(
		TTask& r_Task, const TSchedulerInterface::THandle& c_hScheduler
	){
		r_Task.hScheduler = c_hScheduler;
	} /*SetTaskScheduler(TTask&, const TSchedulerInterface::THandle&)*/

	//------------------------------------------------------------------------
	static void InsertTask(
    	TTask& r_Task, TSafeBuffer& r_Queue, bool bAutodelete
    ){
		r_Queue.Insert(
        	dynamic_cast<TSafeBuffer::TRecord&>(r_Task), bAutodelete
        ) /*r_Queue.Insert*/;
	} /*InsertTask(TTask&, TSafeBuffer&)*/
	//------------------------------------------------------------------------
	static void UnregisterTask(TTask& r_Task, TSafeBuffer& r_Queue)
	{
		r_Queue.Unregister(dynamic_cast<TSafeBuffer::TRecord&>(r_Task));
	} /*UnregisterTask(TTask&, TSafeBuffer&)*/
	//------------------------------------------------------------------------
	static void DeleteTask(TTask& r_Task, TSafeBuffer& r_Queue)
	{
		r_Queue.Delete(dynamic_cast<TSafeBuffer::TRecord&>(r_Task));
	} /*DeleteTask(TTask&, TSafeBuffer&)*/
	//------------------------------------------------------------------------
	static TTask& CurrentTask(TSafeBuffer& r_Queue)
	{
		return dynamic_cast<TTask&>(r_Queue.Current());
	} /*InsertTo(TSafeBuffer&, TTask&)*/

	//------------------------------------------------------------------------
	static void InsertSignal(TSignal& r_Signal, TSafeBuffer& r_Queue)
	{
		r_Queue.Insert(dynamic_cast<TSafeBuffer::TRecord&>(r_Signal));
	} /*InsertTo(TSignal&, TSafeBuffer&)*/
	//------------------------------------------------------------------------
	static TSignal& CurrentSignal(TSafeBuffer& r_Queue)
	{
		return dynamic_cast<TSignal&>(r_Queue.Current());
	} /*InsertTo(TSafeBuffer&, TSignal&)*/

protected:
	//------------------------------------------------------------------------
    TSafeHandleBase hParent;

	//------------------------------------------------------------------------
	virtual TSchedulerInterface* GetParentPtr();
	//------------------------------------------------------------------------
	virtual TSchedulerInterface* GetMasterPtr();
} /*TSchedulerInterface*/;

//============================================================================
NATLIB_NAMESPACE_END

//============================================================================
NLSAFEH_DECLSPEC(
	NATLIB_NAMESPACE_MEMBER(TSchedulerInterface), _NATLIB_TASKS_DECLSPEC_
) /*NLSAFEH_DECLSPEC*/;

#endif /*!defined(__NATLIB_TASKS_NTASKS_H__)*/
