#ifndef __NATLIB_TASKS_NTSFBUF_H__
#define __NATLIB_TASKS_NTSFBUF_H__
/****************************************************************************/
/** ntsfbuf.h                                                              **/
/** ---------------------------------------------------------------------- **/
/** Safe cyclic buffer (queue)                                             **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** August 18, 1997 by Oleg Kagan                                          **/
/****************************************************************************/

// Natlib Basics includes
//============================================================================
#include <nlexcept.h>
#include <nlalloc.h>
#include <nloperat.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TSafeBuffer
//============================================================================
class _NATLIB_TASKS_DECLSPEC_ TSafeBuffer : public virtual TNatlibObject {
public:
	//------------------------------------------------------------------------
    typedef TNatlibObject TParent;
    
	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(
    	TException, TNatlibObject::TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(TXBufferIsEmpty, TException, _NATLIB_TASKS_DECLSPEC_);
	NLEXCEPT_DECLARE(
    	TXRecordIsNotRegistered, TException, _NATLIB_TASKS_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
	NLEXCEPT_DECLARE(TXInvalidRecord, TException, _NATLIB_TASKS_DECLSPEC_);
    NLEXCEPT_DECLARE(TXInvalidChain, TException, _NATLIB_TASKS_DECLSPEC_);

	// TRecord
	//------------------------------------------------------------------------
	typedef class _NATLIB_TASKS_DECLSPEC_ TRecord : 
		public virtual TAllocationAware 
	{
	public:
		//--------------------------------------------------------------------
		NLEXCEPT_DECLARE_BASE(TException, _NATLIB_TASKS_DECLSPEC_);
		NLEXCEPT_DECLARE(
        	TXUnableToUnregister, TException, _NATLIB_TASKS_DECLSPEC_
        ) /*NLEXCEPT_DECLARE*/;

		// Construct unregistered record
		//--------------------------------------------------------------------
		TRecord();

		//--------------------------------------------------------------------
		virtual ~TRecord() {Unregister();}

		//--------------------------------------------------------------------
		bool IsRegistered() const {return pBuffer != null;}
		//--------------------------------------------------------------------
		bool IsRegistered(const TSafeBuffer& c_Buffer) const
		{
			return (pBuffer && (pBuffer == &c_Buffer));
		} /*IsRegistered(const TSafeBuffer&) const*/;
		//--------------------------------------------------------------------
		virtual void Unregister();

		//--------------------------------------------------------------------
		TSafeBuffer* BufferPtr() {return pBuffer;}

		//--------------------------------------------------------------------
        bool IsAutodelete() const {return bAutodelete;}

	private:
		//--------------------------------------------------------------------
		TSafeBuffer* pBuffer;
		bool bAutodelete;

		//--------------------------------------------------------------------
		TRecord* pNext;
		TRecord* pPrevious;
	
		//--------------------------------------------------------------------
		friend TSafeBuffer;
	} typedef_TRecord;

	//------------------------------------------------------------------------
	TSafeBuffer() : pCurrent(null), nRecords(0) {}
	//------------------------------------------------------------------------
	virtual ~TSafeBuffer() {Clear();}

    //------------------------------------------------------------------------
    void AssertValid() const;
    //------------------------------------------------------------------------
    TNatlibStdString Dump() const;

	// Insert already consructed record into the buffer before Current Record
	//------------------------------------------------------------------------
	virtual void Register(TRecord&, bool bAutodelete = true);
	//------------------------------------------------------------------------
	void Insert(TRecord& r_Record, bool bAutodelete = true)
	{
		Register(r_Record, bAutodelete);
	} /*Insert(TRecord, bool = true)*/

	// Unregisters specified record (without destruction)
	//------------------------------------------------------------------------
	virtual void Unregister(TRecord&);
	// Deletes specified record from the Buffer (destruct if needed)
	//------------------------------------------------------------------------
	void Delete(TRecord&);

	//------------------------------------------------------------------------
	virtual void MoveFrom(TSafeBuffer&);

	//------------------------------------------------------------------------
	virtual void Clear();

	//------------------------------------------------------------------------
	nl_ulong_t Size() const {return nRecords;}
	//------------------------------------------------------------------------
	bool IsEmpty() const {return nRecords == 0;}

	// Moves the Current Record Pointer one record forward
	//------------------------------------------------------------------------
	void Next() 
	{
		if (!pCurrent) 
			throw TXBufferIsEmpty(THIS_CLASS);
        pCurrent = pCurrent->pNext;
	} /*Next()*/
	// Moves Current Record Pointer one record backward
	//------------------------------------------------------------------------
	void Previous()
	{
		if (!pCurrent) 
			throw TXBufferIsEmpty(THIS_CLASS);
		pCurrent = pCurrent->pPrevious;
	} /*TSafeBuffer::Previous()*/

	//------------------------------------------------------------------------
	TRecord& Current()
	{
		if (!pCurrent) 
			throw TXBufferIsEmpty(THIS_CLASS);
		return *pCurrent;
	} /*Current()*/
	//------------------------------------------------------------------------
	const TRecord& CurrentConst() const
	{
		if (!pCurrent) 
			throw TXBufferIsEmpty(THIS_CLASS);
		return *pCurrent;
	} /*CurrentConst() const*/
    //------------------------------------------------------------------------
    TRecord* CurrentPtr() {return pCurrent;}
    //------------------------------------------------------------------------
    const TRecord* CurrentPtrConst() const {return pCurrent;}

	//------------------------------------------------------------------------
	void ForEach(TOperationOn<TRecord>&);
	//------------------------------------------------------------------------
	void ForEach(TConstantOperationOn<TRecord>&) const;
	//------------------------------------------------------------------------
	void ForEachConst(TConstantOperationOn<TRecord>& r_Operation) const
	{
		ForEach(r_Operation);
	} /*ForEachConst(TConstantOperationOn<TRecord>&) const*/

    //------------------------------------------------------------------------
	operator TRecord&() {return Current();}
    //------------------------------------------------------------------------
	operator const TRecord&() const {return CurrentConst();}
    //------------------------------------------------------------------------
	operator TRecord*() {return CurrentPtr();}
    //------------------------------------------------------------------------
	operator const TRecord*() const {return CurrentPtrConst();}

protected:
	//------------------------------------------------------------------------
	TRecord* pCurrent;
	nl_ulong_t nRecords;

	//------------------------------------------------------------------------
	NLOBJECT_CLONE_IS_NOT_ALLOWED(TSafeBuffer)
} /*TSafeBuffer*/;

//============================================================================
NATLIB_NAMESPACE_END

#endif /*!defined(__NATLIB_TASKS_NTSFBUF_H__)*/
