#ifndef __NATLIB_NWTHREAD_H__
#define __NATLIB_NWTHREAD_H__
/****************************************************************************/
/** nwthread.h                                                             **/
/** ---------------------------------------------------------------------- **/
/** C++ interface to MS Win32 multithreading model                         **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** January 25, 1998 by Oleg Kagan                                         **/
/****************************************************************************/

//============================================================================
#include <nwerror.h>
#include <nllock.h>

//============================================================================
#ifndef WIN32 /*MS Win32*/
	#error nwthread.h can be used only for Microsoft Win32 applications
#endif /*!defined(WIN23)*/

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TWin32Thread
//============================================================================
class _NATLIB_DECLSPEC_ TWin32Thread : public TNatlibObject {
public:
	// THandle
    //------------------------------------------------------------------------
    typedef HANDLE THandle;

	// TState
	//------------------------------------------------------------------------
    typedef enum {
    	stateFirst = 0,
	    stateInvalid = stateFirst,  // The object is invalid. Currently
                                    //   this happens only when the
                                    //   operating system is unable to start
                                    //   the thread.
    	stateCreated,  // The class has been created but the thread has not
                       //   been started

        stateRunning,  // The thread is running
        stateSuspended,  // The thread has been suspended
        stateFinished,  // The thread has finished execution
        stateAll
    } TState;

	// TPriority
	//------------------------------------------------------------------------
	typedef enum {
		priorityIdle         = THREAD_PRIORITY_IDLE,          // -15
		priorityLowest       = THREAD_PRIORITY_LOWEST,        //  -2
		priorityBelowNormal  = THREAD_PRIORITY_BELOW_NORMAL,  //  -1
		priorityNormal       = THREAD_PRIORITY_NORMAL,        //   0
		priorityAboveNormal  = THREAD_PRIORITY_ABOVE_NORMAL,  //   1
		priorityHighest      = THREAD_PRIORITY_HIGHEST,       //   2
		priorityTimeCritical = THREAD_PRIORITY_TIME_CRITICAL, //  15
    } TPriority;

	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(
    	TException, TNatlibObject::TException, _NATLIB_DECLSPEC_
    ) /*NLEXCEPT_DECLARE*/;
	// The user called Suspend() on an object before calling Start()
	//------------------------------------------------------------------------
    NLEXCEPT_DECLARE(TXSuspendBeforeRun, TException, _NATLIB_DECLSPEC_);
	// The user called Resume() on an object before calling Start()
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXResumeBeforeRun, TException, _NATLIB_DECLSPEC_);
	// The user called Resume() on a thread that was not stateSuspended
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXResumeDuringRun, TException, _NATLIB_DECLSPEC_);
	// The user called Suspend() on an object whose thread had already exited
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXSuspendAfterExit, TException, _NATLIB_DECLSPEC_);
	// The user called Resume() on an object whose thread had already exited
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXResumeAfterExit, TException, _NATLIB_DECLSPEC_);
	// The operating system was unable to create the thread
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXCreationFailure, TException, _NATLIB_DECLSPEC_);
	// The object's destructor was invoked its thread had exited
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXDestroyBeforeExit, TException, _NATLIB_DECLSPEC_);
	// An attempt was made to assign to an object that was not in either the
	// stateCreated or the stateFinished state.
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXAssignError, TException, _NATLIB_DECLSPEC_);
	// Error while killing thread
	//------------------------------------------------------------------------
	NLEXCEPT_DECLARE(TXUnableToKillThread, TException, _NATLIB_DECLSPEC_);

    // Attach to the current running (often primary) thread
    //------------------------------------------------------------------------
    typedef enum {CurrentThread} TCurrentThread;
    TWin32Thread(TCurrentThread);

	// Destroys the object
	//------------------------------------------------------------------------
	virtual ~TWin32Thread();

	//------------------------------------------------------------------------
	NLOBJECT_DECLARE_CLONE(TWin32Thread) = 0;

    // Begins execution of the thread. Returns the handle of the thread
	//------------------------------------------------------------------------
    THandle Start();

	// Suspends execution of the thread
    //------------------------------------------------------------------------
    DWORD Suspend();

    // Resumes execution of a suspended thread
    //------------------------------------------------------------------------
    DWORD Resume();

    // Put this tread to sleep for a given number of milliseconds, with an 
	// optional wakeup on an IO completion. Return true if wakeup is due to 
	// IO completion.
	//------------------------------------------------------------------------
    bool Sleep(DWORD uTimeout, bool bAlertable = false);

    // Sets an internal flag that indicates that the thread should exit.
	// The derived class can check the state of this flag by calling
	// ShouldTerminate().
	//------------------------------------------------------------------------
    virtual void Terminate();

	// Terminates thread unconditianaly (using TerminateThread API call)
	// Note: See API documentation (TerminateThread) for more information
	//------------------------------------------------------------------------
	void Kill(DWORD uExitCode);

	// Blocks the calling thread until the internal thread exits or until the
	// time specified by timeout, in milliseconds,expires. A timeout of
	// TLock::c_uWaitInfinite says to wait indefinitely
	//------------------------------------------------------------------------
    DWORD WaitForExit(nl_ulong_t uTimeout = TLock::c_uWaitInfinite);

    // Combines the behavior of Terminate() and WaitForExit(). Sets an
    // internal flag that indicates that the thread should exit and blocks
    // the calling thread until the internal thread exits or until the time
    // specified by timeout, in milliseconds, expires. A timeout of
    // TLock::c_wWaitInifinite says to wait indefinitely
    //------------------------------------------------------------------------
    DWORD TerminateAndWait(nl_ulong_t uTimeout = TLock::c_uWaitInfinite)
	{
		Terminate(); return WaitForExit(uTimeout);
	} /*TerminateAndWait(nl_ulong_t = TLock::c_uWaitInfinite)*/;

	// Gets the current status of the thread. See TWin32Thread::Status for
    // possible values
	//------------------------------------------------------------------------
    TState GetState() const
	{
		// If the thread is marked as stateRunning it may have terminated 
		// without our knowing it, so we have to check
		//-----------------------------------------------------------
		if (State == stateRunning)
			const_cast<TWin32Thread*>(this)->State = CheckState();
		return State;
	} /*GetState() const*/

	//------------------------------------------------------------------------
    DWORD GetExitCode() const;

	// Gets the priority of the thread
    //------------------------------------------------------------------------
    nl_int_t GetPriority() const;
    // Sets the priority of the thread. Can pass a TPriority for simplicity.
    //------------------------------------------------------------------------
    void SetPriority(nl_int_t);   

protected:
	// Creates an object of type TWin32Thread
	//------------------------------------------------------------------------
    TWin32Thread() :
		uThreadId(0), hThread(null), State(stateCreated),
		bTerminationRequested(false), bAttached(false)
	{} /*TWin32Thread()*/

	// Puts the object into the stateCreated state, just like the default 
	// constructor. Does not copy any of the internal details of the thread 
	// being copied
	//------------------------------------------------------------------------
    TWin32Thread(const TWin32Thread&):
		uThreadId(0), hThread(null), State(stateCreated),
		bTerminationRequested(false), bAttached(false)
	{} /*TWin32Thread(const TWin32Thread&)*/
	
	// Returns a non-zero value to indicate that Terminate() or 
	// TerminateAndWait() has been called. If this capability is being used, 
	// the thread should call ShouldTerminate() regularly, and if it returns 
	// a non-zero value the thread finish its processing and exit
	//------------------------------------------------------------------------
    bool ShouldTerminate() const {return bTerminationRequested;}

	// Alternative to returning from Run()
	//------------------------------------------------------------------------
    void Exit(DWORD uCode) 	
	{
		State = stateFinished; ::ExitThread(uCode);
	} /*Exit(DWORD)*/

	//------------------------------------------------------------------------
	virtual void EvExecutionCompleted();

	// The target object must be in either the stateCreated state or the 
	// stateFinished state. If so, puts the object into the stateCreated 
	// state. If the object is not in either the stateCreated state or the 
	// stateFinished state it is an error and an exception will be thrown
	//------------------------------------------------------------------------
	const TWin32Thread& operator=(const TWin32Thread&);

private:
    //------------------------------------------------------------------------
	static DWORD WINAPI Execute(void*);

    //------------------------------------------------------------------------
	DWORD uThreadId;
    THandle hThread;
    volatile mutable TState State;

	//------------------------------------------------------------------------
    volatile bool bTerminationRequested;
    volatile bool bAttached;

	// The function that does the work. Calling Start() creates a thread that 
	// begins executing Run() with the 'this' pointer pointing to the 
	// TWin32Thread-based object
	//------------------------------------------------------------------------
    virtual nl_int_t Run();         

    //------------------------------------------------------------------------
	TState CheckState() const;
} /*TWin32Thread*/;

//============================================================================
NATLIB_NAMESPACE_END

#endif  /*!defined(__NATLIB_NWTHREAD_H__)*/
