/****************************************************************************/
/** ntsched.cpp                                                            **/
/** ---------------------------------------------------------------------- **/
/** Natlib Tasks non-preemptive scheduler                                  **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** September 01, 1997 by Oleg Kagan                                       **/
/****************************************************************************/

//============================================================================
#include <nlcompat.h>
#pragma hdrstop

// Natlib Basics includes
//============================================================================
#include <nlstring.h>

// Natlib Tasks includes
//============================================================================
#include <ntsched.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TScheduler::TTaskByNameComparator implemtation
//============================================================================
TComparator::TResult TScheduler::TTaskByNameComparator::operator()(

	const TTask::THandle& c_hTask

) const {
	if (!(c_hTask.IsTargetValid() && Etalon().IsTargetValid()))
    	return compUndefined;
	if (Etalon() == c_hTask) return compEqual;

	const TStdString* pc_sEtalonName = &(Etalon()->Name());
    const TStdString* pc_sTaskName = &(c_hTask->Name());

    if (pc_sEtalonName->length() < pc_sTaskName->length())
    	return compLess;
    if (pc_sEtalonName->length() > pc_sTaskName->length())
    	return compGreater;

    TStdStringCaseSensitiveComparator NameComparator;
    NameComparator.SetEtalon(Etalon()->Name());
    return NameComparator(c_hTask->Name());
} /*TScheduler::TTaskByNameComparator::operator()(const TTask::THandle&)*/

// TScheduler implementation
//============================================================================
TScheduler::TScheduler() :
	TSafeHandleAllowed(),
	TAllocationAware(this)
{} /*TScheduler()*/

//----------------------------------------------------------------------------
TScheduler::~TScheduler()
{} /*~TScheduler()*/

//----------------------------------------------------------------------------
void TScheduler::EvTransaction()
{
    PerformTasksStartups();
    DispatchTasksSignals();
	PerformTasksTransactions();
    PerformTasksTermination();
} /*TScheduler::EvTransaction()*/

//----------------------------------------------------------------------------
bool TScheduler::IsFinished()
{
	return (
		StartupQueue.IsEmpty() && 
		ActiveQueue.IsEmpty() && 
		(IdleQueue.IsEmpty() || SignalQueue.IsEmpty()) &&
		TerminationQueue.IsEmpty()
	) /*return*/;
} /*TScheduler::IsFinished()*/

//----------------------------------------------------------------------------
void TScheduler::Start(

	TTask& r_Task,
    bool bAutodelete /*= true*/

){
	if (TaskSchedulerPtr(r_Task) || IsTaskRegistered(r_Task)) 
		throw TXTaskAlreadyStarted(THIS_CLASS);
	SetTaskScheduler(r_Task, THandle(*this));
	InsertTask(r_Task, StartupQueue, bAutodelete);
} /*TScheduler::Start(TTask&, bool = true)*/

//----------------------------------------------------------------------------
void TScheduler::Terminate(

	const TTask::THandle& c_hTask,
    bool bUnconditional /*= false*/

){
#pragma message ("bUnconditional is not used (Continue here)")
	if (!c_hTask.IsTargetValid())
		throw TXInvalidParameter(THIS_CLASS);

	TTask* pTask = &c_hTask.Dereference();
	InsertTask(*pTask, TerminationQueue, pTask->IsAutodelete());
} /*TScheduler::Terminate(const TTask::THandle&, bool = false)*/

//----------------------------------------------------------------------------
void TScheduler::TerminateAll()
{
	StartupQueue.Clear();
	TerminationQueue.MoveFrom(ActiveQueue);
	TerminationQueue.MoveFrom(IdleQueue);
} /*TScheduler::TerminateAll()*/

//----------------------------------------------------------------------------
void TScheduler::Send(

	TSignal& r_Signal,
	const TTask::THandle& c_hTarget

){
	SetSignalDestination(r_Signal, c_hTarget);
	InsertSignal(r_Signal, SignalQueue);
} /*TScheduler::Send(TSignal&, const TTask::THandle&)*/

//----------------------------------------------------------------------------
void TScheduler::Send(

	TSignal& r_Signal,
	const TContainerOf<TTask::THandle>& c_Targets

){
	SetSignalDestination(r_Signal, c_Targets);
	InsertSignal(r_Signal, SignalQueue);
} /*TScheduler::Send(TSignal&, const TConrainerOf<TTask::THandle>&)*/

//----------------------------------------------------------------------------
void TScheduler::InterfaceStart(

	TSchedulerInterface& r_Interface,
	TTask& r_Task,
	bool bAutodelete /*= true*/

){
	throw TNatlibException(THIS_CLASS);
} /*TScheduler::InterfaceStart(TSchedulerInterface&, TTask&, bool = true)*/;

//----------------------------------------------------------------------------
void TScheduler::InterfaceSend(

	TSchedulerInterface& r_Interface,
	TSignal& r_Signal,
	const TTask::THandle& c_hTarget

){
	throw TNatlibException(THIS_CLASS);
} /*TScheduler::InterfaceSend(TSchedulerInterface&, TSignal&, ...)*/;

//----------------------------------------------------------------------------
void TScheduler::InterfaceSend(

	TSchedulerInterface& r_Interface, 
	TSignal& r_Signal, 
	const TContainerOf<TTask::THandle>& c_Targets

){
	throw TNatlibException(THIS_CLASS);
} /*TScheduler::InterfaceSend(TSchedulerInterface&, TSignal&, ...)*/;

//----------------------------------------------------------------------------
void TScheduler::InterfaceShowTaskException(

	TSchedulerInterface& r_Interface, 
	TTask& r_Task, 
	TTask::TPhase Phase, 
	const TNatlibException* pc_Exception /*= null*/
    
){
	throw TNatlibException(THIS_CLASS);
} /*TScheduler::InterfaceShowTaskException(TSchedulerInterface&, ...)*/;

//----------------------------------------------------------------------------
void TScheduler::ShowTaskException(

	TTask& /*r_Task*/,
    TTask::TPhase /*Phase*/,
    const TNatlibException* /*pc_Exception*/ /*= null*/

){
} /*TScheduler::ShowTaskException(TTask&, TPhase, const TNatlibException*)*/

//----------------------------------------------------------------------------
bool TScheduler::IsTaskStarted(

	TTask& r_Task

) const {
	return (
	    TParent::IsTaskRegistered(r_Task, IdleQueue) ||
		TParent::IsTaskRegistered(r_Task, ActiveQueue) 
    ) /*return*/;
} /*TScheduler::IsTaskStarted(TTask&) const*/

//----------------------------------------------------------------------------
void TScheduler::PerformTasksStartups()
{
	nl_ulong_t nPass = StartupQueue.Size();
	do {
		TTask* pTask = null;
		try {
			while (nPass) {
				--nPass;
				TTask* pCurrentTask = &CurrentTask(StartupQueue);
    			CallTaskStartup(*(pTask = pCurrentTask)); pTask = null;
		        if (
					StartupQueue.Size() && 
					(pCurrentTask == &CurrentTask(StartupQueue))
				){
                	InsertTask(
                    	*pCurrentTask, ActiveQueue,
                        pCurrentTask->IsAutodelete()
                    ) /*InsertTask*/;
                } /*if*/
			} /*while*/
		} /*try*/
		catch (TXServiceRequest& r_XServiceRequest) {
			if (!pTask) throw;
    		ProcessXServiceRequest(
        		*pTask, TTask::phaseStartup, r_XServiceRequest
			) /*ProcessXServiceRequest*/;
		} /*catch (TXServiceRequest&)*/
		catch (TNatlibException& r_Exception) {
			if (!pTask) throw;
    		CallTaskException(*pTask, TTask::phaseStartup, &r_Exception);
		} /*catch (TNatlibException&)*/
		catch (...) {
			if (!pTask) throw;
    		CallTaskException(*pTask, TTask::phaseStartup, null);
		} /*catch (...)*/
	} while (nPass);
} /*TScheduler::PerformTasksStartups()*/

//----------------------------------------------------------------------------
void TScheduler::DispatchTasksSignals()
{
	for (nl_ulong_t nPass = SignalQueue.Size(); nPass; --nPass) {
    	TSignal* pSignal = &CurrentSignal(SignalQueue);
		TVectorOf<TTask::THandle>* pDestination = (
			&GetSignalDestination(*pSignal)
		) /*pDestination*/;

        nl_ulong_t nTask = 0, nDestinations = pDestination->Size();
		do {
			TTask* pTask = null;
			try {
				while (nTask < nDestinations) {
					TTask::THandle hTask = pDestination->RawElement(nTask++);
					if (hTask.IsTargetValid()) {
            			TTask* pCurrentTask = &hTask.Dereference();
						if (IsTaskStarted(*pCurrentTask)) {
							CallTaskTransaction(
								*(pTask = pCurrentTask), pSignal
							) /*CallTaskTransaction*/; pTask = null;
						} /*if*/
            		} /*if*/
				} /*while*/
			} /*try*/
			catch (TXServiceRequest& r_XServiceRequest) {
				if (!pTask) throw;
    			ProcessXServiceRequest(
        			*pTask, TTask::phaseTransaction, r_XServiceRequest
				) /*ProcessXServiceRequest*/;
			} /*catch (TXServiceRequest&)*/
			catch (TNatlibException& r_Exception) {
				if (!pTask) throw;
    			CallTaskException(
					*pTask, TTask::phaseTransaction, &r_Exception
				) /*CallTaskException*/;
			} /*catch (TNatlibException&)*/
			catch (...) {
				if (!pTask) throw;
    			CallTaskException(*pTask, TTask::phaseTransaction, null);
			} /*catch (...)*/
		} while (nTask < nDestinations);

		pSignal->Unregister();
        if (pSignal->IsDynamic()) delete pSignal;
	} /*for (nPass)*/
} /*TScheduler::DispatchTasksSignals()*/

//----------------------------------------------------------------------------
void TScheduler::PerformTasksTransactions()
{
	nl_ulong_t nPass = nl_ulong_t(ActiveQueue.Size());
    do {
		TTask* pTask = null;
    	try {
			while (nPass && ActiveQueue.Size()) {
            	--nPass;
   				TTask* pCurrentTask = &CurrentTask(ActiveQueue);
				CallTaskTransaction(*(pTask = pCurrentTask), null); 
				pTask = null;
				if (
					ActiveQueue.Size() &&
					(pCurrentTask == &CurrentTask(ActiveQueue))
				) ActiveQueue.Next();
	    	} /*while (nPass)*/
	    } /*try*/
	    catch (TXServiceRequest& r_XServiceRequest) {
    		if (!pTask) throw;
		    ProcessXServiceRequest(
    	    	*pTask, TTask::phaseTransaction, r_XServiceRequest
			) /*ProcessXServiceRequest*/;
	    } /*catch (TXServiceRequest&)*/
	    catch (TNatlibException& r_Exception) {
    		if (!pTask) throw;
	    	CallTaskException(*pTask, TTask::phaseTransaction, &r_Exception);
	    } /*catch (TNatlibException&)*/
		catch (...) {
	    	if (!pTask) throw;
	    	CallTaskException(*pTask, TTask::phaseTransaction, null);
	    } /*catch (...)*/
    } while (nPass && ActiveQueue.Size());
} /*TScheduler::PerformTasksTransactions()*/

//----------------------------------------------------------------------------
void TScheduler::PerformTasksTermination()
{
	nl_ulong_t nPass = nl_ulong_t(TerminationQueue.Size());
    do {
		TTask* pTask = null;
    	try {
			while (nPass && TerminationQueue.Size()) {
            	--nPass;
   				pTask = &CurrentTask(TerminationQueue);
				CallTaskTerminate(*pTask); pTask = null;
        		DeleteTask(CurrentTask(TerminationQueue), TerminationQueue);
	    	} /*while (nPass)*/
	    } /*try*/
	    catch (TXServiceRequest& r_XServiceRequest) {
			if (!pTask) throw;
    		ProcessXServiceRequest(
	        	*pTask, TTask::phaseTerminate, r_XServiceRequest
			) /*ProcessXServiceRequest*/;
	    } /*catch (TXServiceRequest&)*/
	    catch (TNatlibException& r_Exception) {
			if (!pTask) throw;
	    	CallTaskException(*pTask, TTask::phaseTerminate, &r_Exception);
		} /*catch (TNatlibException&)*/
		catch (...) {
			if (!pTask) throw;
			CallTaskException(*pTask, TTask::phaseTerminate, null);
	    } /*catch (...)*/
    } while (nPass && TerminationQueue.Size());
} /*TScheduler::PerformTasksTermination()*/

//------------------------------------------------------------------------
bool TScheduler::CallTaskCanTerminate(

	TTask& r_Task

){
	try {
		return TParent::CallTaskCanTerminate(r_Task);
    } /*try*/
    catch (TXServiceRequest& r_XServiceRequest) {
    	ProcessXServiceRequest(
        	r_Task, TTask::phaseCanTerminate, r_XServiceRequest
        ) /*ProcessXServiceRequest*/;
    } /*catch (TXServiceRequest&)*/
    catch (TNatlibException& r_Exception) {
    	CallTaskException(r_Task, TTask::phaseCanTerminate, &r_Exception);
    } /*catch (TNatlibException&)*/
	catch (...) {
    	CallTaskException(r_Task, TTask::phaseCanTerminate, null);
    } /*catch (...)*/
    return true;
} /*TScheduler::CallTaskCanTerminate(TTask&)*/

//----------------------------------------------------------------------------
void TScheduler::CallTaskException(

    TTask& r_Task,
    TTask::TPhase Phase,
    TNatlibException* pException

){
	try {
    	TParent::CallTaskException(r_Task, Phase, pException);
    } /*try*/
    catch (TXServiceRequest& r_XServiceRequest) {
    	ProcessXServiceRequest(
        	r_Task, TTask::phaseException, r_XServiceRequest
        ) /*ProcessXServiceRequest*/;
    } /*catch (TXServiceRequest&)*/
    catch (TNatlibException& r_Exception) {
    	ShowTaskException(r_Task, TTask::phaseException, &r_Exception);
        Terminate(TTask::THandle(r_Task));
    } /*catch (TNatlibException&)*/
    catch (...) {
		ShowTaskException(r_Task, TTask::phaseException, null);
        Terminate(TTask::THandle(r_Task));
    } /*catch (...)*/
} /*TScheduler::CallTaskException(TTask&, TPhase, const TNatlibException*)*/

//----------------------------------------------------------------------------
void TScheduler::ProcessXServiceRequest(

	TTask& r_Task,
    TTask::TPhase Phase,
    TXServiceRequest& r_Request

){
	union {
		TXWaitRequest* pXWaitRequest;
		TXEndTransactionRequest* pXEndTransactionRequest;
		TXTerminateRequest* pXTerminateRequest;
		TXGlobalTerminateRequest* pXGlobalTerminateRequest;
	} /*union*/;

	if (
		(pXWaitRequest = 
			dynamic_cast<TXWaitRequest*>(&r_Request)
		) /*pXWaitRequest*/ != null
	){
		if (Phase == TTask::phaseTerminate) return;
		InsertTask(r_Task, IdleQueue, r_Task.IsAutodelete());
	}
	else if (
		(pXEndTransactionRequest = 
			dynamic_cast<TXEndTransactionRequest*>(&r_Request)
		) /*pXEndTransactionRequest*/ != null
	){
		// Ignore
	}
	else if (
		(pXTerminateRequest = 
			dynamic_cast<TXTerminateRequest*>(&r_Request)
		) /*pXTerminateRequest*/ != null
	){
		if (Phase == TTask::phaseTerminate) return;
		if (pXTerminateRequest->Value() || CallTaskCanTerminate(r_Task)) {
			InsertTask(r_Task, TerminationQueue, r_Task.IsAutodelete());
		} /*if*/
	}
	else if (
		(pXGlobalTerminateRequest =
			dynamic_cast<TXGlobalTerminateRequest*>(&r_Request)
		) != null
	){
		TerminateAll();
	} 
	else {
		throw TXUnknownServiceRequest(THIS_CLASS);
	} /*iff*/
} /*TScheduler::ProcessXServiceRequest(TXServiceRequest&)*/

//============================================================================
NATLIB_NAMESPACE_END

