/****************************************************************************/
/** ntsfbuf.h                                                              **/
/** ---------------------------------------------------------------------- **/
/** Safe cyclic buffer (queue)                                             **/
/** ---------------------------------------------------------------------- **/
/** Copyright (C) 1996-98 by OK Thinking Software, All Rights Reserved     **/
/** ---------------------------------------------------------------------- **/
/** August 18, 1997 by Oleg Kagan                                          **/
/****************************************************************************/

//============================================================================
#include <nlcompat.h>
#pragma hdrstop

// Natlib Tasks includes
//============================================================================
#include <ntsfbuf.h>

//============================================================================
NATLIB_NAMESPACE_BEGIN

// TSafeBuffer implementation
//============================================================================
void TSafeBuffer::AssertValid() const
{
	if (pCurrent) {
		const TRecord* pc_Record = pCurrent;
	    size_t nCounted = 0;
	    do {
			pc_Record = pc_Record->pNext;
	        if (++nCounted > nRecords) 
				throw TXInvalidChain(THIS_CLASS);
	    } while (pc_Record != pCurrent);
	    if (nCounted != nRecords) 
			throw TXInvalidChain(THIS_CLASS);

        pc_Record = pCurrent;
	    nCounted = 0;
	    do {
			pc_Record = pc_Record->pPrevious;
	        if (++nCounted > nRecords) 
				throw TXInvalidChain(THIS_CLASS);
	    } while (pc_Record != pCurrent);
	    if (nCounted != nRecords) 
			throw TXInvalidChain(THIS_CLASS);
    }
    else {
    	if (nRecords) 
			throw TXInvalidChain(THIS_CLASS);
    } /*iff*/
} /*TSafeBuffer::AssertValid() const*/

//----------------------------------------------------------------------------
TNatlibStdString TSafeBuffer::Dump() const
{
	TStdString sDump(TParent::Dump());
	// Add size
    //----------
    {
	    char szDump[256];
	    sprintf(szDump, " size: %lu", (unsigned long)Size());
	    sDump += szDump;
    } /*Add size*/
    return sDump;
} /*TSafeBuffer::Dump() const*/

//----------------------------------------------------------------------------
void TSafeBuffer::Register(

	TSafeBuffer::TRecord& r_Record,
	bool bAutodelete /*= true*/

){
	r_Record.Unregister();

	// Insert
	//--------
	r_Record.pBuffer = this;
	r_Record.bAutodelete = bAutodelete;

	if (pCurrent) {
		r_Record.pNext = pCurrent;
		r_Record.pPrevious = pCurrent->pPrevious;;

		pCurrent->pPrevious->pNext = &r_Record;
        pCurrent->pPrevious = &r_Record;
    }
    else {
    	pCurrent = r_Record.pNext = r_Record.pPrevious = &r_Record;
    } /*if*/
    ++nRecords;
} /*TSafeBuffer::Register(TSafeBuffer::TRecord&, bool = true)*/

//----------------------------------------------------------------------------
void TSafeBuffer::Unregister(

	TRecord& r_Record

){
	if (!r_Record.IsRegistered(*this)) 
		throw TXRecordIsNotRegistered(THIS_CLASS);

	if (r_Record.pNext != &r_Record) {
		if (pCurrent == &r_Record) Next();
		r_Record.pPrevious->pNext = r_Record.pNext;
		r_Record.pNext->pPrevious = r_Record.pPrevious;
	}
	else {
		pCurrent = null;
	} /*iff*/

	// Clear pointers in record
	//--------------------------
	r_Record.pBuffer = null;
	r_Record.pNext = r_Record.pPrevious = null;
	--nRecords;
} /*TSafeBuffer::Unregister(TRecord&)*/

//----------------------------------------------------------------------------
void TSafeBuffer::Delete(

	TRecord& r_Record
	
){
	Unregister(r_Record);
	if (r_Record.bAutodelete && r_Record.IsDynamic()) delete &r_Record;
} /*TSafeBuffer::Delete(TRecord&)*/

//----------------------------------------------------------------------------
void TSafeBuffer::MoveFrom(

	TSafeBuffer& r_Source

){
	while (r_Source.Size()) Insert(r_Source.Current());
} /*TSafeBuffer::MoveFrom(TSafeBuffer&)*/

//----------------------------------------------------------------------------
void TSafeBuffer::Clear()
{
	while (pCurrent) Delete(*pCurrent);
} /*TSafeBuffer::Clear()*/

//----------------------------------------------------------------------------
void TSafeBuffer::ForEach(

	TOperationOn<TSafeBuffer::TRecord>& r_Operation

){
	try {
		TRecord* pRecord = CurrentPtr();
		for (nl_ulong_t nPass = Size(); nPass; --nPass) {
			r_Operation.operator()(*pRecord);
			pRecord = pRecord->pNext;			
		} /*for (nPass)*/
	} /*try*/
	catch (const TOperation::TXAbort&) {}
} /*TSafeBuffer::ForEach(TOperationOn<TSafeBuffer::TRecord>&)*/

//----------------------------------------------------------------------------
void TSafeBuffer::ForEach(

	TConstantOperationOn<TSafeBuffer::TRecord>& r_Operation

) const {
	try {
		const TRecord* pc_Record = CurrentPtrConst();
		for (nl_ulong_t nPass = Size(); nPass; --nPass) {
			r_Operation.operator()(*pc_Record);
			pc_Record = pc_Record->pNext;			
		} /*for (nPass)*/
	} /*try*/
	catch (const TOperation::TXAbort&) {}
} /*TSafeBuffer::ForEach(TConstantOperationOn<TSafeBuffer::TRecord>&)*/

// TSafeBuffer::TRecord implementation
//============================================================================
TSafeBuffer::TRecord::TRecord() :
	TAllocationAware(this),
	pBuffer(null), bAutodelete(false), pNext(null), pPrevious(null)
{} /*TSafeBuffer::TRecord()*/

//----------------------------------------------------------------------------
void TSafeBuffer::TRecord::Unregister()
{
	if (pBuffer) {
		pBuffer->Unregister(*this);
		if (pBuffer) 
			throw TXUnableToUnregister(THIS_CLASS);
	} /*if*/
} /*TSafeBuffer::TRecord::Unregister()*/

//============================================================================
NATLIB_NAMESPACE_END
